/*
WaveTomography software implements the algorithm for solving direct and inverse problems of wave tomography in application to medical ultrasound imaging for breast cancer diagnosis.
The software has been developed at Lomonosov Moscow State University. Copyright (c) 2021 Seryozhnikov S.Y, Romanov S.Y.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "k_defs.h"
#include "cmdl.h"
#include "cdata.h"
#include "parallel.h"
#include <math.h>
#include <omp.h>
#include <cctype>
#ifndef TARGET_WINDOWS
#include <unistd.h>
#endif
#include "cdata_hlp.cpp"

#ifdef USE_MPI
#include <mpi.h>
#endif

#define CPU_GRAD_SCALE 65536.0f

static const OPTDEF DatasetOptions[] = { // loaded from <datasetname>.ini
{"-size", OPTDEF::T_INT, 1, offsetof(CData, ReqW)  },
{ "-nsources", OPTDEF::T_INT,   1,  offsetof(CData,GlobalSrcs) },
{ "-nframes",  OPTDEF::T_INT,   1,  offsetof(CData,NFrames) },
{ "",-1,0,0 } };

static const OPTDEF _opt[] = {
{"-size",     OPTDEF::T_INT,   1,  offsetof(CData,ReqW)  },
{"-nsources", OPTDEF::T_INT,   1,  offsetof(CData,GlobalSrcs) },
{"-impulse",  OPTDEF::T_FLOAT, 2,  offsetof(CData,ImpWidth)     },
{"-impstart", OPTDEF::T_FLOAT, 1,  offsetof(CData,ImpStart)     },
{"-impclip",  OPTDEF::T_FLOAT, 2,  offsetof(CData,ImpClipAMin)  },
{"-v0",       OPTDEF::T_FLOAT, 1,  offsetof(CData,Vel0)  },
{"-dx",       OPTDEF::T_FLOAT, 1,  offsetof(CData,dx0)  },
{"-dt",       OPTDEF::T_FLOAT, 1,  offsetof(CData,dt0)  },
{"-nframes",  OPTDEF::T_INT,   1,  offsetof(CData,NFrames) },
{"-LAPL",     OPTDEF::T_FLOAT, 5,  offsetof(CData,LAPL_BASE[1]) },
{"-absorb",   OPTDEF::T_FLOAT,  1,  offsetof(CData,AbsorbA) },
{"-TimeBuffer", OPTDEF::T_FLOAT,     1,  offsetof(CData,TimeBuf) },
{"-SrcRadius", OPTDEF::T_FLOAT,     1,  offsetof(CData,SrcRadius) },
{"-RcvRadius", OPTDEF::T_FLOAT,     1,  offsetof(CData,RcvRadius) },
{"-RcvSpacing", OPTDEF::T_FLOAT,     1,  offsetof(CData,RcvSpacing) },

{"-PhantomSize", OPTDEF::T_FLOAT,     1,  offsetof(CData,PhantomSize) },
{"-PhantomContrast", OPTDEF::T_FLOAT, 1,  offsetof(CData,PhantomContrast) },
{"-SolutionRange", OPTDEF::T_FLOAT, 2,  offsetof(CData,GradCropR0) },
{"-GradFadeOutRadius", OPTDEF::T_FLOAT, 1,  offsetof(CData,TFadeRadius) },
{"-GradFadeInRange", OPTDEF::T_FLOAT, 2,  offsetof(CData,GradStartR0) },
{"-GradFalloffPerIter", OPTDEF::T_FLOAT, 1,  offsetof(CData,GradEraseAtUpdate) },

{"-TargetGrad", OPTDEF::T_FLOAT, 1,  offsetof(CData,TargetGradAmpl)},
{"-StepUpCoeff", OPTDEF::T_FLOAT,     1,  offsetof(CData,StepUpCoeff) },
{"-StepDownCoeff", OPTDEF::T_FLOAT,     1,  offsetof(CData,StepDownCoeff) },
{"-MinThreshold", OPTDEF::T_FLOAT, 1,  offsetof(CData,MinThreshold) },

{"-LimitVel", OPTDEF::T_FLOAT, 2,  offsetof(CData,Limit_C) },
{"-LimitAtt", OPTDEF::T_FLOAT, 1,  offsetof(CData,Limit_A) },
{"-AttGradScale", OPTDEF::T_FLOAT, 1,  offsetof(CData,AttGradScale) },

{"-mode",     OPTDEF::T_INT,   1,  offsetof(CData,Mode)  },
{"-dump",     OPTDEF::T_INT,   1,  offsetof(CData,nDump)  },

{"-run",      OPTDEF::T_INT,   2,  offsetof(CData,RunMethod) },
{"-memshear", OPTDEF::T_INT,   6,  offsetof(CData,MemoryShear) },
{"-memalign", OPTDEF::T_INT,   1,  offsetof(CData,MemAlign) },
{"-ialign",   OPTDEF::T_INT,   2,  offsetof(CData,ImgAlignX) },
{"-nthreads", OPTDEF::T_INT,   1,  offsetof(CData,NThreads) },
{"-CellY",    OPTDEF::T_INT,   2,  offsetof(CData,BLKs.FwdY) },
{"-CellW",    OPTDEF::T_INT,   2,  offsetof(CData,BLKs.FwdW) },
{"-vector",   OPTDEF::T_INT,   1,  offsetof(CData,VectorSize) },

{"-direct",  OPTDEF::T_STR64, 1,  offsetof(CData,DirectBaseName) },
{"-inverse",  OPTDEF::T_STR64, 1,  offsetof(CData,InverseBaseName) },
{"-restart",  OPTDEF::T_STR64, 1,  offsetof(CData,RestartName) },
{"-MaxIter",  OPTDEF::T_INT, 1,  offsetof(CData,MaxIterations) },
{"-SaveIter",  OPTDEF::T_INT, 1,  offsetof(CData,SaveIterations) },

{"-CPUMemory", OPTDEF::T_INT,   1,  offsetof(CData,CPUMemorySizeMB) },
{"-autotune", OPTDEF::T_INT,   1,  offsetof(CData,fAutoTune) },
{"-DemoMode", OPTDEF::T_INT,   1,  offsetof(CData,fDemoMode) },

{"-perftest", OPTDEF::T_INT,   2,  offsetof(CData,PerfTest) },
{"-SkipScalars", OPTDEF::T_INT, 1,  offsetof(CData,fSkipScalars) },
{"-TestIncrSize", OPTDEF::T_INT,   2,  offsetof(CData,TestIncrSize) },
{"-TestIncrSrc", OPTDEF::T_INT,   1,  offsetof(CData,TestIncrSrc) },
{"-TestVolume",  OPTDEF::T_FLOAT, 1,  offsetof(CData,TestVolume) },
{"-TestDuration",  OPTDEF::T_FLOAT, 1,  offsetof(CData,TestDuration) },
{"-TestSizes",  OPTDEF::T_INT, 3,  offsetof(CData,TestSizes[0]) },

{"-WaveExportFormat",  OPTDEF::T_STR64, 1,  offsetof(CData,WaveExportFmtNames) },
{"-ImageExportFormat",  OPTDEF::T_STR64, 1,  offsetof(CData,ImageExportFmtNames) },
{"-PBMScale",           OPTDEF::T_FLOAT, 2,  offsetof(CData,PBM_CLim) },


{"",-1,0,0} };

const OPTDEF *const CData::Options = _opt;
const char* const CData::AutotuneFile = "autotune.dat";
const char* const CData::PerftestFile = "perftest.dat";
const char* const CData::INIFile = "wavetm.ini";

inline int CollectRetval(int in)
{
	int ret = in;
#ifdef USE_MPI
	MPI_Allreduce(&in, &ret, 1, MPI_INT, MPI_BOR, MPI_COMM_WORLD);
#endif
	if ((in!=0)&&(in!=E_FINISHED)) printf("%s Error code %i ", _PID, in);
	return ret;
}

// Clamp the block size to GPU capabilities
// input - CellSize* in the struct
// output - B
void CData::ValidateBlockSizes(BlockSizes& B) const
{
	cl_int4 ret;
	B.FwdY = CLAMP(BLKs.FwdY, 5, 80);
	B.BwdY = CLAMP(BLKs.BwdY, 5, 80);
	B.FwdY -= B.FwdY % 5;		B.BwdY -= B.BwdY % 5;

	// must divide the image height
	while ( InH%B.FwdY ) B.FwdY -= 5;
	B.FwdY = __min(B.FwdY, 5);
	while ( InH%B.BwdY ) B.BwdY -= 5;
	B.BwdY = __min(B.BwdY, 5);

#ifdef TARGET_GPU
	if ( EnableGPU && pGPU && pGPU->fReady )
	{// must not exceed  max threads
		B.FwdW = CLAMP(BLKs.FwdW, 8, pGPU->MaxBlockSize);
		B.BwdW = CLAMP(BLKs.BwdW, 8, pGPU->MaxBlockSize);
		// must not exceed local memory
		// some GPU drivers reserve some local memory
		// float8 per thread per Y-line is used by the program
		const int WBLimit = (pGPU->LocalMemSize - 256) / (BLKs.BwdY * 32);
		B.BwdW = __min(B.BwdW, WBLimit);
	}
#endif
}


int DecodeFmtName(const char* name)
{
	static const 	struct { char ext[4]; int mask; }
	S[] = { { "MAT",EXPORT_MAT }, { "PBM",EXPORT_PBM },
					 { "RAW",EXPORT_RAW },{ "RST",CData::SAVE_RST },
	{ "FUL",EXPORT_FULL },
	{"\0"},0 };

	int ret = 0;	if ( !name ) return ret;

	for ( int i = 0; S[i].mask; ++i )
		if ( strstr(name, S[i].ext) ) ret |= S[i].mask;
	return ret;
}

// fill unspecified parameters with defaults 
// or clear all parameters to unspecified (0) 
void CData::SetParameters(bool clear)
{
	if ( clear )
	{
		Vel0 = Vel = 0;
		dx0 = dt0 = dx = dt = 0;
		ImpFreq = ImpWidth = ImpStart = 0;
		ImpClipAMin = ImpClipAMax = 0;
		PhantomSize = SrcRadius = RcvRadius = 0;
		TimeBuf = 0; NFrames = 0;
		AbsorbA = 0; FadeTime = 0;
		return;
	}

	BLKs.SetDefaults();

	WaveExportFormats = DecodeFmtName(WaveExportFmtNames);
	ImageExportFormats = DecodeFmtName(ImageExportFmtNames);

	Vel0 = (Vel0 > 0) ? Vel0 : 1.5;
	Vel = Vel0;

	if ( GlobalSrcs <= 0 ) { GlobalSrcs = 15; }
	{ cl_uint2 res = SelectPart(GlobalSrcs);
	StartGlobalSrc = res.s[0];	NSrcs = res.s[1] - res.s[0];
	}

	const float DZ_Mb = DZ * 1.0e-6f;

#ifdef TARGET_GPU
	if ( EnableGPU )
	{
		HostSrcs = 1; GPUSrcs = floor(GPUMemorySize / DZ_Mb);
		GPUSrcs = CLAMP(GPUSrcs, 1, NSrcs);
	} else
#endif
	{
		GPUSrcs = 0;  HostSrcs = __max(1, floor(CPUMemorySizeMB / DZ_Mb));
		HostSrcs = __min(HostSrcs, NSrcs);
	}

	// negative dx specifies relative to the default
	if ( dx0 == 0 ) dx0 = -1;
	if ( dx0 < 0 ) dx0 *= -FIELD_SIZE_mm / ReqW;

	// negative dt specifies dt/dx
	dt0 *= (dt0 > 0) ? 1 : -dx0;
	if ( (dt0 < dx0*0.05f) || (dt0 > dx0*0.6f) )  dt0 = dx0 * 0.4f;
	dt = dt0; dx = dx0;

	float FramesPerDX = dx / dt / Vel;

	if ( ImpClipAMin <= 0 )ImpClipAMin = 0.73;
	if ( ImpClipAMax <= ImpClipAMin )  ImpClipAMax = 0.95;

	if ( ImpFreq <= 0 ) ImpFreq = 2;
	if ( ImpWidth <= 0 )
	{
		ImpWidth = 9 * dx*ImpFreq;
		ImpWidth = __max(ImpWidth, 1.6*ImpFreq);
	}

	if ( ImpStart <= 0 )	ImpStart = 1.5f*ImpWidth;
	const float rmin = 0.5*dx*__min(ReqW, ReqH);

	// negative R specifies relative to the field size, 0=default
	if ( RcvRadius <= 0 )

	{
		if ( RcvRadius == 0 ) RcvRadius = DefRcvRadius * rmin;
		if ( RcvRadius < 0 ) RcvRadius *= -rmin;
	}
	if ( SrcRadius <= 0 )		SrcRadius = RcvRadius + ImpStart;

	// 0=default
	if ( GradStartR0 == 0.0f ) GradStartR0 = -0.2;// ImpWidth;
	if ( GradStartR1 == 0.0f ) GradStartR1 = -0.4;//ImpWidth * 2;
	// Negative specifies relative to RcvRadius
	if ( GradStartR0 < 0.0f ) GradStartR0 *= -RcvRadius;
	if ( GradStartR1 < 0.0f ) GradStartR1 *= -RcvRadius;
	GradStartR0 = CLAMP(GradStartR0, ImpWidth*0.5, RcvRadius*0.45);
	GradStartR1 = CLAMP(GradStartR1, GradStartR0*1.01, RcvRadius*0.5);

	EndInvTime = GradStartR0 * FramesPerDX / dx;
	GradMixDuration = (GradStartR1 - GradStartR0) * FramesPerDX / dx;

	if ( GradCropR0 == 0 ) GradCropR0 = -0.64;
	if ( GradCropR1 == 0 ) GradCropR1 = -0.8;

	if ( GradCropR0 < 0 ) GradCropR0 *= -RcvRadius;
	if ( GradCropR1 < 0 ) GradCropR1 *= -RcvRadius;
	if ( GradCropR0 >= GradCropR1 ) GradCropR0 = GradCropR1 - 4;
	GradCropR0 = CLAMP(GradCropR0, 0, RcvRadius*0.99);
	GradCropR1 = CLAMP(GradCropR1, GradCropR0 + dx, RcvRadius);


	if ( PhantomSize <= 0 )
	{
		PhantomSize = 0.92*RcvRadius;
		PhantomSize = __min(PhantomSize, 2 * (SrcRadius - 2 * ImpWidth));
		PhantomSize = __min(PhantomSize, 120);
	}

	// Clamp the total phase shift so the auto-generated phantoms are solvable
	if ( PhantomContrast <= 0 )		PhantomContrast = 80.0f / PhantomSize;

	if ( TimeBuf <= 0 ) TimeBuf = 1.35;
	if ( NFrames < 4 )	NFrames = RcvRadius * 2.0f / Vel / dt * TimeBuf;

	if ( AbsorbA < 0 ) AbsorbA = 0.1f;

	if ( TFadeRadius == 0 ) TFadeRadius = -0.16;
	if ( TFadeRadius < 0 ) TFadeRadius *= -RcvRadius;
	if ( FadeTime <= 0 ) FadeTime = 1.0f / (TFadeRadius*dx / dt / Vel);

	// Normalize the Laplacian matrix
	LAPL_BASE[0] = LAPL_BASE[7] = -4.0f*(LAPL_BASE[5] + LAPL_BASE[4] + LAPL_BASE[4] + LAPL_BASE[3] + LAPL_BASE[2] + LAPL_BASE[1]);
	for ( int i = 0; i < 7; ++i )LAPL.s[i] = LAPL_BASE[i] / (dx*dx);

	{ // Boundary condition coefficients
		const double vdt = Vel * dt,
			K1 = (dt*vdt*dx*8.) / (4.*dx + 4.*vdt + AbsorbA * dt*dx),
			K2 = 1. / (dt * dx * 2.),
			K3 = 1. / (vdt*dt * 2.),
			K4 = AbsorbA / (8.*vdt),
			K5 = Vel / (dx*dx * 4.);

		// Accessible by GPU and CPU
		Kuh = (K2 - K3 - K4 - 2 * K5)*K1;
		Kdh = (-K2 - K3 + K4)*K1;
		Kdl = (K2 - K3 + K4 - 2 * K5)*K1;
		Kmid = K3 * 2 * K1;
		Kdif = K5 * K1;
	}
}// fn

void CData::Clear(int what) const
{
	ENTER
		int i;
	if ( !(Mainbuf.data) ) return;
	if ( what&CLR_COEFF )
	{
		cl_uint V = _PackFloatCA(Vel, 0);
		for ( int I = (MSizeA >> 2) - 1; I >= 0; --I )
			((cl_uint*)Buf_a)[I] = V;
	}

	if ( what&CLR_DATA )  memset(Mainbuf.data + MSizeA * 2, 0, Mainbuf.len - MSizeA * 2);
	if ( what&CLR_W )
		for ( i = 0; i < HostSrcs; ++i )
		{
			ASSERT(MSizeA * (4 + 4 * i) + MSizeA * 2 <= Mainbuf.len);
			memset(Mainbuf.data + MSizeA * (4 + 4 * i), 0, MSizeA * 2);
		}
	//puts("Cleared");
	if ( what&CLR_GRAD )
	{
		ASSERT(Mainbuf.len >= MSizeA * 2);
		memset(Mainbuf.data + MSizeA, 0, MSizeA);
	}
	if ( what&CLR_DETS )  if ( pDetectors ) memset(pDetectors, -1, MaxDetectors * sizeof(DETECTOR));
	if ( what&CLR_BND )  if ( pBoundLines ) memset(pBoundLines, -1, InH * sizeof(BOUNDLINE));

	//puts("ret");
	RETURN();
}// fn

int CData::Init(int w, int h, int d)
{
	int retval = 0;
	ENTER;
	Close();

	if ( w <= 0 ) w = 640;

	if ( fDemoMode )
	{
		w = CLAMP(w, 480, 800); 	w = align(w, 160);
		if ( d < 0 ) d = 15;
		d = CLAMP(d, 10, 20);
	}

	ReqW = w;  ReqH = (h <= 0) ? w : h;
	GlobalSrcs = (d <= 0) ? 15 : d;


	InW = align(ReqW, ImgAlignX); InH = align(ReqH, ImgAlignY);
	FullW = align(InW + 2 * OutFrameX, ImgAlignX) + ImgShearX;
	FullH = InH + 2 * OutFrameY;
	DY = FullW * 4; 	MSizeA = DY * FullH;
	DZ = MSizeA * 4; // u0/u1/w0/w1
	ReqOffsetB = (((InW - ReqW) >> 1) << 2) + ((InH - ReqH) >> 1)*DY;
	InOffsetB = align(OutFrameX * 4, ImgAlignX * 4) + DY * OutFrameY;
	SetParameters();

	size_t memsz = DZ * HostSrcs + MSizeA * 2 + MemAlign + 8 * ImgAlignX;
	if ( !Mainbuf.alloc(memsz, "Mainbuf") )
	{
		fprintf(stderr, "%s CData::Init - allocation failure [%zu Mb]. Try decreasing -CPUMemory. ", _PID, memsz >> 20);
		retval = 1; goto _fail;
	}

	for ( int i = 0; i < 6; ++i ) MemoryShear[i] &= 0x7F8;
	Buf_a = align((cl_ulong)Mainbuf.data, MemAlign);
	pCoeff = Buf_a + MemoryShear[4] + InOffsetB;
	pGrad = Buf_a + MSizeA + MemoryShear[5] + InOffsetB;
	if ( HostSrcs )
	{
		pU[0] = Buf_a + MSizeA * 2 + MemoryShear[0] + InOffsetB;
		pU[1] = Buf_a + MSizeA * 3 + MemoryShear[1] + InOffsetB;
		pW[0] = Buf_a + MSizeA * 4 + MemoryShear[2] + InOffsetB;
		pW[1] = Buf_a + MSizeA * 5 + MemoryShear[3] + InOffsetB;
	} else pU[0] = pU[1] = pW[0] = pW[1] = pGrad;

	{ size_t LBSize = InH * sizeof(BOUNDLINE) + MaxDetectors * sizeof(DETECTOR);
	Linebuf.alloc(LBSize, "Linebuf ");
	if ( !Linebuf.data )
	{ fprintf(stderr, "%s CData::Init - LineBuf [%z] allocation failure ", _PID, LBSize);  retval = 1; goto _fail; }
	}
	pBoundLines = (BOUNDLINE*)align((cl_ulong)Linebuf.data, sizeof(BOUNDLINE));
	pDetectors = (DETECTOR*)(pBoundLines + InH);

	Clear(-1);

	GenerateEmitters();
	GenerateDetectors();
	GenerateBound();

#ifdef TARGET_GPU
	if ( EnableGPU )
	{
		if ( !pGPU->fReady ) InitGPU();
		if ( !pGPU->fReady )
		{ retval = CL_DEVICE_NOT_AVAILABLE;	goto _fail; }
	}
#endif

	if ( InitDataCache() )
	{
		fprintf(stderr, "%sCData init failed! ", _PID);	Close();
		retval = -1; goto _fail;
	}

#ifdef TARGET_GPU
	if ( EnableGPU && !GLayerBuf.Addr )
	{
		if ( AllocateOnGPU() )
		{ retval = CL_INVALID_MEM_OBJECT; goto _fail; }
	}
#endif
	IFLOG(LOG_MULTI) printf("%s Init: Ns=%i of %i from %i ", _PID, NSrcs, GlobalSrcs, StartGlobalSrc);
_fail:  RETURN(retval);
}

int CData::Init(RestartStruct& RST)
{
	*(static_cast<RestartStruct*>(this)) = RST;
	return Init(ReqW, ReqH, GlobalSrcs);
}


int CData::InitGPU(GPUContext* pGpu, cl_uint device)
{
	int retval = 0;
	omp_set_num_threads(NThreads > 0 ? NThreads : (NThreads = omp_get_max_threads()));
	if ( RunArch ) ShaderObj::ActiveArch = RunArch;

#ifdef TARGET_GPU
	if ( !EnableGPU ) return retval;
	EnableGPU = 0;
	pGPU = pGpu;
	if ( !pGPU ) return CL_DEVICE_NOT_AVAILABLE;
	if ( !pGPU->CLContext )
	{

		if ( retval = pGPU->Init(device >> 8, device & 0xFF, KA_List, "gpu.h", "k2d.cl") )
		{ fprintf(stderr, "%sGPU initialization error. Platform:%i device:%i ", _PID, device >> 8, device & 0xFF); return retval; }

		struct { size_t param, limit; const char* msg; } const T[] =
		{ { pGPU->MaxBlockSize , 16, "thread block size"},
			{ pGPU->GlobalMemSize  , 1 << 26, "global memory size"},
			{ pGPU->LocalMemSize  , 512, "local memory size"},
			{pGPU->MaxAllocationSize,1 << 26,"allocation size"}, {0,0,NULL} };

		for ( auto X = T; X->limit; ++X )
		{
			if ( X->param < X->limit )
			{ fprintf(stderr, "%sGPU %s invalid (found: %zu, min: %zu) ", _PID, X->msg, X->param, X->limit); return CL_INVALID_DEVICE; }
		}
		EnableGPU = 1;
	} else EnableGPU = 1;
#else
	EnableGPU = 0;
#endif	
	return retval;
}


int FACTORIZE(int Dim, int Cell)
{
	while ( Dim % Cell ) --Cell;
	int XMul = Dim / Cell;
	Cell = Dim / XMul;
	while ( Dim%Cell ) --Cell;
	return Cell;
}


void CData::XferToGPU(bool Up, cl_int what, cl_uint SrcI) const
{
	if ( !EnableGPU ) return;
#ifdef TARGET_GPU
	ENTER;
	constexpr int NItems = 6;
	constexpr cl_uint HostSections[NItems] = { 2,3,4,5,0,1 },
		GPUSections[NItems] = { 1,2,3,4,0,0 };
	const CLBuffer *pGBuf[NItems] = { &GLayerBuf,&GLayerBuf, &GLayerBuf,&GLayerBuf,&GCoeffBuf,&GLayerBuf };
	const cl_ulong MSizeB = InW * InH * 4;
	const size_t org[4] = { 0,0,0,0 }, rgn[4] = { InW,InH,1,0 };
	bool fImg = false;
	memblk tmp;
	RectWin CoeffWin((char*)pCoeff, InW, InH, 32, DY);

	what &= (1 << NItems) - 1;
	clEnqueueBarrier(pGPU->CLQueue[0]);
	IFLOG(LOG_KRNL) printf("\n XferToGPU %s mask=0x%2X src=%i", Up ? "upload" : "download", what, SrcI);

//	if ( Up ) GBoundBuf.Xfer(true, BoundCache.data, GBoundBuf.Addr, (NumBoundaryA + NumDetectorsA) * 4, pGPU->CLQueue[0]);

	if ( (what & SAVE_COEFF) && !GCoeffBuf.Addr )
	{
		printf("\n XferToGPU upload image.coeff ");
		fImg = true; what &= ~0x10;
		tmp.alloc(MSizeB);
		if ( !tmp.data ) return;
		if ( Up )
		{
			CoeffWin.copymem(tmp.data, InW * 4, true);
			GLayerBuf.Xfer(Up, (void*)pW[0], pgLW.s[0], MSizeB, pGPU->CLQueue[0]);
			clEnqueueCopyBufferToImage(pGPU->CLQueue[0], GLayerBuf.Mem, GCoeffBuf.Mem, pgLW.s[0], org, rgn, 0, NULL, NULL);
		}
	}


	for ( int idx = 0; idx < NItems; ++idx )
	{
		if ( what&(1 << idx) )
			pGBuf[idx]->Xfer(Up, (char*)Buf_a + HostSections[idx] * MSizeA,
											 pGBuf[idx]->Addr + GPUSections[idx] * MSizeA, MSizeA, pGPU->CLQueue[0]);
	}

	if ( fImg && !Up )
	{
		printf("\n XferToGPU download image.coeff ");
		clEnqueueCopyImageToBuffer(pGPU->CLQueue[0], GCoeffBuf.Mem, GLayerBuf.Mem, org, rgn, pgLW.s[0] - GLayerBuf.Addr, 0, NULL, NULL);
		GLayerBuf.Xfer(Up, tmp.data, pgLW.s[0], MSizeB, pGPU->CLQueue[0]);
		CoeffWin.copymem(tmp.data, 0, Up);
	}
	if ( !Up ) clFinish(pGPU->CLQueue[0]);
	else clEnqueueBarrier(pGPU->CLQueue[0]);
	RETURN()
#endif
}



int CData::SolveDirect(bool testmode)
{ // MPI-parallel executes same as sequential, no sync points
	ENTER;
	int cl_error = 0;
	STR64 tmpn;

	int SrcIncr = EnableGPU ? GPUSrcs : HostSrcs;
	GenerateAbsorbField();

	for ( StartSrc = 0; StartSrc < NSrcs; StartSrc += SrcIncr )
	{
		ActiveSrcs = __min(NSrcs - StartSrc, SrcIncr);
		if ( !ActiveSrcs ) break;
		IFLOG(LOG_MULTI) printf("%s Direct: Sources (global) #%02i -- %02i", _PID, StartSrc + StartGlobalSrc, StartSrc + ActiveSrcs - 1 + StartGlobalSrc);
		CL_TRY(RunDirect(true));
		if ( !testmode ) for ( int Src = StartSrc; Src < (StartSrc + ActiveSrcs); ++Src )
		{
			snprintf(tmpn, sizeof(tmpn), "%s.exd.%03i", DirectBaseName, Src + StartGlobalSrc);
			ExData.Dump(tmpn, Src - StartSrc,
#ifdef TARGET_GPU
									EnableGPU ? pGPU->CLQueue :
#endif
									NULL);
		}
	}

	if ( !MyId )
	{
		snprintf(tmpn, sizeof(tmpn), "%s.ini", DirectBaseName);
		FILE* fp = fopen(tmpn, "wb");
		if ( fp ) { DatasetOptions->Print((char*)this, fp);		fclose(fp); }
		IFLOG(LOG_DUMP) { SaveInitial(NULL, SAVE_RST | SAVE_COEFF); }
	}
	PRINT_CL_ERROR;
	RETURN(cl_error);
}


float CData::ComputeResidual(void)
{
	ENTER;
	double sum = 0;
#ifdef TARGET_GPU
	if ( EnableGPU )	ExData.TransferGPU(-1, false, pGPU->CLQueue[0]);
	IFLOG(LOG_DUMP) ExData.Dump(EnableGPU ? "exdG" : "exdC", -1);
#endif
	for ( int S = 0; S < ExData.LinesCommon - 1; ++S )
	{
		float* const p = (float*)ExData.HostAddr(-1, S);
#pragma simd
		for ( int D = NumDetectors - 1; D >= 0; --D ) { sum += p[D];	p[D] = 0; }
	}
#ifdef TARGET_GPU
	if ( EnableGPU )	ExData.TransferGPU(-1, true, pGPU->CLQueue[0]);
#endif
	RETURN(sum);
}


void CData::BackstepCoeff(float fc, const cl_ushort2* pPrevCoeff)
{

	cl_ushort2 *pC = (cl_ushort2*)pCoeff; const cl_ushort2 *pG = (cl_ushort2*)pPrevCoeff;

	if ( fc == 0 ) { memcpy(pC, pG, ReqH*FullW * 4); return; }

	for ( int Y = 0; Y < ReqH; ++Y, pC += FullW, pG += FullW )
	{
#pragma simd
		for ( int X = 0; X < ReqW; ++X )
		{
			const cl_ushort2 ca = pC[X], caOld = pG[X];
			const float cc = ((float)ca.s[0] * (float)(ca.s[0])), co = ((float)caOld.s[0] * (float)(caOld.s[0]));
			float cnew = round(sqrt(cc * fc + co * (1 - fc)));
			float a_new = round((float)ca.s[1] * fc + (float)caOld.s[1] * (1.0f - fc));
			pC[X].s[0] = cnew;// CLAMP(cnew, CImin, CImax);
			pC[X].s[1] = __max(a_new, 0.0f);
		}
	}
}


int CData::Process(void)
{
	ENTER;
	int retval = 0;


#ifdef TARGET_GPU
	if ( EnableGPU )
	{
		if ( !pGPU->fReady )
		{ if ( InitGPU() ) return CL_DEVICE_NOT_AVAILABLE; }
		if ( !pGPU->fReady ) return CL_DEVICE_NOT_AVAILABLE;
	}
#endif
	if ( NumProcesses == 1 )
	{ // tests only supported in single-process mode because the number of sources can change
		if ( PerfTest == 1 )
		{ DetermineCPUCache(PerfTestParam); retval = 0; goto ret; }

		if ( TestIncrSize[0] )
		{
			TestRecord TR;

			int WorkSize = ReqW, StepNum;
			for ( StepNum = 0;
					 WorkSize <= TestIncrSize[1];
					 WorkSize += TestIncrSize[0], ++StepNum )
			{
				if ( TestIncrSrc > 0 )
				{
					int SrcBak = GlobalSrcs;
					do PerfTestRun(WorkSize, PerfTestParam, TR);
					while ( (GlobalSrcs -= TestIncrSrc) > 0 );
					GlobalSrcs = SrcBak;
				} else PerfTestRun(WorkSize, PerfTestParam, TR);
			}
			retval = 0; goto ret;
		}
	}

	if ( *DirectBaseName )
	{// solve Direct
		int p = 0;
		if ( retval = Init(ReqW, ReqH, GlobalSrcs) ) goto skip;
		p = LoadPhantom(NULL, PHANTOM_ALLOW_AUTO | PHANTOM_ALLOW_REINIT); // allow reinit from .RST
		if ( p == PHANTOM_NONE ) { retval = 1; goto skip1; }
		if ( !MyId && (p == PHANTOM_AUTO) )
		{
			STR64 a; sprintf(a, "auto_%s", DirectBaseName);
			SaveInitial(a, SAVE_COEFF | ImageExportFormats);
			printf("\nSaving auto-phantom '%s' ", a);
		}
		SaveSrcRcv(); // uses MPI

		if ( !MyId ) printf("\nGenerating dataset '%s' ", DirectBaseName);
		retval = SolveDirect();
	skip1:	Close();
	skip:;
	}
	if ( CollectRetval(retval) ) goto ret;

	if ( *InverseBaseName )
	{// solve inverse
		STR64 tmpn; int res;
		snprintf(tmpn, sizeof(tmpn), "%s.ini", InverseBaseName);
		printf("\nLoading dataset '%s' ", InverseBaseName);
		DatasetOptions->ParseFile(tmpn, (char*)this);
		if ( !MyId )
		{ IFLOG(LOG_INFO) DatasetOptions->Print((char*)this, stdout); }

		strncpy(DirectBaseName, InverseBaseName, sizeof(DirectBaseName));
		// used by auto-save functions for file names

		retval = Init(ReqW, ReqH, GlobalSrcs);
		if ( CollectRetval(retval) ) return 1;

		CurrentIteration = 0;
		Clear(CLR_COEFF);

		res = 0;
		// Don't allow reinit or auto-number here as the problem is already specified
		// '#NNN' may be a file name			
		if ( *RestartName )  res = LoadPhantom(RestartName, 0);
		//else res = LoadRST(InverseBaseName); // auto-load restartfile = datasetname on startup. Disabled by default.
		if ( (res == PHANTOM_RST) && !MyId )
			printf("Restarting iteration #%3i from %s ", CurrentIteration, RestartName);
		retval = SolveInverse(0);
		Close();
	}

ret: RETURN(retval);
}



int CData::DetermineCPUCache(int sz)
{
	ENTER;
	fDemoMode = 0; EnableGPU = 0;	nDump = 0;

	int retval = 0, Found = 0;
	float DMin = 1e30;
	size_t ExDataSize;

	Close(); SetParameters(true);

	CPUMemorySizeMB = 200;
	if ( sz <= 0 ) sz = 360;
	sz = CLAMP(sz, 320, 1600); sz -= (sz % 40);

	HostSrcs = (CPUMemorySizeMB << 20) / (sz*sz * 16);
	HostSrcs = CLAMP(HostSrcs, 1, MAX_SRCS - 1);

	if ( TestVolume < 0 ) TestVolume = 5e8;
	NFrames = TestVolume / (sz*sz*HostSrcs);
	NFrames = CLAMP(NFrames, 4, 2000);

	if ( Init(sz, sz, HostSrcs) ) { retval = 1; goto _ret; }

	EndInvTime = 2;
	ExDataSize = NumDetectorsA * NFrames * 4;
	if ( !ExDataFileMaps.init(0, NSrcs, ExDataSize, NULL) )
	{ retval = 1; goto _ret; };

	StartSrc = 0; ActiveSrcs = 1; EndInvTime = 2;
	RepeatTest = TestDuration > 0 ? -TestDuration * 1000 : -1000; // ms to run

	printf("\n\n CPU cache test. Threads=%i. Max srcs=%i\nSize(MB) DirectMPPS  InvMPPS  TotalMPPS  Srcs", NThreads, HostSrcs);

	if ( retval = RunDirect(false) )		goto _ret; // cache preload

	while ( ActiveSrcs <= HostSrcs )
	{
		float MB = MSizeA / 1048576.0f * (ActiveSrcs * 4 + 2);

		if ( retval = RunDirect(false) )		goto _ret;
		printf("\n%6.1f   %8.2f  ", MB, Stat_TaskMPPS);
		float DM = 1.0f / Stat_TaskMPPS;

		if ( retval = RunInverse() ) goto _ret;

		DM += 1.0f / Stat_TaskMPPS;
		printf(" %8.2f  %8.2f     %i ", Stat_TaskMPPS, 1.0f / DM, ActiveSrcs);
		if ( DM < DMin ) { DMin = DM; Found = MB; }
		ActiveSrcs += (ActiveSrcs < 10) ? 1 : (ActiveSrcs < 24) ? 2 : (ActiveSrcs < 64) ? 4 : 8;
	}
	printf("\nOptimal: -CPUMemory %i (MB)\n", Found);
	{// local block
		FILE* fd = fopen(CData::INIFile, "a+");
		if ( fd )
		{ fprintf(fd, "\n-CPUMemory %i //added by CPU cache test\n", Found); fclose(fd); }
	}
_ret:
	Close();
	if ( retval )	 printf("%s DetermineCPUCache aborted ", _PID);
	RETURN(retval);
}

void CData::LoadOptimizedSettings(void)
{
	memblk opt(AutotuneFile);
	if ( !opt.data ) goto Auto;

	{// local block
		TestRecord* pTR = (TestRecord*)opt.data;
		int NumRecs = opt.len / sizeof(TestRecord);
		int dS = 0x7FFFFFFF, Found = -1;

		for ( int i = 0; i < NumRecs; ++i )
		{
			if ( !EnableGPU == !pTR[i].NThreads ) continue; // use NT=0 records if EnableGPU=1
			int d = ((int)ReqW - (int)pTR[i].size); if ( d < 0 ) d = -d;
			if ( d < dS ) { dS = d; Found = i; }
		}

		if ( (Found >= 0) && (fabs((float)dS) / ReqW < 0.3) )
		{
			if ( !MyId ) { pTR[Found].Print(); puts(" Found "); }
			BLKs = pTR[Found];	return;
		}
	}

Auto:
	if ( fAutoTune )
	{ TestRecord TR;  ComputeOptimizedSettings(TR); 	BLKs = TR; }
}

int CData::PerfTestRun(int sz, int optimize, TestRecord& TR)
{ // uses TestVolume
	ENTER;
	// mode 0 - execute one test, fill TestRecord
	// mode 1 - find optimal block sized, fill TestRecord on optimal settings
	// mode 2 - run test, apply optimized settings from step 1 
	// mode & 4 - run at iteration settings

	fDemoMode = 0;
	int retval = 0;

	int SrcIncr;
	cl_float2 F_B_mpps;

	Close();

	if ( !(optimize & 4) ) //opt &4 = run at iteration settings, else run at defaults
	{
		SetParameters(true);
		RepeatTest = (TestDuration > 0) ? -TestDuration * 1000 : 1;
		NFrames = TestVolume / (sz*sz*NSrcs);
		if ( NFrames > sz * 3 )
		{
			NFrames = sz * 3;
			if ( RepeatTest > 0 )
			{
				RepeatTest = TestVolume / (sz*sz*NSrcs*NFrames);
				RepeatTest = __max(RepeatTest, 1);
			}
		}
		NFrames = __max(NFrames, 8);
	}

	TR.NThreads = EnableGPU ? 0 : NThreads;
	TR.size = sz;	ValidateBlockSizes(TR);
	TR.NSrcs = GlobalSrcs;

	if ( retval = Init(sz, sz, GlobalSrcs) )  goto _ret;

	if ( !ExDataFileMaps.init(0, NSrcs, NumDetectorsA * NFrames * 4, NULL) )
	{ retval = 1; goto _ret; };

	switch ( optimize & 3 )
	{
		case 2: if ( !MyId ) printf("\n Load optimized settings");
			LoadOptimizedSettings(); // run at optimized settings - fallthrough

		case 0: // run at current settings
			F_B_mpps = SimulatedRun();
			TR.FwdMPPS = F_B_mpps.s[0];
			TR.BwdMPPS = F_B_mpps.s[1];
			TR.flags = 0;
			*(static_cast<BlockSizes*>(&TR)) = BLKs;
			if ( !MyId ) FWrite(PerftestFile, &TR, sizeof(TR), true);
			TR.Print();
			break;

		case 1:  	ComputeOptimizedSettings(TR);

	}	 // switch

_ret: Close();
	if ( retval ) fprintf(stderr, "Test aborted");
	RETURN(retval);
}//fn


void CData::ComputeOptimizedSettings(TestRecord &TR)
{
	static const int SpanTests[] = { 5,10,20,40,0 };
	bool fLocalExdata = false;

	if ( !Mainbuf.data ) return;

	printf("\n Computing optimized settings for size=%i (%i*%i) ", ReqW, FullW, FullH);
	TR.NThreads = EnableGPU ? 0 : NThreads;
	TR.NSrcs = NSrcs;
	TR.FwdMPPS = TR.BwdMPPS = 0;

	TestRecord T_Curr;
	T_Curr.NThreads = EnableGPU ? 0 : NThreads;
	T_Curr.NSrcs = NSrcs;

	const size_t ExDataSize = NumDetectorsA * NFrames * 4;
	if ( ExDataFileMaps.NumOpen(0, NSrcs, ExDataSize) < NSrcs )
	{ fprintf(stderr, "\n %s: ExData not initialized", __FUNCTION__);  return; }

	for ( int CW = 32; CW <= (EnableGPU ? 128 : 32); CW <<= 1 )
		for ( int i = 0; SpanTests[i]; ++i )
		{
			const int CurSpan = SpanTests[i];
			T_Curr.size = TR.size = ReqW;
			T_Curr.flags = 1;
			T_Curr.FwdY = T_Curr.BwdY = BLKs.FwdY = BLKs.BwdY = CurSpan;
			T_Curr.FwdW = T_Curr.BwdW = CW;

			const cl_float2 F_B_mpps = SimulatedRun();

			T_Curr.FwdMPPS = F_B_mpps.s[0];
			T_Curr.BwdMPPS = F_B_mpps.s[1];

			if ( TR.FwdMPPS < F_B_mpps.s[0] )
			{ TR.FwdMPPS = F_B_mpps.s[0]; TR.FwdY = CurSpan; TR.FwdW = CW; }
			if ( TR.BwdMPPS < F_B_mpps.s[1] )
			{ TR.BwdMPPS = F_B_mpps.s[1]; TR.BwdY = CurSpan; TR.BwdW = CW; }
			if ( !MyId ) FWrite(PerftestFile, &T_Curr, sizeof(T_Curr), true);
			IFLOG(LOG_INFO) { T_Curr.Print(); }
		} // loop param
	if ( !MyId ) FWrite(AutotuneFile, &TR, sizeof(TR), true);	 TR.Print(); puts(" =optimal");
}


cl_float2 CData::SimulatedRun(void)// return megapixels per second fwd, bwd
{ // CData, GPU and ExDataFileMaps must be initialized!	
	const int SrcIncr = EnableGPU ? GPUSrcs : HostSrcs;
	cl_float2 ret = { 0.0f,0.0f };
	cl_ulong FwdVol = 0, BwdVol = 0;

	IFLOG(LOG_INFO) { printf("\nActiveSrcs: %i ", SrcIncr); }
	EndInvTime = 2;

	ActiveSrcs = __min(NSrcs, SrcIncr);
	if ( !ActiveSrcs ) return ret;

	RunDirect(false); // cache preload

	for ( int StartSrc = 0; StartSrc < NSrcs; StartSrc += SrcIncr )
	{
		ActiveSrcs = __min(NSrcs - StartSrc, SrcIncr);
		if ( !ActiveSrcs ) break;

		RunDirect(false);
		FwdVol += (cl_ulong)Stat_FrameCount * ActiveSrcs*InW*InH;
		ret.s[0] += Stat_TaskTime;
		//printf("\n %.2fs %ifrm %.2f fps AS=%i W=%i H=%i ", Stat_TaskTime, Stat_FrameCount,(float)Stat_FrameCount / Stat_TaskTime, ActiveSrcs,InW,InH);
		RunInverse();
		BwdVol += (cl_ulong)Stat_FrameCount * ActiveSrcs*InW*InH;
		//printf("\n %.2fs %ifrm  %.2f fps ", Stat_TaskTime, Stat_FrameCount,(float)Stat_FrameCount / Stat_TaskTime);
		ret.s[1] += Stat_TaskTime;
	}// Srcs

	ret.s[0] = (float)FwdVol*1e-6f / ret.s[0];
	ret.s[1] = (float)BwdVol*1e-6f / ret.s[1];
	return ret;
}


int CData::SolveInverse(bool testmode)
{ // .pCoeff contains initial apprx, RST parameters assumed valid
	ENTER;
	int cl_error = 0, MinHit = 0;	
	cl_ulong IterTime, StartTime = Microseconds();
	STR64 tmpn;	char tmpstr[256];


	float fc = 0, tmpf = 0; bool step_up = false;
	float		DefTargetStep = 0.12f, MaxTargetStep = 0.3f, MinRes = 1e30;
	int StepDownFired = 0, DMin = 0;
	FILE* fdLog = NULL;

	size_t ExDataSize = NumDetectorsA * NFrames * 4;
	int SrcIncr = EnableGPU ? GPUSrcs : HostSrcs;


	if ( CGScale <= 0 ) CGScale = (GradScale.s[0] = 30.0f / NSrcs / dt / dt)*CPU_GRAD_SCALE;
	else GradScale.s[0] = CGScale / CPU_GRAD_SCALE;
	if ( AGScale <= 0 ) AGScale = (GradScale.s[1] = 3.0f / NSrcs / dt)*CPU_GRAD_SCALE;
	else GradScale.s[1] = AGScale / CPU_GRAD_SCALE;
	ParGradScale.s[0] = GradScale.s[0] * GPUSrcs;
	ParGradScale.s[1] = GradScale.s[1] * GPUSrcs;

	OldResidual = 0;
	ComputeResidual(); // clear

	memblk PrevCoeff(ReqH*DY), FullGrad(ReqH*FullW * sizeof(cl_float2));

	if ( !PrevCoeff.data || !FullGrad.data ) { cl_error = 1;  goto pre_iter_errors; }
	memcpy(PrevCoeff.data, (char*)pCoeff, ReqH*DY);

	snprintf(tmpn, sizeof(tmpn), "%s.exd", InverseBaseName);
	if ( testmode ) 	ExDataFileMaps.init(0, NSrcs, ExDataSize, NULL);
	else
		if ( !ExDataFileMaps.assert(0, NSrcs, ExDataSize, ExDataSize, tmpn, StartGlobalSrc) )
		{ fprintf(stderr, "%s Data files absent for srcs %i -- %i ", _PID, StartGlobalSrc, StartGlobalSrc + NSrcs - 1); cl_error = 1;  goto pre_iter_errors; }
	if ( fAutoTune ) LoadOptimizedSettings();

	if ( !MyId )
	{
		STR64 logname;	sprintf(logname, "%s.log", InverseBaseName);
		fdLog = fopen(logname, "a+"); if ( !fdLog )fdLog = stdout;
		fprintf(fdLog, "\n\nTask: '%s' size: %i  nsources: %i\n", InverseBaseName, ReqW, NSrcs);
		IFLOG(LOG_INFO) Options->Print((const char*)this, fdLog);
	}

pre_iter_errors:

	if ( CollectRetval(cl_error) ) return (cl_error ? cl_error : 1);

// CurrentIteration is read from .RST or set to 0 or 1 prior to calling
	while ( !cl_error && (++CurrentIteration <= MaxIterations) )
	{
		float Residual = 0;
		Clear(CLR_GRAD);

#ifdef TARGET_GPU
		XferToGPU(true, SAVE_GRAD);
		MRange CLSRng("cls",
									64, GParGrad.Len >> 10, 64, GParGrad.Len >> 10,
									64, 1, 0, 1);

		if ( GParGrad.Addr )
		{
			CL_TRY(clSetKernelArg((*pGPU)[k_CLS], 0, 8, &(GParGrad.Addr)));
			CL_TRY((*pGPU)[k_CLS].ExecWait(pGPU->CLQueue[0], CLSRng));
		}
#endif

		GenerateAbsorbField();
		int LoadedSrc = -1;
		for ( StartSrc = 0; StartSrc < NSrcs; StartSrc += SrcIncr )
		{

			ActiveSrcs = __min(NSrcs - StartSrc, SrcIncr);
			if ( !ActiveSrcs ) break;
			IFLOG(LOG_MULTI) printf("%sSources #%02i -- %02i ", _PID, StartSrc+StartGlobalSrc, StartSrc + StartGlobalSrc+ActiveSrcs - 1);
			CL_TRY(RunDirect(false));
			IFLOG(LOG_STATS | LOG_PROFILE) printf(" Direct: %.2f MPps ", Stat_TaskMPPS);

#ifdef TARGET_GPU
			if ( EnableGPU && (StartSrc != LoadedSrc) )
			{ // load active sources from Exdata memory into GPU only if the active subset had changed
				for ( int sd = 0; sd < ActiveSrcs; ++sd )
					ExData.TransferGPU(sd, true, pGPU->CLQueue[0], ExDataFileMaps[sd + StartSrc].access);
				LoadedSrc = StartSrc;
			}
#endif					
			CL_TRY(RunInverse());
			IFLOG(LOG_STATS | LOG_PROFILE) printf(" Inverse: %.2f MPps ", Stat_TaskMPPS);

#ifdef TARGET_GPU
			if ( GParGrad.Addr )
			{
				int GW = InW / 4, LW = FACTORIZE(GW, 128);
				MRange SumGradRng("SumGrad", GW, InH, GW, InH, LW, 1, 0, 1);
				SETARG0(k_SumParGrad);
				CL_TRY(clSetKernelArg((*pGPU)[k_SumParGrad], 1, 4, &(ActiveSrcs)));
				CL_TRY((*pGPU)[k_SumParGrad].ExecWait(pGPU->CLQueue[0], SumGradRng));
			}

			if ( EnableGPU ) { IFLOG(LOG_DUMP) ExData.CLBuffer::Dump("ExDataB.dmp", pGPU->CLQueue[0]); }
#endif			
			Residual += ComputeResidual(); //printf(" %.2f ", Residual);
			if ( (Residual == 0) || (Residual != Residual) )
			{ fprintf(stderr, "\nIncorrect data "); cl_error = 1; goto cl_error_occured; }
			IFLOG(LOG_DUMP)
			{
				sprintf(tmpn, "Gs%02i_i%02i", StartSrc + StartGlobalSrc, CurrentIteration);
				SaveInitial(tmpn, SAVE_GRAD | EXPORT_MAT);
			}
		} // sources

		if ( EnableGPU ) XferToGPU(false, SAVE_GRAD);
		float4 estgrad = EstimateGrad();
		float GradAmpl = estgrad[0] / (GradScale.s[0] * dt*dt*dt*dt);

		IFLOG(LOG_DUMP)
		{ // for all processes
			if ( EnableGPU ) XferToGPU(true, SAVE_GRAD);
			sprintf(tmpn, "GR%02i_i%02i", StartSrc + StartGlobalSrc, CurrentIteration);
			SaveInitial(tmpn, SAVE_GRAD | EXPORT_MAT);
		}
#ifdef USE_MPI
		if ( NumProcesses > 1 )
		{// Unify the gradients via MPI 
			cl_float2* const dst = ((cl_float2*)FullGrad.data);
			cl_short2* const src = ((cl_short2*)(pGrad + ReqOffsetB));
#pragma simd
			for ( int X = ReqH * FullW; X >= 0; --X )
			{ // unpack short to float
				dst[X].s[0] = src[X].s[0] / GradScale.s[0];
				dst[X].s[1] = src[X].s[1] / GradScale.s[1];
			}
			dst->s[0] = GradAmpl; dst->s[1] = Residual;
			MPI_Allreduce(MPI_IN_PLACE, FullGrad.data, InW*InH*2, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
			GradAmpl = dst->s[0] / NumProcesses; Residual = dst->s[1];
			dst->s[0] = dst->s[1] = 0;

			if ( !MyId && (G_Settings.LogLevel&LOG_DUMP) )
			{
				const int D[4] = { 2, FullW,ReqH, 0 };
				sprintf(tmpn, "GRAD_i%02i", CurrentIteration);
				MATHeader(D, MATHeader::miSINGLE, tmpn).WriteFile(FullGrad.data);
			}
		}
#endif		
		IterTime = Microseconds() - StartTime;

		if ( StepSize <= 0 )
		{
			const float dtc0 = 1.0f / (dt*dt*Vel*Vel),
				PlanDiff = fabs(1.0f / (dt*dt*(Vel + TargetGradAmpl)*(Vel + TargetGradAmpl)) - dtc0);
			if ( StepSize == 0 ) StepSize = -1;
			StepSize *= -PlanDiff / GradAmpl;
		}

		if ( G_Settings.LogLevel&LOG_INFO )
		{
			snprintf(tmpstr, sizeof(tmpstr), "%s #%02i Est.grad C: %.2f fill, %.2f over; A: %.2f fill, %.2f over. ", _PID, CurrentIteration, estgrad[0] / 32768.0f, estgrad[2], estgrad[1] / 32768.0f, estgrad[3]);
			puts(tmpstr);		if ( NumProcesses == 1 ) fputs(tmpstr, fdLog);
		}

		if ( TargetStepCoeff <= 0 ) TargetStepCoeff = DefTargetStep;
		float DN = OldResidual <= 0 ? 0 : (1 - Residual / OldResidual) * 100;
		if ( Residual < MinRes ) { MinRes = Residual; DMin = 0; } else ++DMin;
		if ( (OldResidual == 0) || (Residual < OldResidual) )
		{
			fc = StepSize;
			if ( OldResidual == 0 ) tmpf = 1;
			else
				if ( TargetStepCoeff <= 0 ) tmpf = StepUpCoeff;
				else
				{
					tmpf = 1.0f - Residual / OldResidual;
					if ( StepDownFired ) tmpf = 1.0f;
					else
					{ // Try to keep average descent rate
						TargetStepCoeff = 0.75f*TargetStepCoeff + 0.25f*tmpf;
						tmpf = (1.0f + TargetStepCoeff) / OldResidual * Residual;
						tmpf = CLAMP(tmpf, 0.6f, 1.25f)*StepUpCoeff;
					}
				}

			StepSize *= tmpf;
			if ( !MyId )
			{
				if ( fDemoMode )			snprintf(tmpstr, sizeof(tmpstr), "\n%02i:  %.2fs", CurrentIteration, 1e-6f*(float)IterTime);
				else		snprintf(tmpstr, sizeof(tmpstr), "\n%02i: StepUp N=%.3e Old=%.3e D=%.3f%% L=%.2e DL=%.2f Ts=%.2f%%  %.2fs", CurrentIteration, Residual, OldResidual, DN, StepSize, tmpf, TargetStepCoeff * 100, 1e-6f*(float)IterTime);
				fputs(tmpstr, fdLog); 	puts(tmpstr);
			}
			step_up = 1;
			memcpy(PrevCoeff.data, (char*)pCoeff, ReqH*DY);
			UpdateCoeff(fc, NumProcesses > 1 ? (cl_float2*)FullGrad.data : NULL);
			StepDownFired = 0;
		} else
		{
			step_up = 0;
			StepSize *= (fc = StepDownCoeff);
			TargetStepCoeff *= StepDownCoeff;

			BackstepCoeff((StepDownFired > 1) ? -0.05 : 0.2, (cl_ushort2*)(PrevCoeff.data));
			if ( StepDownFired > 1 ) OldResidual = 0; // possible local minimum - try another direction

			if ( !MyId )
			{
				if ( fDemoMode )		snprintf(tmpstr, sizeof(tmpstr), "\n%02i:  %.2fs", CurrentIteration, 1e-6f*(float)IterTime);
				else	snprintf(tmpstr, sizeof(tmpstr), "\n%02i: StepDn N=%.3e Old=%.3e D=%.3f%% L=%.2e DL=%.2f  Ts=%.2f%%  %.2fs", CurrentIteration, Residual, OldResidual, DN, StepSize, tmpf, TargetStepCoeff * 100, 1e-6f*(float)IterTime);
				fputs(tmpstr, fdLog); 	puts(tmpstr);
			}
			StepDownFired++;
		}


		if ( !MyId )
		{
			sprintf(tmpn, "%s_last", InverseBaseName);
			SaveInitial(tmpn, SAVE_COEFF | SAVE_RST);

			if ( CurrentIteration % SaveIterations == 0 )
			{
				sprintf(tmpn, "%s_i%03i", InverseBaseName, CurrentIteration);
				SaveInitial(tmpn, SAVE_COEFF | ImageExportFormats);
				printf(" Saved %s ", tmpn);
			}
		}
		if ( (OldResidual > 0) && ((OldResidual - Residual) > 0) )
		{ if ( fabs(OldResidual - Residual) / (OldResidual + Residual) < MinThreshold ) ++MinHit; else MinHit = 0; }
		if ( step_up ) 	OldResidual = Residual;
		if ( (MinHit > 8) || (DMin > 15) )
		{
			if ( !MyId )
			{
				fprintf(fdLog, "\nFinished DnHit=%i DMin=%i SD=%i \n", MinHit, DMin, StepDownFired);
				printf("\nFinished");
			}
			cl_error = E_FINISHED;
		}

		GradScale.s[0] *= (estgrad[2] > 0.5) ? 0.05 : (estgrad[2] > 0.05) ? 0.12 : 12000.0f / estgrad[0];
		GradScale.s[1] *= (estgrad[3] > 0.5) ? 0.05 : (estgrad[3] > 0.05) ? 0.12 : 12000.0f / estgrad[1];
		CGScale = GradScale.s[0] * CPU_GRAD_SCALE;
		AGScale = GradScale.s[1] * CPU_GRAD_SCALE;
		ParGradScale.s[0] = GradScale.s[0] * GPUSrcs;
		ParGradScale.s[1] = GradScale.s[1] * GPUSrcs;
		if ( !MyId )
		{
			if ( CheckFile("stop") >= 0 )
			{ printf("\nExit: stop-file"); unlink("stop");  cl_error = 1; }
			fflush(fdLog);
		}

		cl_error = CollectRetval(cl_error); // exit iteration loop on error/ stop condition
	}// iterations

	if ( !MyId )
	{
		sprintf(tmpn, "finished_%s", InverseBaseName);
		SaveInitial(tmpn, SAVE_COEFF | ImageExportFormats);
	}
	PRINT_CL_ERROR
		if ( !MyId && (fdLog != stdout) ) fclose(fdLog);
	if ( cl_error == E_FINISHED ) cl_error = 0;
	RETURN(cl_error);
}


int CData::RunLoop(std::function<int(int)> *callback, bool Direct)
{ // Runs the simulation callback forward or backward in time. 
	// Provides dump, time count functionality, etc.
	const cl_ulong StartTime = Microseconds();
	const int End = Direct ? (T_START + NFrames) : EndInvTime;
	const int Rep = (RepeatTest < 0) ? 1 : RepeatTest;
	cl_int cl_error = 0;
	Stat_FrameCount = 0;
rep: // RepeatTest < 0 specifies milliseconds to run, else repeat count
	for ( int i = 0; i < Rep; ++i )
	{
		T = Direct ? T_START : (NFrames + T_START - 1); // ;
		while ( (Direct ? (T < End) : (T > End)) && !cl_error )
		{
			int Span = (Mode&MODE_CHOP) ? 1 : NFrames;		if ( !Direct ) Span = -Span;
			if ( Direct ? (T + Span >= End) : (T + Span < End) )	Span = End - T;
			if ( nDump )
			{
				int nextdump = Direct ? T + (nDump - (T%nDump)) : T - 1 - ((T - 1) % nDump);
				if (Span<0? ( T + Span < nextdump ): (T + Span > nextdump)) Span = nextdump - T;
			}
			IFLOG(LOG_TSTEP) printf("\n T=%i span=%i", T, Span); if ( !Span ) break;
			{ cl_error |= (*callback)(Span); } T += Span; Stat_FrameCount += (Span < 0 ? -Span : Span);
			if ( nDump && (T%nDump == 0) ) { SaveInitial(NULL, SAVE_UW | WaveExportFormats | SAVE_TSTEP | SAVE_ALT_NAME); }
		}
	}
	cl_ulong MppsTime = Microseconds() - StartTime;
	if ( (RepeatTest < 0) && (MppsTime / 1000 < -RepeatTest) ) goto rep;
	Stat_TaskTime = 1e-6f*(float)MppsTime;
	Stat_TaskMPPS = 1e-6f / Stat_TaskTime * ActiveSrcs*Stat_FrameCount*InW*InH;
	return cl_error;
};


int CData::RunDirect(bool fDirect)
{
	ENTER;

	int Span, cl_error = 0; cl_ulong MppsTime;
	T = T_START;

	if ( !ActiveSrcs ) { fprintf(stderr, "%sRunDirect: Empty set (ActiveSrcs=0)", _PID); RETURN(-1); }

#ifdef TARGET_GPU
	if ( !EnableGPU ) goto CPU_Mode;
	{
		struct BlockSizes BSZ;
		ValidateBlockSizes(BSZ);
		const int SectionWidth = fDirect ? NumDetectorsA : NumBoundaryA;

		int GH = InH / BSZ.FwdY, LH,// (FullH*GPUSrcs + BLKs.FwdY - 1) / BLKs.FwdY,
			GW = InW / 4, LW = FACTORIZE(GW, BSZ.FwdW);
		MRange3  F44Rng("F44Rng", GW, GH, ActiveSrcs, GW, GH, ActiveSrcs, LW, 1, 1, 0, 1);
		MRange  BoundR1("BoundR1", SectionWidth, ActiveSrcs,
										SectionWidth, ActiveSrcs, DTCTR_ALIGN, 1, 0, 1);
		LW = FACTORIZE(InW, 64);		LH = FACTORIZE(InH, 16);
		MRange3  NRBC_Hrz("NRBC_H", InW, 2, ActiveSrcs, InW, 2, ActiveSrcs, LW, 1, 1, 0, 1);
		MRange3  NRBC_Vrt("NRBC_V", InH, 2, ActiveSrcs, InH, 2, ActiveSrcs, LH, 1, 1, 0, 1);


		IFLOG(LOG_GPU) GPU_DATA::Print(-1);
		if ( cl_error = UploadToGPU() ) { EnableGPU = 0; goto cl_error_occured; }

		if ( ImpWidth == 0.0f )// skip initial pulse generation - loaded from data
			XferToGPU(true, SAVE_COEFF | SAVE_UU);

		else
		{
			MRange CLSRng("cls",
										64, (GLayerBuf.Len - MSizeA) >> 10, 64, (GLayerBuf.Len - MSizeA) >> 10,
										64, 1, 0, 1);
			cl_ulong tmpl = GLayerBuf.Addr + MSizeA;
			CL_TRY(clSetKernelArg((*pGPU)[k_CLS], 0, 8, &tmpl));
			XferToGPU(true, SAVE_COEFF);
			CL_TRY((*pGPU)[k_CLS].ExecWait(pGPU->CLQueue[0], CLSRng));
			GenerateInitialPulse();
			IFLOG(LOG_DUMP) SaveInitial("initpulseG", SAVE_UU | SAVE_TSTEP | EXPORT_MAT);
		}


		SETARG0(k_ForwardWave44);
		SETARG0(k_SaveExData);
		SETARG0(k_NRBCond_Hrz);
		SETARG0(k_NRBCond_Vrt);
		CL_TRY(clSetKernelArg((*pGPU)[k_ForwardWave44], 2, sizeof(cl_int), &BSZ.FwdY));

		cl_ulong tmpl = fDirect ? ExData.GPUAddr(0, 0) : BoundCh.GPUAddr(0, 0);
		if ( !tmpl )
		{ cl_error = CL_INVALID_MEM_OBJECT;  goto cl_error_occured; }
		CL_TRY(clSetKernelArg((*pGPU)[k_SaveExData], 1, sizeof(cl_ulong), &tmpl));


		std::function<int(int)> DirGPU = [this, &F44Rng, &BoundR1, &NRBC_Hrz, &NRBC_Vrt] (int _Span) ->int
		{
			const bool fProfiling = (G_Settings.LogLevel & LOG_GPUPROF);
			cl_int cl_error = 0;
			clFinish(pGPU->CLQueue[0]);
			for ( int I = 0; (I < _Span) && !cl_error; ++I )
			{
				cl_int TT = I + T;
				IFLOG(LOG_TSTEP) { printf(" DirT=%i  ", TT); }

				if ( !fSkipScalars )
				{
					cl_ulong  tmpl = pgLU.s[TT & 1];
					cl_int winc = pgLU.s[1 - (TT & 1)] - tmpl;

					CL_TRY(clSetKernelArg((*pGPU)[k_NRBCond_Vrt], 1, sizeof(cl_ulong), &tmpl));
					CL_TRY(clSetKernelArg((*pGPU)[k_NRBCond_Vrt], 2, sizeof(cl_int), &winc));
					CL_TRY((*pGPU)[k_NRBCond_Vrt].ExecNoWait(pGPU->CLQueue[0], NRBC_Vrt, fProfiling));

					CL_TRY(clSetKernelArg((*pGPU)[k_NRBCond_Hrz], 1, sizeof(cl_ulong), &tmpl));
					CL_TRY(clSetKernelArg((*pGPU)[k_NRBCond_Hrz], 2, sizeof(cl_int), &winc));
					CL_TRY((*pGPU)[k_NRBCond_Hrz].ExecNoWait(pGPU->CLQueue[0], NRBC_Hrz, fProfiling));

					CL_TRY(clSetKernelArg((*pGPU)[k_SaveExData], 2, sizeof(cl_int), &TT));
					CL_TRY((*pGPU)[k_SaveExData].ExecNoWait(pGPU->CLQueue[0], BoundR1, fProfiling));
				}

				CL_TRY(clSetKernelArg((*pGPU)[k_ForwardWave44], 1, sizeof(cl_int), &TT));
				CL_TRY((*pGPU)[k_ForwardWave44].ExecNoWait(pGPU->CLQueue[0], F44Rng, fProfiling));

				CL_TRY(clEnqueueBarrier(pGPU->CLQueue[0]));
				//CL_TRY(clEnqueueWaitForEvents(pGPU->CLQueue[0], 1, &((*pGPU)[k_NRBCond_Vrt].evt)));															 

				if ( fProfiling )
				{
					(*pGPU)[k_ForwardWave44].Profile(true);
					(*pGPU)[k_NRBCond_Hrz].Profile(true);
					(*pGPU)[k_NRBCond_Vrt].Profile(true);
					(*pGPU)[k_SaveExData].Profile(true);
				}
			}

			PRINT_CL_ERROR
				clFinish(pGPU->CLQueue[0]);
			return cl_error;
		};

		RunLoop(&DirGPU, true);

		IFLOG(LOG_DUMP)
		{
			ExData.CLBuffer::Dump("exdata.dmp", pGPU->CLQueue[0]);
			BoundCh.CLBuffer::Dump("bound.dmp", pGPU->CLQueue[0]);
		}
		IFLOG(LOG_GPU) GPUStat = pGPU->PrintStats(true);

	}
cl_error_occured:
	if ( cl_error ) printf("%s%s", _PID, ClErrorName(cl_error));
	RETURN(cl_error);
#endif

CPU_Mode:
	Clear(CLR_DATA);
	GenerateInitialPulse();

	IFLOG(LOG_DUMP) SaveInitial("initpulse", SAVE_UU | SAVE_TSTEP | EXPORT_MAT);

	constexpr int ASYNC = Parallel::RUN_ASYNC;
	Parallel PP[] =
	{ fDirect ?
		Parallel(GetShader("SaveExData2D",ShaderObj::FL_2D, 0, 1),              (cl_ulong)this, NumDetectorsA,ActiveSrcs,0,1,ExData.LineWidthB, Parallel::RUN_2D_YMARCH | ASYNC) :
		Parallel(GetShader("SaveBound2D",ShaderObj::FL_2D, 0, 1),              (cl_ulong)this, NumBoundaryA,ActiveSrcs,0,1,BoundCh.LineWidthB, Parallel::RUN_2D_YMARCH | ASYNC),

		Parallel(GetShader("BoundaryCond",ShaderObj::FL_1D, 0, 0),             (cl_ulong)this, ActiveSrcs * 4, 0, 0, 0, 0, Parallel::RUN_1D),
		Parallel(GetShader("ForwardWaveUni", ShaderObj::FL_2D, 0,VectorSize * 2),(cl_ulong)this, InW, InH*ActiveSrcs, 0, BLKs.FwdY, DY, Parallel::RUN_2D_YMARCH),
		Parallel()
	};

	for ( int i = sizeof(PP) / sizeof(PP[0]) - 2; i >= 0; --i )
		if ( !PP[i].pObj )
		{ printf("\nRunDirect: shader[%i] not found", i); return 1; }

	std::function<int(int)> DirectCPU = [this, &PP] (int Span) ->int
	{ return PP[fSkipScalars ? 2 : 0].Run(RunMethod, T, Span); };

	RunLoop(&DirectCPU, true);

	IFLOG(LOG_KRNL) PP->Print();
	RETURN(0);
}

int CData::RunInverse(void)
{
	int cl_error = 0; cl_ulong MppsTime; cl_ulong tmpl;
	T = NFrames + T_START - 1;
	//SaveInitial(NULL, SAVE_UW | EXPORT_MAT | EXPORT_FULL | SAVE_TSTEP | SAVE_ALT_NAME);	
	ENTER;

#ifdef TARGET_GPU
	if ( !EnableGPU ) goto CPU_Mode;
	{ struct BlockSizes BSZ;
	ValidateBlockSizes(BSZ);

	int GH = InH / BSZ.BwdY, GW = InW / 4,
		LW = FACTORIZE(GW, BSZ.BwdW),
		LH = FACTORIZE(InH, BSZ.BwdW);

	MRange  F44Rng("F44Rng", GW, GH, GW, GH, LW, 1, 0, 1);
	MRange3 F44Par("F44Par", GW, GH, ActiveSrcs, GW, GH, ActiveSrcs, LW, 1, 1, 0, 1);
	MRange  BoundR1("BndR1", NumBoundaryA, ActiveSrcs,
									NumBoundaryA, ActiveSrcs, DTCTR_ALIGN, 1, 0, 1);

	MRange  ExDataR1("ExdR1", NumDetectorsA, ActiveSrcs,
									 NumDetectorsA, ActiveSrcs, DTCTR_ALIGN, 1, 0, 1);
	LW = FACTORIZE(InW, 64);		LH = FACTORIZE(InH, 16);
	MRange3  NRBC_Hrz("NRBC_H", InW, 2, ActiveSrcs, InW, 2, ActiveSrcs, LW, 1, 1, 0, 1);
	MRange3  NRBC_Vrt("NRBC_V", InH, 2, ActiveSrcs, InH, 2, ActiveSrcs, LH, 1, 1, 0, 1);

	if ( !pGPU->fReady ) return CL_DEVICE_NOT_AVAILABLE;
	if ( !GLayerBuf.Len ) return CL_INVALID_MEM_OBJECT;
	if ( UploadToGPU() ) return CL_INVALID_MEM_OBJECT;

	std::function<int(int)> InvGPU =
		[this, &F44Rng, &F44Par, &BoundR1, &NRBC_Hrz, &ExDataR1, &NRBC_Vrt, &cl_error] (int _Span) ->int
	{
		const bool fProfilingEvent = (G_Settings.LogLevel&LOG_GPUPROF);
		cl_ulong tmpl; int winc;
		cl_int cl_error = 0;
		if ( _Span >= 0 ) return 0;
		_Span = -_Span;
		IFLOG(LOG_TSTEP) printf("\n Inverse: T=%i .. %i ", T, T - _Span + 1);

		clFinish(pGPU->CLQueue[0]);
		for ( int I = 0; (I < _Span) && !cl_error; ++I )
		{
			cl_int TT = T - I, TT1 = TT;
			float nvscale = (NFrames - TT) *FadeTime;
			nvscale = CLAMP(nvscale, 0.0f, 1.0f);

			IFLOG(LOG_TSTEP) printf(" InvT=%i ", TT);
			CL_TRY(clSetKernelArg((*pGPU)[k_BackwardWave44], 1, sizeof(cl_int), &TT));
			CL_TRY(clSetKernelArg((*pGPU)[k_BackwardWave44s], 1, sizeof(cl_int), &TT));
			CL_TRY(clSetKernelArg((*pGPU)[k_BackwardWave44pg], 1, sizeof(cl_int), &TT));

			if ( !fSkipScalars )
			{
				tmpl = pgLW.s[TT & 1];	winc = pgLW.s[1 - (TT & 1)] - tmpl;

				CL_TRY(clSetKernelArg((*pGPU)[k_NRBCond_Vrt], 1, sizeof(cl_ulong), &tmpl));
				CL_TRY(clSetKernelArg((*pGPU)[k_NRBCond_Vrt], 2, sizeof(cl_int), &winc));
				CL_TRY((*pGPU)[k_NRBCond_Vrt].ExecNoWait(pGPU->CLQueue[0], NRBC_Vrt, true));

				CL_TRY(clSetKernelArg((*pGPU)[k_NRBCond_Hrz], 1, sizeof(cl_ulong), &tmpl));
				CL_TRY(clSetKernelArg((*pGPU)[k_NRBCond_Hrz], 2, sizeof(cl_int), &winc));
				CL_TRY((*pGPU)[k_NRBCond_Hrz].ExecNoWait(pGPU->CLQueue[0], NRBC_Hrz, fProfilingEvent));

				CL_TRY(clSetKernelArg((*pGPU)[k_LoadExData], 2, sizeof(cl_int), &TT1));
				CL_TRY(clSetKernelArg((*pGPU)[k_LoadExData], 3, sizeof(float), &nvscale));
				CL_TRY(clSetKernelArg((*pGPU)[k_LoadBound], 2, sizeof(cl_int), &TT));

				CL_TRY((*pGPU)[k_LoadBound].ExecNoWait(pGPU->CLQueue[0], BoundR1, fProfilingEvent));

				CL_TRY((*pGPU)[k_LoadExData].ExecNoWait(pGPU->CLQueue[0], ExDataR1, fProfilingEvent));

				CL_TRY(clEnqueueBarrier(pGPU->CLQueue[0]));
				//clEnqueueWaitForEvents(pGPU->CLQueue[0], 1, &(*pGPU)[k_NRBCond_Vrt].evt);
				//if ( !fProfilingEvent ) clReleaseEvent((*pGPU)[k_NRBCond_Vrt].evt);
			}

			if ( GParGrad.Addr )
			{ CL_TRY((*pGPU)[k_BackwardWave44pg].ExecNoWait(pGPU->CLQueue[0], F44Par, fProfilingEvent)); } else
			{ CL_TRY((*pGPU)[(GPUSrcs == 1) ? k_BackwardWave44s : k_BackwardWave44].ExecNoWait(pGPU->CLQueue[0], F44Rng, fProfilingEvent)); }

			CL_TRY(clEnqueueBarrier(pGPU->CLQueue[0]));

			if ( fProfilingEvent )
			{
				(*pGPU)[k_LoadExData].Profile(true);
				(*pGPU)[k_LoadBound].Profile(true);
				(*pGPU)[k_NRBCond_Vrt].Profile(true);
				(*pGPU)[k_NRBCond_Hrz].Profile(true);
				(*pGPU)[GParGrad.Addr ? k_BackwardWave44pg : (GPUSrcs == 1) ? k_BackwardWave44s : k_BackwardWave44].Profile(true);
			}
		} // for time

		PRINT_CL_ERROR;
		clFinish(pGPU->CLQueue[0]);
		return cl_error;
	}; // lambda

	SETARG0(k_BackwardWave44);
	SETARG0(k_BackwardWave44s);
	SETARG0(k_BackwardWave44pg);
	SETARG0(k_LoadExData);
	SETARG0(k_SaveExData);
	SETARG0(k_LoadBound);
	SETARG0(k_NRBCond_Hrz);
	SETARG0(k_NRBCond_Vrt);
	CL_TRY(clSetKernelArg((*pGPU)[k_BackwardWave44], 2, sizeof(cl_int), &BSZ.FwdY));
	CL_TRY(clSetKernelArg((*pGPU)[k_BackwardWave44s], 2, sizeof(cl_int), &BSZ.FwdY));
	CL_TRY(clSetKernelArg((*pGPU)[k_BackwardWave44pg], 2, sizeof(cl_int), &BSZ.FwdY));
	CL_TRY(clSetKernelArg((*pGPU)[k_BackwardWave44], 3, sizeof(cl_int), &ActiveSrcs));
	tmpl = ExData.GPUAddr(0, 0);
	CL_TRY(clSetKernelArg((*pGPU)[k_LoadExData], 1, sizeof(cl_ulong), &tmpl));
	CL_TRY(clSetKernelArg((*pGPU)[k_SaveExData], 1, sizeof(cl_ulong), &tmpl));
	tmpl = BoundCh.GPUAddr(0, 0);
	CL_TRY(clSetKernelArg((*pGPU)[k_LoadBound], 1, sizeof(cl_ulong), &tmpl));

	RunLoop(&InvGPU, false);

	IFLOG(LOG_GPU) GPUStat = pGPU->PrintStats(true);

	PRINT_CL_ERROR; 	RETURN(cl_error);
	} // localblock


#endif

CPU_Mode:

	constexpr int ASYNC = Parallel::RUN_ASYNC;
	Parallel PP[] =
	{
		Parallel(GetShader("LoadExData2D",ShaderObj::FL_2D, 0, 1),              (cl_ulong)this, NumDetectorsA,ActiveSrcs,0,1,ExData.LineWidthB, Parallel::RUN_2D_YMARCH | ASYNC),
		Parallel(GetShader("LoadBound2D",ShaderObj::FL_2D, 0, 1),              (cl_ulong)this, NumBoundaryA,ActiveSrcs,0,1,BoundCh.LineWidthB, Parallel::RUN_2D_YMARCH | ASYNC),
		Parallel(GetShader("BoundaryCond",ShaderObj::FL_1D, 0, 0),          (cl_ulong)this, ActiveSrcs * 4, 1, 0, 0, 0, Parallel::RUN_1D),
		// H=1 apply bound.cond. to W
		Parallel(GetShader("BackwardWave",  ShaderObj::FL_2D, 0,VectorSize),(cl_ulong)this, InW, InH, 0, BLKs.BwdY, DY, Parallel::RUN_2D_YMARCH),
		Parallel()
	};

	for ( int i = sizeof(PP) / sizeof(PP[0]) - 2; i >= 0; --i )
		if ( !PP[i].pObj )
		{ printf("\nRunInverse: shader[%i] not found", i); return 1; }

	Clear(CLR_W);

	std::function<int(int)>InvCPU = [this, &PP] (int Span) ->int
	{ return PP[fSkipScalars ? 3 : 0].Run(RunMethod, T, Span); };

	RunLoop(&InvCPU, false);

	IFLOG(LOG_KRNL) PP->Print();
	RETURN(cl_error);
}


void CData::Test(const char* ShName)
{
	ENTER;
	LoadInitial();
	if ( !MSizeA ) { puts("Test: Object not ready"); return; }

	int NT = omp_get_max_threads();
	NT = NThreads ? NThreads : NT;
	omp_set_num_threads(NT);

	if ( RunArch ) ShaderObj::ActiveArch = RunArch;

	Parallel PP[] =
	{ Parallel(GetShader(ShName,ShaderObj::FL_2D,0,VectorSize),
							(cl_ulong)this, InW, InH, 0, BLKs.FwdY, DY,
							 Parallel::RUN_2D_YMARCH | Parallel::RUN_LAST_ITEM) };

	T = T_START;
	NT = nDump ? nDump : NFrames;
	while ( T <= NFrames )
	{
		PP->Run(RunMethod, T, NT);		T += NT;
		if ( nDump ) Dump();
	}
	PP->Print();  ++TestNumber;
	return;

	RETURN();
} //fn  */


void CData::GenerateAbsorbField(void)
{
	const float SZ = __min(InW, InH)*0.5f,
		RcR = __max(SrcRadius, RcvRadius) / dx + 10;
	float NLines = (SZ - RcR)*0.9f;
	NLines = __max(NLines, 1);
	float R0 = SZ - NLines - 1,
		CX = InW * 0.5f, CY = InH * 0.5f;

	for ( int J = 0; J < InH; ++J )
	{
		const float CJ = (J - CY)*(J - CY);
		int* const dest = (int*)(pCoeff + J * DY);
		for ( int I = 0; I < InW; ++I )
		{
			float R = sqrt((I - CX)*(I - CX) + CJ) - R0;
			R = __min(R, NLines);
			R = __max(R, 0);
			R /= NLines; R *= R; 	R *= AbsorbA;
			if ( R > 0 ) dest[I] = _PackFloatCA(Vel, R);
		}
	}
}


void CData::GenerateEmitters(void)
{
	NSrcs = __min(NSrcs, MAX_SRCS);

	const float SrcR = SrcRadius / dx,
		hW = InW * 0.5f, hH = InH * 0.5f,
		AngleStep = 2 * M_PI / GlobalSrcs;
	
	for ( int I = 0; I < NSrcs; ++I )
	{
		const float A = AngleStep * (StartGlobalSrc + I) + M_PI * 0.625;
		SrcPos[I].s[0] = SrcR * cos(A) + hW;
		SrcPos[I].s[1] = -SrcR * sin(A) + hH;
		//printf("%s GenEmitters: ST=%i  [%i]=%f %f\n", StartGlobalSrc, I, SrcPos[i].s[0], SrcPos[i].s[1]);
	}
}

void CData::GenerateDetectors(void)
{ // Generate detector circle with spacing

	const float RcvR = RcvRadius / dx,
		AngleStep = __max(RcvSpacing / dx, 0.25f) / RcvR;
	int ND, N, i, Idx;
	DETECTOR D0;

	N = floor(2 * M_PI / AngleStep);
	for ( i = ND = 0; i < N && ND < MaxDetectors; ++i )
	{
		const float A = AngleStep * i;
		const DETECTOR D(RcvR*cos(A), -RcvR * sin(A));
		if ( D.val != D0.val ) D0 = pDetectors[ND++] = D;
	}

	NumDetectors = ND;
	NumDetectorsA = align(ND, BND_ALIGN);
	QSort<DETECTOR>(pDetectors, 0, NumDetectors - 1);
}


void CData::SaveSrcRcv(void)
{
	STR64 tmpn;

	const int D[4] = { 2, NumDetectorsA, 0, 0 };
	const int S[4] = { 2, GlobalSrcs, 0, 0 };

	snprintf(tmpn, sizeof(tmpn), "%s.rcv", *DirectBaseName ? DirectBaseName : "_");
	if ( !MyId )
	{
		MATHeader(D, MATHeader::miINT16, tmpn).WriteFile(pDetectors);
		FWrite(tmpn, pDetectors, sizeof(DETECTOR)*NumDetectors);
	}

	snprintf(tmpn, sizeof(tmpn), "%s.src", *DirectBaseName ? DirectBaseName : "_");
	
	//for (int i = 0; i < NSrcs; ++i)	{		printf("%s SRC[%i]= %f %f \n", _PID, i, SrcPos[i].s[0], SrcPos[i].s[1]);	}

#ifdef USE_MPI
	if ( NumProcesses > 1 )
	{
		int ret = 0;
		constexpr int sz = sizeof(SrcPos[0]);		
		memblk AllSrcPos(GlobalSrcs * sz + 2 * sizeof(int)*NumProcesses);
		if ( CollectRetval(!AllSrcPos.data) ) return;		

		int* const pCounts = (int*)(AllSrcPos.data + (GlobalSrcs * sz));
		int* const pDispls = pCounts + NumProcesses;
		for ( int i = 0; i < NumProcesses; ++i )
		{
			cl_uint2 r = SelectPart(GlobalSrcs, i);
			pDispls[i] = r.s[0] * sz;
			pCounts[i] = (r.s[1] - r.s[0])*sz;
		}
		ret=MPI_Gatherv(SrcPos, NSrcs*sz, MPI_CHAR,
								AllSrcPos.data, pCounts, pDispls, MPI_CHAR,
								0, MPI_COMM_WORLD);
		if(!MyId) MATHeader(S, MATHeader::miSINGLE, tmpn).WriteFile(AllSrcPos.data);
} else
#endif
MATHeader(S, MATHeader::miSINGLE, tmpn).WriteFile(SrcPos);
}


void CData::GenerateBound(void)
{ /* Generates BoundLines
	Fills detector positions in BoundLines - requires detectors built
	Fills boundary ranges for each line
	Computes boundary size for allocation by InitDataCache()
	*/

	if ( !NumDetectors ) GenerateDetectors();
	const int Xc = (InW >> 1), Yc = (InH >> 1);
	int X0, Y, DI = 0;
	DETECTOR D;
	BOUNDLINE L; // Span0Start Span0Len Span1Start Span1Len BoundIdx 0 DetIdx NDets

	Clear(CLR_BND);

	for ( Y = 0; Y < InH; ++Y )
	{
		const int Yo = Y - Yc;
	readd: D = pDetectors[DI];
		if ( D.y() < Yo )
		{ if ( ++DI < NumDetectors ) goto readd; else break; }

		if ( D.y() > Yo ) continue;

		L.DetIdx = X0 = DI;
		while ( ((pDetectors[DI]).y() == Yo) && (DI < NumDetectors) ) ++DI;
		L.NDets = DI - X0;
		pBoundLines[Y] = L;
	}


	int BI = 0;
	if ( BoundThick > 0 )
	{
		const float RcvPix = (RcvRadius / dx + BoundThick + 1.0f);
		for ( Y = 0; Y < InH; ++Y )
		{
			DI = 0;
			L = pBoundLines[Y]; L.BoundIdx = BI;
			for ( int X = 0; X < InW; ++X )
				if ( fabs(hypot((float)(X - Xc), (float)(Y - Yc)) - RcvPix) < BoundThick )
				{
					if ( !(DI & 1) ) { X0 = X; ++DI; }
				} else
					if ( DI & 1 )
					{
						L.spans[DI - 1] = X0; 	L.spans[DI] = X - X0; BI += (X - X0);
						(++DI) &= 3;
					}// for X - if
				pBoundLines[Y] = L;
		}// for Y
	}
	NumBoundary = BI;
	NumBoundaryA = align(BI, BND_ALIGN);

	IFLOG(LOG_DUMP)
	{
		int S[4] = { sizeof(BOUNDLINE) >> 1, InH, 0, 0 };
		MATHeader(S, MATHeader::miINT16, "boundlines").WriteFile(pBoundLines);
		FWrite("boundlines.dat", pBoundLines, sizeof(BOUNDLINE)*InH);
	}

}//fn

inline float approx_acos(float M)
{ return (((-0.018793*M + 0.074261)*M - 0.2121144)*M + 1.5707288)*(M > 1.0f ? 0 : sqrt(1.0f - M)); }


void CData::GenerateInitialPulse(void)
{
	ENTER
		if ( ImpStart <= 0 )
		{
			ImpStart = 1.5f*ImpWidth;// *1.5f + 8.0f*dx;
			printf("\nImpStart auto-set to %.3f mm ", ImpStart);
#ifdef TARGET_GPU
			if ( EnableGPU ) UploadToGPU();
#endif
		}

#ifdef TARGET_GPU
	if ( EnableGPU )
	{
		int cl_error = 0;
		int cw = FACTORIZE(InW, 32),
			ch = FACTORIZE(InH, pGPU->MaxBlockSize / cw);
		MRange3 InitRng("InitRng", InW, InH, ActiveSrcs, InW, InH, ActiveSrcs, cw, ch, 1, 0, 1);

		SETARG0(k_Initialize);
		clSetKernelArg((*pGPU)[k_Initialize], 1, 4, &StartSrc);
		cl_error |= (*pGPU)[k_Initialize].ExecWait(pGPU->CLQueue[0], InitRng);
		if ( cl_error ) puts(ClErrorName(cl_error));

		IFLOG(LOG_DUMP)
		{
			GLayerBuf.Dump("initpGPU.dmp", pGPU->CLQueue[0]);
		}
	} else
#endif

	{
		const float WaveShift = dt / dx * Vel,
			ImpClipLen = ImpClipAMax - ImpClipAMin,
			Amp = 100.0f / sqrt(ImpWidth),
			CX = InW * 0.5f, CY = InH * 0.5f,
			BegImp = ImpStart / dx,
			WidthImp = ImpWidth / dx,
			EndImp = BegImp + WidthImp, RWidthImp = 1.0f / WidthImp;
		const int UsedSrcs = __min(ActiveSrcs, HostSrcs);


#pragma omp parallel for		
		for ( int SrcIdx = 0; SrcIdx < UsedSrcs; ++SrcIdx )
		{
			float SrcX = SrcPos[SrcIdx + StartSrc].s[0], SrcY = SrcPos[SrcIdx + StartSrc].s[1];
			float OrthX = CX - SrcX, OrthY = CY - SrcY,
				OrthLen = hypot(OrthX, OrthY);
			OrthX /= OrthLen; OrthY /= OrthLen;

			int I, J,
				IMAX = SrcY + EndImp + 2, JMAX = SrcX + EndImp + 2,
				IMIN = SrcY - EndImp - 2, JMIN = SrcX - EndImp - 2;

			IMIN = __max(IMIN, 0);	JMIN = __max(JMIN, 0);
			IMAX = __min(IMAX, InH);	JMAX = __min(JMAX, InW);

			for ( I = IMIN; I < IMAX; ++I )
			{
				const cl_ulong Base = I * DY + SrcIdx * DZ;
				float* const Dest0 = (float*)(pU[0] + Base);
				float* const Dest1 = (float*)(pU[1] + Base);
				float DotProdY = (I - SrcY)*OrthY, DotLenY = (I - SrcY)*(I - SrcY);

#pragma simd
				for ( J = JMIN; J < JMAX; ++J )
				{
					float R = sqrt(DotLenY + (J - SrcX)*(J - SrcX));
					R = __max(R, 1.0f);
					float f0 = R - BegImp, f1 = f0 - WaveShift;
					f0 = __max(f0, 0); f1 = __max(f1, 0);
					f0 = __min(f0, WidthImp)*RWidthImp*M_PI,
						f1 = __min(f1, WidthImp)*RWidthImp*M_PI;

					float fade = (DotProdY + (J - SrcX)*OrthX) / R;
					fade = (M_PI - approx_acos(fade)) / M_PI;
					fade = CLAMP(fade, ImpClipAMin, ImpClipAMax) - ImpClipAMin;
					fade = CLAMP(fade, 0, ImpClipLen);
					fade *= Amp / (ImpClipLen*sqrt(R));
					f0 = sin(f0)*sin(f0*ImpFreq)*fade;
					f1 = sin(f1)*sin(f1*ImpFreq)*fade;
					Dest0[J] = f0;		Dest1[J] = f1;
				} // for J - fill line
			} // for I - next line
		} // parfor
	}// if !GPU
	RETURN();
}// fn
// */


int CData::InitDataCache(void) //new
{ /* builds boundary list from Linebuf, 	uploads to line 0 in BoundCH
	builds detector list , 	uploads to line 0 in ExData
	*/
	int retval = 1, Y, I, BI;
	cl_uint *DestBound = NULL, *DestDetectors = NULL;
	const int X0 = (InW >> 1), Y0 = (InH >> 1);

	ENTER;

	if ( !NumDetectors ) GenerateDetectors();
	if ( !NumBoundary ) GenerateBound();

	if ( !Linebuf.data )
	{ puts("\nInitDataCache: Linebuf not ready "); goto ret; }

	BoundCh.Initialize(NumBoundaryA * 4, 1, NFrames);
	ExData.Initialize(NumDetectorsA * 4, 1 + __max(HostSrcs, GPUSrcs), NFrames);
	if ( BoundCh.Alloc(HostSrcs) || ExData.Alloc(HostSrcs) ) goto ret;

	DestBound = (cl_uint*)BoundCh.HostAddr(-1, 0);
	DestDetectors = (cl_uint*)ExData.HostAddr(-1, ExData.LinesCommon - 1);

	for ( Y = 0; Y < InH; ++Y )
	{
		const int BO = Y * FullW;
		const auto L = pBoundLines[Y];

		BI = L.BoundIdx;
		if ( DestBound )
		{
			for ( I = 0; I < L.spans[1]; ++I ) DestBound[BI++] = ((BO + L.spans[0] + I) << 2);
			for ( I = 0; I < L.spans[3]; ++I ) DestBound[BI++] = ((BO + L.spans[2] + I) << 2);
		}

		BI = L.DetIdx, I = L.NDets;
		if ( DestDetectors ) for ( I = 0; I < L.NDets; ++I )
		{
			DestDetectors[BI++] = ((pDetectors[BI].x() + BO) << 2);
		}
	}


	if ( DestDetectors ) for ( I = 0; I < NumDetectors; ++I )
	{
		DETECTOR d = pDetectors[I];
		DestDetectors[I] = (d.x() + X0) * 4 + (d.y() + Y0)*DY;
	}
	if ( DestBound )
	{
		Y = DestBound[NumBoundary ? (NumBoundary - 1) : 0];
		for ( I = NumBoundary; I < NumBoundaryA; ++I ) DestBound[I] = Y;
	}
	if ( DestDetectors )
	{
		Y = DestDetectors[NumDetectors ? (NumDetectors - 1) : 0];
		for ( I = NumDetectors; I < NumDetectorsA; ++I ) DestDetectors[I] = Y;
	}
	IFLOG(LOG_DUMP)
	{
		if ( DestBound )	ExportImage(EXPORT_RAW, RectWin(DestBound, NumBoundaryA, 1, 32), "bound_idx");
		if ( DestDetectors ) ExportImage(EXPORT_RAW, RectWin(DestDetectors, NumDetectorsA, 1, 32), "det_idx");
	}
	retval = 0;
ret:	RETURN(retval);
}

float4 CData::EstimateGrad(void) const
{
	const float rdt2 = 1.0 / (dt * dt);

	int XC = (InW >> 1), YC = (InH >> 1);
	cl_short2* pG = (cl_short2 *)(pGrad);
	cl_uint* pC = (cl_uint *)(pCoeff);
	cl_short2 mx, mn;
	mx.s[0] = mx.s[1] = 0x8000; mn.s[0] = mn.s[1] = 0x7FFF;
	int OverAreaC = 0, OverAreaA = 0, TotalArea = 1;

	float GradMixR0pix = GradCropR0 / dx,
		GradMixRW = 1.0f / (GradCropR1 / dx - GradMixR0pix);

	for ( int Y = 0; Y < InH; ++Y )
	{
		float YYC = (Y - YC)*(Y - YC);
		for ( int X = 0; X < InW; ++X )
		{
			float R = sqrt(float((X - XC)*(X - XC) + YYC));
			R -= GradMixR0pix; R *= GradMixRW;
			R = 1.0f - R;		R = __min(R, 1.0f);
			if ( R <= 0 ) { pG[X] = { 0,0 }; continue; }

			++TotalArea;
			cl_short2 s = pG[X];
			s.s[0] *= R; s.s[1] *= R;	if ( R < 1 ) pG[X] = s;
			OverAreaC += (((s.s[0] > 32500) || (s.s[0] < -32500)) ? 1 : 0);
			OverAreaA += (((s.s[1] > 32500) || (s.s[1] < -32500)) ? 1 : 0);
			mx.s[0] = __max(mx.s[0], s.s[0]);
			mn.s[0] = __min(mn.s[0], s.s[0]);
			mx.s[1] = __max(mx.s[1], s.s[1]);
			mn.s[1] = __min(mn.s[1], s.s[1]);
		}
		pC += FullW, pG += FullW;
	}

	float4 retval =
	{ (float)(__max(mx.s[0], -mn.s[0])) ,
		(float)(__max(mx.s[1], -mn.s[1])),
		(float)OverAreaC / (float)TotalArea,
		(float)OverAreaA / (float)TotalArea };
	return retval;
}

void CData::UpdateCoeff(float fc, cl_float2* fullgrad) const
{
	const float rdt2 = 1.0 / (dt * dt);
	int XC = (ReqW >> 1), YC = (ReqH >> 1);
	cl_uint* pC = (cl_uint *)(pCoeff + ReqOffsetB);

	float GradMixR0pix = GradCropR0 / dx, GradMixRW = 1.0f / (GradCropR1 / dx - GradMixR0pix);	

	for ( int Y = 0; Y < ReqH; ++Y )
	{
		float YYC = (Y - YC)*(Y - YC);

		cl_float2 * const pG = fullgrad ? (fullgrad + Y * FullW) : (cl_float2*)pU[0];
		if ( !fullgrad )
		{
			cl_short2* const pgrad = (cl_short2 *)(pGrad + ReqOffsetB + Y * DY);
#pragma simd
			for ( int X = 0; X < ReqW; ++X )
			{
				pG[X].s[0] = pgrad[X].s[0] / GradScale.s[0];
				pG[X].s[1] = pgrad[X].s[1] / GradScale.s[1];
			}
		}

		for ( int X = 0; X < ReqW; ++X )
		{
			float R = sqrt(float((X - XC)*(X - XC) + YYC));
			R -= GradMixR0pix; R *= GradMixRW;
			R = CLAMP(R, 0.0f, 1.0f);
			R = 1.0f - R * GradEraseAtUpdate;

			cl_float2 ca = _UnpackFloatCA(pC[X]);

			if ( !*(cl_long*)(pG + X) ) continue;

			float dtdtcc = rdt2 / (ca.s[0] * ca.s[0]),
				cg = pG[X].s[0] * fc * rdt2*rdt2;
			dtdtcc -= cg;
			dtdtcc = sqrt(1.0 / dtdtcc) / dt;
			dtdtcc = CLAMP(dtdtcc, Limit_C.s[0], Limit_C.s[1]);
			ca.s[0] = dtdtcc;

			float a2dt = ca.s[1];
			a2dt -= pG[X].s[1] * fc*AttGradScale / dt;
			a2dt = CLAMP(a2dt, 0, Limit_A);
			ca.s[1] = a2dt;

			pC[X] = _PackFloatCA(ca.s[0] * R + Vel * (1.0f - R), ca.s[1] * R);
		}
		pC += FullW;
	}
}
