#pragma once

#include "cmdl.h"

#include <cmath>
#include <functional>

#include "gpu_util.h"
#include "gpudata.h"

enum
{
	EXPORT_FIRST = 0x100,
	EXPORT_RAW = 0x100, EXPORT_MAT = 0x200, EXPORT_PBM = 0x400, // available file formats
	EXPORT_LAST = 0x400, EXPORT_FORMATS = EXPORT_LAST - EXPORT_FIRST,
	EXPORT_MASK = 0x700, // used to loop over file formats
	EXPORT_FULL = 0x800 // output full array, not clipped to the requested area
};

enum{ E_FINISHED=0x8000 };

struct BOUNDLINE
{
	cl_short spans[4];	cl_int BoundIdx;
	cl_ushort DetIdx; cl_short NDets;
};

struct DETECTOR
{
	cl_int val;
	inline DETECTOR() { val = 0x80008000; }
	inline DETECTOR(float x, float y)
	{ val = ((int)round(x) & 0xFFFF) | ((int)round(y) << 16); }
	inline bool operator<=(DETECTOR d) { return (val <= d.val); }
	inline int y() { return (val >> 16); }
	inline int x() { return (int)(short)(val & 0xFFFF); }
};//struct

// File header used to restart hte iterations
struct RestartStruct
{
	char Signature[4];
	cl_uint ReqW = 640, ReqH = 640, CurrentIteration = 0;

	float dx0 = 0.0f, dt0 = 0.0f, Vel0 = 1.5,
		OldResidual = 0, StepSize = 0, StepIncr = 0, TargetStepCoeff = 0,
		CGScale = 0, AGScale = 0, rsvd[3];

	RestartStruct()
	{
		memset(this, 0, sizeof(*this));
		Signature[0] = 'R'; Signature[1] = 'S'; Signature[2] = 'T'; Signature[3] = 'F';
	}

	void Print(int level = 0) const
	{
		printf("%s---RST---  %4i * %4i  #%-3i  dx=%.4f dt=%.4f ", _PID,
					 ReqW, ReqH, CurrentIteration, dx0, dt0);
		if ( level > 0 )
			printf("%s Res=%.4f Mul=%.4f KL=%.4f TGsc=%.4f GScl=[%.4e %.4e]", _PID,
						 OldResidual, StepSize, StepIncr, TargetStepCoeff, CGScale, AGScale);
	}
};


struct TestRecord :BlockSizes
{
	unsigned short size, flags, NThreads, NSrcs;
	float FwdMPPS, BwdMPPS;

	void Print(void) const
	{
		printf("\n sz=%4i th=%2i srcs=%2i Fwd:%8.2f  Bwd:%8.2f  blks: %3i %3i %3i %i ",
			(int)size, (int)NThreads, (int)NSrcs, FwdMPPS, BwdMPPS, FwdY, BwdY, FwdW, BwdW);
	}
};


class CDataConfig : public RestartStruct
{ // Plain data structure. No virtuals.	
public:
	float LAPL_BASE[8] = { -3.46,  0.631, 0.2575, 0.00825, -0.0165, 0.00125,0,0 };

	cl_uint  OutFrameX = 8, OutFrameY = 8,
		ImgAlignY = 40, ImgAlignX = 32, ImgShearX = 16,
		ReqOffsetB = 0,

		NThreads = 0, MaxGPUThreads = 256,
		RunMethod = 0, RunArch = 0,
		VectorSize = 0, Mode = 0, nDump = 0,
		EnableGPU = 1,

		TestNumber = 0, TestIncrSize[2] = { 0,0 }, TestIncrSrc = 0,
		MaxIterations = 250, SaveIterations = 10;

	float StepUpCoeff = 1.25f, StepDownCoeff = 0.3f,MinThreshold=0.0003;

	int TestSizes[3] = { 0,0,0 },
		RepeatTest = 1, fAutoTune = 0, fSkipScalars = 0,
		PerfTest = 0, PerfTestParam = 0,
		fDemoMode =
#ifdef DEMO_MODE
		1;
#else
		0;
#endif		

	float TestVolume = 5.0e8f, TestDuration = 0.8f,
		GPUStat = 0.0f,
		Stat_TaskTime = 0.0f, Stat_TaskMPPS = 0.0f; int Stat_FrameCount = 0;

	float PBM_CLim = 0.2f, PBM_ALim = 0.015f;

	float RcvRadius = 0.0f, RcvSpacing = 0.3f, BoundThick = 1.44f,
		SrcRadius = 0.0f, DefRcvRadius = 0.65,
		PhantomSize = 0.0f, PhantomContrast = 0.0f,
		TimeBuf = 1.35,
		ContrastC = 1, ContrastA = 1,
		AbsorbA = -1,

		TFadeRadius = 0, FadeTime = 0,
		TargetGradAmpl = 0.004,
		GradStartR0 = -0.25, GradStartR1 = -0.45,
		GradCropR0 = 0, GradCropR1 = 0,
		GradEraseAtUpdate = 0.0f;


	float Limit_A = 0.05,AttGradScale=1.0f;
	cl_float2 Limit_C = { 1.0, 2.5 };

	cl_uint MaxDetectors = 8192 * 16, MaxBoundary = 8192 * 16,
		GPUMemorySize = 3600,
		CPUMemorySizeMB =
#ifdef TARGET_ARM 
		50,
#else
		22,
#endif
		NSrcs = 0, GlobalSrcs=0, StartGlobalSrc=0;

	int WaveExportFormats = 0, ImageExportFormats = 0;

	STR64 DirectBaseName = "", InverseBaseName = "", RestartName = "",
		WaveExportFmtNames = "MAT", ImageExportFmtNames = "MAT";
}; // struct


class CData : public CDataConfig, public GPU_DATA
{// Plain data structure. No virtuals.	
public:
	static constexpr int SAVE_RST = 0x10000,
		SAVE_U0 = 1, SAVE_U1 = 2, SAVE_W0 = 4, SAVE_W1 = 8,
		SAVE_COEFF = 0x10, SAVE_GRAD = 0x20,
		SAVE_UU = SAVE_U0 | SAVE_U1, SAVE_WW = SAVE_W0 | SAVE_W1,
		SAVE_UW = SAVE_UU | SAVE_WW,
		SAVE_TSTEP = 0x1000, SAVE_ALT_NAME = 0x2000;

	static constexpr int MODE_DIRECT = 1, MODE_DUMP = 0x10, MODE_CHOP = 0x20;

	static constexpr int CLR_COEFF = 1, CLR_DATA = 2, CLR_W = 4, CLR_GRAD = 8, CLR_DETS = 0x10, CLR_BND = 0x20;

	static constexpr int C_SCALE_BITS = 6, A_SCALE_BITS = 3;
	static constexpr float C_SCALE = 1.0f / (float)(1ULL << (32 - C_SCALE_BITS)),
		A_SCALE = 1.0f / (float)(1ULL << (16 - A_SCALE_BITS)),
		FIELD_SIZE_mm = 250.0f;

	static constexpr int T_START = 1, BND_ALIGN = DTCTR_ALIGN, H_Align = 32, V_Align = 10;

	static constexpr int PHANTOM_NONE = 0, PHANTOM_AUTO = 1, PHANTOM_FILE = 2, PHANTOM_RST = 4,
		PHANTOM_ALLOW_REINIT = 1, PHANTOM_ALLOW_AUTO = 2;

	static const OPTDEF *const Options;
	static const char *const AutotuneFile, *const PerftestFile, *const INIFile;

	int  HostSrcs = 0, StartSrc = 0, ActiveSrcs = 0, HostBoundLines = 0;

	cl_ulong Buf_a, MSizeA, pU[2], pW[2], pGrad, pPrevCoeff, pCoeff;
	BOUNDLINE* pBoundLines;
	DETECTOR* pDetectors;
	int T = T_START; // struct cleared to zero at startup up to this point

	memblk  Mainbuf, Linebuf, SrcList;
	FileMapArray ExDataFileMaps;

	//-------------------- 

	void _clr(void)
	{
		memset(&Buf_a, 0, (char*)(&T) - (char*)(&Buf_a));
		NumDetectors = NumBoundary = 0;
	}

	void Close(void)
	{
		ENTER;		Release();
		BoundCh.Free(); ExData.Free();	Mainbuf.free();
		Linebuf.free();	SrcList.free(); ExDataFileMaps.close();
		_clr();
		RETURN();
	}

	inline CData() { _clr(); }
	inline ~CData() { Close(); }

	// Init-Load-Save
	int Init(int w, int h, int d = 1);
	int Init(RestartStruct& RST);
	void Clear(int what) const;

	int LoadInitial(const char* name = NULL);
	void SaveInitial(const char* name = NULL, int mode = SAVE_RST | SAVE_UW | EXPORT_MAT, int Src0 = -1);
	void SavePhantom(const char* name = NULL, int mode = 1);
	int LoadRST(const char* name = NULL);
	int LoadPhantom(const char* name = NULL, int Flags = PHANTOM_ALLOW_AUTO);
	int LoadPhantom(memblk& m, bool AllowReinit = false);
	void Print(int level = 0) const;
	void Dump(void) const;

	// GPU
	int InitGPU(GPUContext* pGpu = NULL, cl_uint device = 0);
	void XferToGPU(bool Up, cl_int what, cl_uint SrcI = 0) const;

	// Main operations
	int RunLoop(std::function<int(int)> *callback, bool Direct = true);
	int RunDirect(bool fDirect = true);
	int RunInverse(void);
	int SolveDirect(bool testmode = false);
	int SolveInverse(bool testmode = false);
	int Process(void);
	// Used by main operations
	float ComputeResidual(void);
	void UpdateCoeff(float k, cl_float2* fullfrad=NULL) const;
	float4 EstimateGrad(void) const;
	void BackstepCoeff(float fc, const cl_ushort2* pPrevCoeff);

	// Secondary ops	
	void GenerateEmitters(void);
	void GenerateDetectors(void);
	void SaveSrcRcv(void);
	void GenerateInitialPulse(void);
	void GenerateAbsorbField(void);
	void GenerateBound(void); // Detectors must be built
	int InitDataCache(void); // 2 above must be built
	void ExportDataCache(int mode = EXPORT_MAT); // 2 above must be built

	// Phantom
	void GeneratePhantom(cl_ulong seed = 0) const;
	void GenerateSimplePhantom(void) const;
	void GenerateResTarget(float CX, float CY) const;
	void Circle(float X, float Y, float R, float dC, float a = 0.0f) const;

	// Config
	void SetParameters(bool clear = false);
	void ValidateBlockSizes(BlockSizes& B) const;
	void Interactive(void);

	// Perf-tests
	int DetermineCPUCache(int sz = 360);
	int PerfTestRun(int sz, int mode, TestRecord& TR);
	cl_float2 SimulatedRun(void);
	void LoadOptimizedSettings(void);
	void ComputeOptimizedSettings(TestRecord &TR);
	void Test(const char* ShName);

	inline cl_uint _PackFloatCA(float c, float a) const
	{
		c = c ? (1.0f / (c * dt * C_SCALE)) : 0;
		a = (a / (dt*A_SCALE));
		c = __min(c, 0xFFFFFFFFULL);		a = __min(a, 0xFFFF);
		c = __max(c, 0.0f);   		a = __max(a, 0.0f);
		return (((cl_uint)(round(c)) >> 16) | ((cl_uint)(round(a)) << 16));
	}

	inline cl_float2 _UnpackFloatCA(cl_uint V) const
	{
		cl_float2 F;
		F.s[1] = (float)(V >> 16)*CData::A_SCALE*dt;
		V <<= 16;
		F.s[0] = V ? (1.0f / (dt*V*CData::C_SCALE)) : Vel;
		return F;
	}

}; // struct




