/*
WaveTomography software implements the algorithm for solving direct and inverse problems of wave tomography in application to medical ultrasound imaging for breast cancer diagnosis.
The software has been developed at Lomonosov Moscow State University. Copyright (c) 2021 Seryozhnikov S.Y, Romanov S.Y.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// Helper functions


void CData::Print(int level) const
{
	RestartStruct::Print(level);
	GPU_DATA::Print(level);
	if ( level > 0 ) printf(
		"%s---HOST---\n   Buf->%012llX   end=%012llX   NSrcs=%i [host %i]\n"
		"    U0->%012llX   U1->%012llX   DY=0x%04llX  DZ=0x%08llX\n"
		"    W0->%012llX   W1->%012llX\n     C->%012llX    G->%012llX\n"
		"  Line->%012llX   end=%012llX  BndLs->%012llX\n"
		"  Dets->%012llX ", _PID,
		Buf_a, Mainbuf.data + Mainbuf.len, NSrcs, HostSrcs,
		pU[0], pU[1], DY, DZ, pW[0], pW[1], pCoeff, pGrad,
		Linebuf.data, Linebuf.data + Linebuf.len, pBoundLines,
		pDetectors);
}


void ExportImage(int mode, const RectWin& RW, const char* name)
{
	if ( !name ) name = "untitled";
	char tmpname[256];
	strncpy(tmpname, name, 250);
	int namelen = strlen(name);
	if ( mode&EXPORT_MAT )
	{
		int Dim[] = { RW.LWidth,RW.Height,0,0 };
		MATHeader MH(Dim, MATHeader::miSINGLE, name);
		FILE* fp = MH.WriteFile();
		RW.copyfile(fp, true); fclose(fp);
	}
	if ( mode&EXPORT_PBM )
	{
		strcpy(tmpname + namelen, ".pbm");
		DumpFloatImagePBM(tmpname, RW);
	}
	if ( mode&EXPORT_RAW )
	{
		strcpy(tmpname + namelen, ".dat");
		FILE* fp = fopen(tmpname, "wb");
		RW.copyfile(fp, true); fclose(fp);
	}
}


void CData::Dump(void) const
{
	ENTER;
	char vname[32];
	int Dim[4] = { InW,InH,NSrcs,0 };
	MATHeader MH;

	for ( int i = 1; i >= 0; --i )
	{
		snprintf(vname, 27, "U%iT%03i%c", TestNumber, T, i + 'a');
		MH.WriteFile(Dim, MATHeader::miSINGLE, vname, (void*)pU[i]);

		sprintf(vname, "W%iT%03i%c", TestNumber, T, i + 'a');
		MH.WriteFile(Dim, MATHeader::miSINGLE, vname, (void*)pU[i]);

	}
	RETURN();
}


int CData::LoadRST(const char* name)
{
	int retval = PHANTOM_NONE;
	if ( !name ) name = InverseBaseName;
	if ( !name ) return PHANTOM_NONE;
	if ( !*name ) return PHANTOM_NONE;
	STR64 tmpn;	memblk M;
	snprintf(tmpn, sizeof(tmpn), "%s.rst", name);

	M.ReadFile(tmpn); if ( !M.data || M.len < sizeof(RestartStruct) ) return PHANTOM_NONE;
	RestartStruct *pR = (RestartStruct*)M.data;
	if ( memcmp(pR->Signature, Signature, sizeof(Signature)) )
	{ fprintf(stderr, "\nIncorrect restart file '%s' ", name); return PHANTOM_NONE; }
	if ( (pR->ReqW != ReqW) || (pR->ReqH != ReqH) )
	{ fprintf(stderr, "\nRestart file of different size '%s' ", name); return PHANTOM_NONE; }
	if ( retval = LoadPhantom(M) ) 	*(RestartStruct*)this = *pR; // copy if nonzero retval;
	return retval; // return any retval
}


int CData::LoadPhantom(const char* name, int Flags)
{
	if ( !name ) name = DirectBaseName;
	if ( !name ) return PHANTOM_NONE;
	if ( !*name ) return PHANTOM_NONE;
	if ( (name[0] == '#') && (Flags&PHANTOM_ALLOW_AUTO) )
	{
		const cl_ulong F = atol(name + 1);
		Clear(CLR_COEFF);
		GeneratePhantom(F);
		IFLOG(LOG_INFO) printf("%sPhantom generated #%llu ", _PID, F);
		return PHANTOM_AUTO;
	} else
	{
		STR64 tmpn;
		memblk M;
		M.ReadFile(name);
		return LoadPhantom(M, Flags&PHANTOM_ALLOW_REINIT);
	}
}


int CData::LoadPhantom(memblk& m, bool AllowReinit)
{
	ENTER;
	int retval = PHANTOM_NONE, X, Y; cl_uint *p, *dest;
	size_t worksz;
	const char *errmsg = NULL;

	Clear(CLR_COEFF);
	if ( !m.data )
	{ errmsg = "LoadPhantom: empty input ";  goto RetErr; }
	p = (cl_uint*)m.data;

	if ( m.len < sizeof(RestartStruct) ) goto RawPhantom;
	if ( m.data[0] == 'R' && m.data[1] == 'S' &&  m.data[2] == 'T' && m.data[3] == 'F' )
	{
		RestartStruct* pRST = (RestartStruct*)m.data;
		if ( m.len < pRST->ReqW*pRST->ReqH * 4 + sizeof(RestartStruct) )
		{ errmsg = "LoadPhantom: incomplete RST file ";	goto RetErr; }
		if ( (pRST->ReqW != ReqW) || (pRST->ReqH != ReqH) )
		{ // reinit if allowed and size changed in .rst
			if ( !AllowReinit ) {  errmsg = "LoadPhantom: .RST data size mismatch ";	goto RetErr; }			
			if ( Init(*pRST)) { errmsg = "LoadPhantom: Error applying restart data ";	goto RetErr; }
		} else *(RestartStruct*)this = *pRST;
		p = (cl_uint*)(m.data + sizeof(RestartStruct));
		retval = PHANTOM_RST;
	}
RawPhantom:
	{
		worksz = (ReqW*ReqH);
		if ( m.len >= worksz * 8 )
		{
			IFLOGFILE puts(" Loading float phantom ");
			float* pf = (float*)m.data;
			for ( Y = 0; Y < ReqH; ++Y )
				for ( dest = (cl_uint*)(pCoeff + Y * DY + ReqOffsetB),
						 X = 0; X < ReqW; ++X )
				{
					float c = *pf++, a = *pf++;	dest[X] = _PackFloatCA(c, a);
				}
			retval = PHANTOM_FILE;
			goto Loaded;
		}

		if ( m.len < worksz * 4 )
		{
			errmsg = "Incorrect phantom size ";
		RetErr: printf("%s%s", _PID, errmsg ? errmsg : "Error"); return retval;
		}
	}

	IFLOGFILE printf("%sLoading int phantom ", _PID);
	RectWin((char*)(pCoeff + ReqOffsetB), ReqW, ReqH, 32, DY)
		.copymem((char*)p, 0, false);

Loaded:
_ret: IFLOG(LOG_INFO) printf("%sPhantom loaded. ", _PID);
	RETURN(retval);
}//fn


void CData::SaveInitial(const char* name, int mode, int Src0)
{
	ENTER;
	int ret; char* srcptr;

	int Dim[] = { ReqW,ReqH,0 };
	STR64 tmpname, suffix;
	if ( !name ) name = DirectBaseName;
	char AltChr = mode & SAVE_ALT_NAME ? 'b' : '_';
	RectWin RW = (mode&EXPORT_FULL) ? RectWin(NULL, FullW, FullH, 32, DY)
		: RectWin((void*)(InOffsetB + ReqOffsetB), ReqW, ReqH, 32, DY);

	*suffix = 0;

	if ( mode & SAVE_COEFF )
	{
		if ( mode & SAVE_RST )
		{
			if ( (mode&SAVE_TSTEP) && (CurrentIteration) )
				sprintf(suffix, "_i%02i", CurrentIteration);
			dx0 = dx; dt0 = dt; Vel0 = Vel;
			snprintf(tmpname, 63, "%s%s.rst", name, suffix);
			FILE* fd = fopen(tmpname, "wb");
			if ( fd )
			{
				fwrite(this, 1, sizeof(RestartStruct), fd);
				RectWin((char*)(pCoeff + ReqOffsetB), ReqW, ReqH, 32, DY)
					.copyfile(fd, true);
				fclose(fd);
			}
		}
		if ( mode & EXPORT_PBM )
		{
			memblk M(ReqW*ReqH * 6 + sizeof(PBMHEADER));
			if ( !M.data ) return;
			PBMHEADER* pB = (PBMHEADER*)M.data;
			char *dest = M.data + sizeof(PBMHEADER);
			pB->Init(ReqW, ReqH, PBMHEADER::RGB16);
			const float cs = 65536.0f / PBM_CLim, as = 65536.0f / PBM_ALim;

			for ( int J = ReqH; J; --J )
			{
				cl_uint* SrcCoeff = (cl_uint*)(pCoeff + ReqOffsetB + DY * J);
				for ( int I = 0; I < ReqW; ++I )
				{
					const cl_float2 ca = _UnpackFloatCA(SrcCoeff[I]);
					const float cc = (ca.s[0] - Vel)*cs;
					const int
						cp = round(CLAMP(cc, 0.0f, 65535.0f)),
						aa = round(CLAMP(ca.s[1] * as, 0.0f, 65535.0f)),
						cm = round(CLAMP(-cc, 0.0f, 65535.0f));
					dest[0] = (cp >> 8); dest[1] = (cp & 0xFF);
					dest[2] = (aa >> 8); dest[3] = (aa & 0xFF);
					dest[4] = (cm >> 8); dest[5] = (cm & 0xFF);
					dest += 6;
				} // loop pixels
			}// loop Y
			snprintf(tmpname, 63, "%s%s.pbm", name, suffix);
			M.WriteFile(tmpname);
		} // pbm mode

		if ( mode&EXPORT_MAT )
		{
			memblk M(ReqW*ReqH * 8);
			if ( M.data )
			{
				int dim[] = { 2,ReqW,ReqH,0 };
				cl_float2* dest = (cl_float2*)M.data;
				cl_uint* src = (cl_uint*)pCoeff;
				snprintf(tmpname, 63, "%s%s", name, suffix);
				for ( int J = ReqH; J; --J )
				{
					cl_uint* SrcCoeff = (cl_uint*)(pCoeff + ReqOffsetB + DY * J);
					for ( int I = 0; I < ReqW; ++I )
						*dest++ = _UnpackFloatCA(SrcCoeff[I]);
				}
				MATHeader(dim, MATHeader::miSINGLE, tmpname).WriteFile(M.data);
			}
		}

		if ( mode&EXPORT_RAW )
		{
			sprintf(tmpname, "%s.cci", name);
			FILE* fp = fopen(tmpname, "wb");
			if ( fp )
			{
				RectWin((char*)(pCoeff + ReqOffsetB), ReqW, ReqH, 32, DY)
					.copyfile(fp, true); fclose(fp);
			}
		}

	}// rst mode

	memblk gpubuf;
	if ( EnableGPU ) { if ( !gpubuf.alloc(MSizeA) ) return; }

	if ( mode&SAVE_GRAD )
	{
#ifdef TARGET_GPU
		if ( EnableGPU )
		{
			GLayerBuf.Xfer(false, gpubuf.data, pgGrad - InOffsetB, FullH*DY, pGPU->CLQueue[0]);
			srcptr = gpubuf.data + InOffsetB;
		} else
#endif
			srcptr = (char*)pGrad;

		RW.ptr = srcptr + ((mode&EXPORT_FULL) ? (-InOffsetB) : ReqOffsetB);
		FILE* fp = NULL;
		STR64 tmpname;
		snprintf(tmpname, sizeof(tmpname), "%s.grd", name);
		if ( mode&EXPORT_MAT )
		{
			int dim[] = { 2,RW.LWidth,RW.Height,0 };
			fp = MATHeader(dim, MATHeader::miINT16, tmpname).WriteFile();
			RW.copyfile(fp, true);	fclose(fp);
		}
		if ( mode&EXPORT_RAW )
		{
			fp = fopen(tmpname, "wb");
			RW.copyfile(fp, true);	fclose(fp);
		}

	}


	*suffix = 0;
	int SrcOffs, SrcStart, SrcEnd;
	if ( Src0 < 0 ) { SrcStart = 0;  SrcOffs = StartSrc; SrcEnd = ActiveSrcs; } else { SrcStart = Src0 - StartSrc; SrcOffs = Src0; SrcEnd = SrcStart + 1; }
	for ( int Src = SrcStart; Src < SrcEnd; ++Src )
	{
		int CurSrc = Src + SrcOffs;
		sprintf(suffix, "_s%02i", CurSrc);
		if ( mode&SAVE_TSTEP )
		{
			if ( CurrentIteration )
				sprintf(suffix, "_s%02i%c%03ii%02i", CurSrc, AltChr, T, CurrentIteration);
			else
				sprintf(suffix, "_s%02i%c%03i", CurSrc, AltChr, T);
		}

		struct { cl_ulong ptr;  char sig[4]; }
		UWs[4] = { {pU[0],"u0"},{pU[1],"u1"},{pW[0],"w0"},{pW[1],"w1"} },
			gUWs[4] = { {pgLU.s[0],"u0"},{pgLU.s[1],"u1"},{pgLW.s[0],"w0"},{pgLW.s[1],"w1"} };

#define PRINTNAME sprintf(tmpname, "%s_%s%s", name, UWs[i].sig,suffix)
		for ( int mode2 = EXPORT_FIRST; mode2 <= EXPORT_LAST; mode2 <<= 1 )
			if ( mode&mode2 )
			{
				int mode3 = mode & (~EXPORT_MASK | mode2);
				int cond = SAVE_U0;
				for ( int i = 0; i < 4; ++i, cond <<= 1 )
					if ( mode&cond )
					{
						PRINTNAME;
#ifdef TARGET_GPU
						if ( EnableGPU )
						{
							memset(gpubuf.data, 1, gpubuf.len);
							GLayerBuf.Xfer(false, gpubuf.data, gUWs[i].ptr - InOffsetB + Src * DZ, FullH*DY, pGPU->CLQueue[0]);
							srcptr = gpubuf.data + InOffsetB;
						} else
#endif
							srcptr = (char*)UWs[i].ptr + Src * DZ;

						RW.ptr = srcptr + ((mode&EXPORT_FULL) ? (-(cl_long)InOffsetB) : ReqOffsetB);
						ExportImage(mode, RW, tmpname);
					}
			}
	}
#undef PRINTNAME
	RETURN();
}

int CData::LoadInitial(const char* name)
{
	constexpr int NExts = 4;
	char TryExt[NExts][8] = { "",".rst",".ccf",".cci" };
	int i, ret = 0;
	STR64 tmpname; FILE* fp;

	ENTER;
	T = T_START;
	if ( !name ) name = DirectBaseName;
	if ( !name || !*name ) return 0;

	for ( i = 0; i < NExts; ++i )
	{
		snprintf(tmpname, sizeof(tmpname) - 1, "%s%s", name, TryExt[i]);
		if ( LoadPhantom(tmpname) ) { ++ret;  break; }
	}

	if ( !MSizeA )
	{
		puts("\nStarting with zero data");
		if ( Init(ReqW, ReqH) ) return 0;
	}

	char UWExt[4][4] = { ".u0",".u1",".w0",".w1" };
	cl_ulong UWPtrs[4] = { pU[0], pU[1], pW[0], pW[1] };
	for ( i = 0; i < 4; ++i )
	{
		snprintf(tmpname, sizeof(tmpname) - 1, "%s%s", name, UWExt[i]);
		RectWin RW((char*)UWPtrs[i] + ReqOffsetB, ReqW, ReqH, 32, DY);
		RW.copyfile(fp = fopen(tmpname, "rb"), false);
		if ( fp )
		{
			IFLOG(LOG_FILES) printf("\nLoadInitial '%s'", tmpname);
			fclose(fp);
		}fp = NULL;
	}
//	 SaveInitial("_saved");
	RETURN(ret);
}



void CData::GenerateResTarget(float x, float y) const
{
	Circle((x + 1.0f), (y + 1.0f), -0.5f, 0.2, 0.0f);
	Circle((x + 1.0f), y, 0.5f, -0.05f, 0.0f);
	Circle((x + 2.5f), y, 1.0f, -0.2f, 0.0f);
	Circle((x + 2.5f), (y + 3.0f), -2.0f, 0.05f, 0.0f);
}



void CData::GeneratePhantom(cl_ulong seed) const
{

	Clear(CLR_COEFF);
	if ( !seed ) return;
	if ( seed == 1 ) { GenerateSimplePhantom(); return; }

	ENTER;
	constexpr  int MAX_OBJS = 8;
	float DL = 2;

	xsrng RNG; volatile float vola;
#define rng (vola=RNG)
	if ( seed ) RNG.seed(seed ^ 0x987654321FEDCBA9ULL, seed + (seed << 31) + 0x123456789ABC75DULL);

	float F = 0,
		X0 = PhantomSize * (0.08 * rng - 0.04),
		Y0 = PhantomSize * (0.08 * rng - 0.04),
		R0 = PhantomSize * 0.5* (0.9 + rng * 0.1),
		DC0 = 0.025 * 40 / R0, A0 = 0.04 + rng * 0.03,
		BaseDiam = PhantomSize * 0.25;

	Circle(X0, Y0, R0*1.03, 0.1, 0.01);
	Circle(X0, Y0, R0, -0.02 * PhantomContrast, 0.0008);

	int I, NN;
	for ( NN = 0; NN < MAX_OBJS && BaseDiam > 2; ++NN )
	{
		BaseDiam *= (0.65 + rng * 0.1);
		float r = sqrt(rng)*(R0 - BaseDiam)*0.9, phi = rng * 2 * M_PI,
			x = X0 + r * cos(phi), y = Y0 - r * sin(phi),
			Diam = BaseDiam * (0.85f + rng * 0.3f),
			dc = 1.0f / Diam * (rng - 0.4f),
			ab = 0.11f / Diam * rng;// rng;
		ab = CLAMP(ab, 0, 0.08);
		dc *= (dc < 0) ? 0.3 : 1;
		dc = CLAMP(dc, -0.05, 0.15);
		Circle(x, y, Diam*0.5f, dc*PhantomContrast, ab);
	}; // loop	
	// resolution target
	Circle(1.5, 1, 4.0, -0.02, 0.0);// %
	Circle(0, 0, 1.0, 0.15, 0.03);// %	
	Circle(3, 0, 1.0, 0.15, 0.03);// %
	Circle(3, 2, 0.5, 0.15, 0.03);// %
	RETURN();
}// fn


// Debug phantom
void CData::GenerateSimplePhantom(void) const
{
	Clear(CLR_COEFF);
	float  L = 75 * 0.5, r = L * 0.08;

	Circle(-L * 0.3, -L * 0.5, r, 0.1, 0);
	Circle(-L * 0.3, -L * 0.2, r, 0.05, 0);
	Circle(-L * 0.3, 0, r, -0.05, 0);
	Circle(-L * 0.3, L * 0.3, r, -0.1, 0);

	Circle(L * 0.3, -L * 0.5, r, 0, 0.02);
	Circle(L * 0.3, -L * 0.2, r, 0, 0.04);
	Circle(L * 0.3, 0, r, 0, 0.06);
	Circle(L * 0.3, L * 0.3, r, 0, 0.8);
}

#undef rng

void CData::Circle(float CX, float CY, float R, float dC, float a) const
{ // coordinates in millimeters
	// if (R<0) draw simplex
	// output scaled by contrast
	dC *= ContrastC; a = __max(a, 0.0f)*ContrastA;

	CX = CX / dx + 0.5f*ReqW;	CY = CY / dx + 0.5f*ReqH; 	R /= dx;
	const float soften = 1.25;
	//printf("\ncirc X=%f Y=%f R=%f ", CX, CY, R);

	int X, Y, RCX = (int)round(CX), RCY = (int)round(CY);
	float DCX = CX - RCX, DCY = CY - RCY;

	int Ylim = ceil(fabs(R) + soften + 2);	Ylim = __min(Ylim, ReqH - RCY - 1);

	if ( R > 0 )
		for ( Y = __max(-RCY, (int)floor(-R - soften - 2)); Y <= Ylim; ++Y )
		{
			cl_uint *ptr = (cl_uint*)(pCoeff + ReqOffsetB) + (RCY + Y)*FullW;
			float rr = R * R - Y * Y;
			int DX = ceil(sqrt(__max(rr + 2 + soften, 0)));
			int XLim = (ReqW - RCX - 1); XLim = __min(XLim, DX);

			if ( Y >= -RCY )
				for ( X = __max((-DX), (-RCX)); X <= XLim; ++X )
				{
					float D = (R - sqrt((X - DCX)*(X - DCX) + (Y - DCY)*(Y - DCY))) / soften;
					D = __min(1, D);	D = __max(0, D);
					int xx = RCX + X;
					cl_float2 ca = _UnpackFloatCA(ptr[xx]);
					ca.s[0] = (1 - D)*ca.s[0] + (Vel + dC)*D;
					ca.s[1] = (1 - D)*ca.s[1] + a * D;
					ptr[xx] = _PackFloatCA(ca.s[0], ca.s[1]);
				} // for X
		}// for Y
	else
		for ( Y = __max(-RCY, (int)floor(R - soften - 2)); Y <= Ylim; ++Y )
		{
			cl_uint *ptr = (cl_uint*)(pCoeff + ReqOffsetB) + (RCY + Y)*FullW;
			float rr = -R - fabs(Y);
			int DX = ceil(__max(rr + 2 + soften, 0));
			int XLim = (ReqW - RCX - 1); XLim = __min(XLim, DX);

			if ( Y >= -RCY )
				for ( X = __max((-DX), (-RCX)); X <= XLim; ++X )
				{
					float D = (-R - fabs(X - DCX) - fabs((Y - DCY))) / soften;
					D = __min(1, D);	D = __max(0, D);
					int xx = RCX + X;
					cl_float2 ca = _UnpackFloatCA(ptr[xx]);
					ca.s[0] = (1 - D)*ca.s[0] + (Vel + dC)*D;
					ca.s[1] = (1 - D)*ca.s[1] + a * D;
					ptr[xx] = _PackFloatCA(ca.s[0], ca.s[1]);
				} // for X
		}// for Y	
}//fn


void CData::Interactive(void)
{
	if ( !fDemoMode || *InverseBaseName || *DirectBaseName || PerfTest || PerfTestParam ) return;
	STR64 tmps; cl_ulong tmpl;
	puts("\n *** Interactive demo ***\n Image size (480/640/800) ? ");
	scanf("%63s", tmps); ReqW = atoi(tmps);
	ReqW = CLAMP(ReqW, 480, 800); ReqW = align(ReqW, 160);
	printf(" -size %i selected\nNumber of emitters (10 - 20) ? ", ReqW);
	scanf("%63s", tmps); GlobalSrcs = atoi(tmps);
	NSrcs = CLAMP(GlobalSrcs, 10, 20);
	printf("   %i selected\n Phantom number (any) ? ", NSrcs);
	scanf("%63s", tmps);
	tmpl = atol(tmps);
	if ( tmpl < 2 ) tmpl = 2;
	snprintf(DirectBaseName, sizeof(DirectBaseName), "#%llu", tmpl);
	snprintf(InverseBaseName, sizeof(InverseBaseName), "#%llu", tmpl);
	printf(" %s selected", DirectBaseName);
#ifdef TARGET_GPU
	printf("\n Use GPU? Y/N "); scanf("%1s", tmps);
	if ( (*tmps == 'y') || (*tmps == 'Y') )
	{
		EnableGPU = 1; G_Settings.GPU_device = 0;
		puts("\n GPU selected");
	} else
	{
		puts("\n CPU selected");
		EnableGPU = 0; G_Settings.GPU_device = -1;
	}
#endif
}