/*
WaveTomography software implements the algorithm for solving direct and inverse problems of wave tomography in application to medical ultrasound imaging for breast cancer diagnosis.
The software has been developed at Lomonosov Moscow State University. Copyright (c) 2021 Seryozhnikov S.Y, Romanov S.Y.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "setup.h"

#define _USE_MATH_DEFINES 
#include <math.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#ifndef TARGET_WINDOWS
#include <unistd.h>
#include <sys/time.h>
#include <sys/mman.h>
#else
#include <io.h>
#endif // TARGET_WINDOWS


#include "k_defs.h"
#include "cmdl.h"


char _PID[32] = "\n>";
int MyId = 0, NumProcesses = 1,
_STACKCOUNT = 2;

int  OPTDEF::T_SIZE[OPTDEF::MAXTYPE] = { sizeof(int),sizeof(float),sizeof(double),64,0 };
const char*  OPTDEF::T_FMT[OPTDEF::MAXTYPE] = { "%i","%f","%lg","%63s","" };


#ifndef TARGET_WINDOWS

const char* ErrName(cl_int n)
{
	static const char* errnames[122] = { " 0:OK ",
 " EPERM 1 Operation not permitted ",
 " ENOENT 2 No such file or directory ",
 " ESRCH 3 No such process ",
 " EINTR 4 Interrupted system call ",
 " EIO 5 I/O error ",
 " ENXIO 6 No such device or address ",
 " E2BIG 7 Arg list too long ",
 " ENOEXEC 8 Exec format error ",
 " EBADF 9 Bad file number ",
 " ECHILD 10 No child processes ",
 " EAGAIN 11 Try again ",
 " ENOMEM 12 Out of memory ",
 " EACCES 13 Permission denied ",
 " EFAULT 14 Bad address ",
 " ENOTBLK 15 Block device required ",
 " EBUSY 16 Device or resource busy ",
 " EEXIST 17 File exists ",
 " EXDEV 18 Cross-device link ",
 " ENODEV 19 No such device ",
 " ENOTDIR 20 Not a directory ",
 " EISDIR 21 Is a directory ",
 " EINVAL 22 Invalid argument ",
 " ENFILE 23 File table overflow ",
 " EMFILE 24 Too many open files ",
 " ENOTTY 25 Not a typewriter ",
 " ETXTBSY 26 Text file busy ",
 " EFBIG 27 File too large ",
 " ENOSPC 28 No space left on device ",
 " ESPIPE 29 Illegal seek ",
 " EROFS 30 Read-only file system ",
 " EMLINK 31 Too many links ",
 " EPIPE 32 Broken pipe ",
 " EDOM 33 Math argument out of domain of func ",
 " ERANGE 34 Math result not representable ",
 " EDEADLK 35 Resource deadlock would occur ",
 " ENAMETOOLONG 36 File name too long ",
 " ENOLCK 37 No record locks available ",
 " ENOSYS 38 Function not implemented ",
 " ENOTEMPTY 39 Directory not empty ",
 " ELOOP 40 Too many symbolic links encountered ",
 " EWOULDBLOCK 41 Operation would block ",
 " ENOMSG 42 No message of desired type ",
 " EIDRM 43 Identifier removed ",
 " ECHRNG 44 Channel number out of range ",
 " EL2NSYNC 45 Level 2 not synchronized ",
 " EL3HLT 46 Level 3 halted ",
 " EL3RST 47 Level 3 reset ",
 " ELNRNG 48 Link number out of range ",
 " EUNATCH 49 Protocol driver not attached ",
 " ENOCSI 50 No CSI structure available ",
 " EL2HLT 51 Level 2 halted ",
 " EBADE 52 Invalid exchange ",
 " EBADR 53 Invalid request descriptor ",
 " EXFULL 54 Exchange full ",
 " ENOANO 55 No anode ",
 " EBADRQC 56 Invalid request code ",
 " EBADSLT 57 Invalid slot ",
 " EDEADLOCK 58 File locking deadlock error ",
 " EBFONT 59 Bad font file format ",
 " ENOSTR 60 Device not a stream ",
 " ENODATA 61 No data available ",
 " ETIME 62 Timer expired ",
 " ENOSR 63 Out of streams resources ",
 " ENONET 64 Machine is not on the network ",
 " ENOPKG 65 Package not installed ",
 " EREMOTE 66 Object is remote ",
 " ENOLINK 67 Link has been severed ",
 " EADV 68 Advertise error ",
 " ESRMNT 69 Srmount error ",
 " ECOMM 70 Communication error on send ",
 " EPROTO 71 Protocol error ",
 " EMULTIHOP 72 Multihop attempted ",
 " EDOTDOT 73 RFS specific error ",
 " EBADMSG 74 Not a data message ",
 " EOVERFLOW 75 Value too large for defined data type ",
 " ENOTUNIQ 76 Name not unique on network ",
 " EBADFD 77 File descriptor in bad state ",
 " EREMCHG 78 Remote address changed ",
 " ELIBACC 79 Can not access a needed shared library ",
 " ELIBBAD 80 Accessing a corrupted shared library ",
 " ELIBSCN 81 .lib section in a.out corrupted ",
 " ELIBMAX 82 Attempting to link in too many shared libraries ",
 " ELIBEXEC 83 Cannot exec a shared library directly ",
 " EILSEQ 84 Illegal byte sequence ",
 " ERESTART 85 Interrupted system call should be restarted ",
 " ESTRPIPE 86 Streams pipe error ",
 " EUSERS 87 Too many users ",
 " ENOTSOCK 88 Socket operation on non-socket ",
 " EDESTADDRREQ 89 Destination address required ",
 " EMSGSIZE 90 Message too long ",
 " EPROTOTYPE 91 Protocol wrong type for socket ",
 " ENOPROTOOPT 92 Protocol not available ",
 " EPROTONOSUPPORT 93 Protocol not supported ",
 " ESOCKTNOSUPPORT 94 Socket type not supported ",
 " EOPNOTSUPP 95 Operation not supported on transport endpoint ",
 " EPFNOSUPPORT 96 Protocol family not supported ",
 " EAFNOSUPPORT 97 Address family not supported by protocol ",
 " EADDRINUSE 98 Address already in use ",
 " EADDRNOTAVAIL 99 Cannot assign requested address ",
 " ENETDOWN 100 Network is down ",
 " ENETUNREACH 101 Network is unreachable ",
 " ENETRESET 102 Network dropped connection because of reset ",
 " ECONNABORTED 103 Software caused connection abort ",
 " ECONNRESET 104 Connection reset by peer ",
 " ENOBUFS 105 No buffer space available ",
 " EISCONN 106 Transport endpoint is already connected ",
 " ENOTCONN 107 Transport endpoint is not connected ",
 " ESHUTDOWN 108 Cannot send after transport endpoint shutdown ",
 " ETOOMANYREFS 109 Too many references: cannot splice ",
 " ETIMEDOUT 110 Connection timed out ",
 " ECONNREFUSED 111 Connection refused ",
 " EHOSTDOWN 112 Host is down ",
 " EHOSTUNREACH 113 No route to host ",
 " EALREADY 114 Operation already in progress ",
 " EINPROGRESS 115 Operation now in progress ",
 " ESTALE 116 Stale NFS file handle ",
 " EUCLEAN 117 Structure needs cleaning ",
 " ENOTNAM 118 Not a XENIX named type file ",
 " ENAVAIL 119 No XENIX semaphores available ",
 " EISNAM 120 Is a named type file ",
 " EREMOTEIO 121 Remote I/O error " };
	static char tmp[72];
	if ( n < 0 || n>121 ) { sprintf(tmp, " ??errno=%i ", n); return tmp; } else return errnames[n];
}//fn


#else

void sleep(int sec) { Sleep(sec * 1000); }

const char* ErrName(cl_int n)
{
	static char tmpb[1024];
	LPSTR messageBuffer = tmpb + 6;
	sprintf(tmpb, "%04i: ", n);
	size_t size = FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
															 NULL, n, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
															 (LPSTR)messageBuffer, 1016, NULL);
	return tmpb;
}
#endif

cl_uint2 SelectPart(cl_uint N, cl_int pid)
{
	if ( pid < 0 ) pid = MyId;
	cl_uint2 F = { 0,0 };
	if ( !N || !NumProcesses || ( pid >= NumProcesses )) return F;		
	F.s[0] = floor(N*pid / NumProcesses);
	F.s[1]= floor(N*(pid + 1) / NumProcesses);
	return F;
}

char* checked_malloc(size_t s, const char* name)
{
	size_t len;
	char* data = (char*)malloc((len = align(s, 4)) + 32);
	if ( !data )
	{
		printf("%s malloc %s fail [%llXh]", _PID, name ? name : "", (cl_ulong)len);
		return NULL;
	}
	register cl_uint *pI = (cl_uint*)data, *pJ = (cl_uint*)(data + len + 16);
	pI[0] = pI[1] = pI[2] = pI[3] = pJ[0] = pJ[1] = pJ[2] = pJ[3] = _MCHK_;
	data += 16;
	memset(data, 0, len);
	if ( G_Settings.LogLevel&LOG_MALLOC ) printf("%s malloc %s @%llXh [%llXh]", _PID, name ? name : "", (cl_ulong)data, (cl_ulong)s);
	return data;
}//fn

bool checked_free(void* data, size_t len, const char* name)
{
	bool retval = true; char cc = '.';
	if ( !data ) return retval;
	register cl_uint *pI = (cl_uint*)(((char*)data) - 16),
		*pJ = (cl_uint*)(((char*)data) + align(len, 4));
	retval = ((pI[0] == _MCHK_) && (pI[1] == _MCHK_) && (pI[2] == _MCHK_) && (pI[3] == _MCHK_));
	if ( !retval ) cc = 'B';
	else if ( len )
	{
		retval = ((pJ[0] == _MCHK_) && (pJ[1] == _MCHK_) && (pJ[2] == _MCHK_) && (pJ[3] == _MCHK_));
		if ( !retval ) cc = 'E';
	}
	free(((char*)data) - 16);
	if ( !retval || (G_Settings.LogLevel&LOG_MALLOC) ) printf("%s   free %s @%llXh [%llXh]%c ", _PID, name ? name : "", (cl_long)data, (cl_long)len, cc);
	return retval;
}// fn



void filemap::print(const char* prefix)
{
	printf("%s filemap: len=%lli err=%s fd=%llX mapd=%llX ",
				 prefix ? prefix : "\n", len, ErrName(error), (long long int)fd, (long long int)mapd);
}

int filemap::init(HANDLE _fd, int myid) // file-access mode
{
	access = NULL; len = 0; log = 0; error = 0; fd = _fd; mapd = INVALID_HANDLE_VALUE;
	len = LSEEK(fd, 0, SEEK_END);
	LSEEK(fd, 0, SEEK_SET);
	return 0;
}


int filemap::init(const char* name, cl_long l, int myid) // name, len -> file mmap;
{                                    // NULL, 0   - no-op
	char tmp[64]; error = 0;           // name, 0   - auto-length file mmap
	const char* errmsg = "open";       // name, -1  - attempt file alloc-read, fallback to mmap, fallback to file access        
	close(); len = l;                  // name, -2  - file access only (old)
	int retval = 0;                    // NULL, len - attempt memory buffer, fallback to mmap "tmp_myid", fallback to file access         
#ifdef TARGET_WINDOWS                // NULL, <=0 - tmpfile access (old)
	const int shft = (sizeof(char)*(sizeof(cl_ulong) - sizeof(long)));
	unsigned long ll = (l >> shft);
	LARGE_INTEGER ll0;  ll0.QuadPart = 0LL;
#endif
	if ( !name )
	{
		fd = mapd = INVALID_HANDLE_VALUE;
		if ( !name && !len ) return 0; // no-op, successful
		if ( len > 0 ) access = checked_malloc(len, (char*)"fmap"); else len = -2LL;
		if ( access )
		{
			IFLOG(LOG_FMAP) { printf((char*)"\nfilemap: alloc [%lli] @%llXh", len, (cl_ulong)access); }
			retval = 1;  goto finish;
		} else
		{
			name = tmp; sprintf(tmp, (char*)"tmpb_%03i.dat", myid);
			if ( log ) printf((char*)"\nfilemap: tmpfile '%s' len=%lli", tmp, len);
		}
	}

#ifdef TARGET_WINDOWS
	fd = CreateFileA(name, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if ( fd == INVALID_HANDLE_VALUE ) { error = GetLastError();  goto finish; }
	SetFilePointerEx(fd, ll0, (PLARGE_INTEGER)&len, FILE_END); // sets len to filesize

	if ( l > 0 )
	{
		if ( len >= l ) retval = 1;
		else { SetFilePointer(fd, len = l, (PLONG)&ll, FILE_BEGIN); SetEndOfFile(fd); }
	}

	SetFilePointer(fd, 0, NULL, FILE_BEGIN);
	// .len set and file checked 
	if ( l == -1LL )
	{
		access = checked_malloc(len, (char*)"fmap");
		if ( access )
		{
			ReadFile(fd, access, len, &ll, NULL);
			if ( log ) printf((char*)"%s filemap alloc-read '%s' [%lli] @%llXh", _PID, name, len, (cl_ulong)access);
		} else l = len; // if alloc fails try mmap;
	}
#ifndef DISABLE_FMAP
	if ( l >= 0 )
	{
		errmsg = "CreateFileMapping";
		mapd = CreateFileMapping(fd, NULL, PAGE_READWRITE, len >> 32, len & 0xFFFFFFFF, NULL);
		if ( mapd == INVALID_HANDLE_VALUE ) { error = GetLastError();  goto finish; }
		errmsg = "MapViewOfFile";
		access = (char*)MapViewOfFile(mapd, FILE_MAP_ALL_ACCESS, 0, 0, len);
		if ( !access ) { error = GetLastError(); goto finish; }
		if ( log ) printf("%s filemap mapped fd=%lli '%s' [%lli] @%llXh", _PID, (cl_long)fd, name, len, (cl_ulong)access);
		if ( len >= l ) retval = 1;
	}
#endif // DISABLE_FMAP
#else // -- target linux --
	fd = open(name, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if ( fd == INVALID_HANDLE_VALUE ) { error = errno; goto finish; }
	errmsg = "ftruncate";
	len = lseek(fd, 0, SEEK_END);
	if ( l > 0 )
	{
		if ( len >= l ) retval = 1;
		else if ( posix_fallocate(fd, 0, len = l) ) { error = errno;  goto finish; }
	}
	lseek(fd, 0, SEEK_SET);
	// .len set and file checked 
	if ( l == -1LL )
	{
		access = checked_malloc(len, "fmap");
		if ( access ) { read(fd, access, len); if ( log ) printf("%s filemap alloc-read %s [%llu]", _PID, name, len); } else l = len;
	}
#ifndef DISABLE_FMAP
	if ( l >= 0 )
	{
		errmsg = "mmap";
		access = (char*)mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
		if ( access == MAP_FAILED ) { error = errno; access = NULL;  goto finish; }
		mapd = ~INVALID_HANDLE_VALUE;
		if ( log ) printf("%s filemap mapped fd=%i '%s' [%lli] @%llXh", _PID, (int)fd, name, len, (cl_long)access);
		if ( len >= l ) retval = 1;
	}
#endif // DISABLE_FMAP
#endif // target windows
finish:
	if ( error ) printf("%s filemap: %s '%s' :%s", _PID, errmsg, name ? name : "(NULL)", ErrName(error));
	return retval;
}//fn

void filemap::close(void)
{
	if ( mapd != INVALID_HANDLE_VALUE )
	{
#ifdef TARGET_WINDOWS
		if ( access ) UnmapViewOfFile(access);
		CloseHandle(mapd);
		if ( log )
			printf("%s filemap unmap %lli @%llX", _PID, (long long)mapd, (long long)access);
#else
		if ( access ) munmap(access, len);
#endif
	} else if ( access )
	{
		checked_free(access, len, (char*)"fmap");
		if ( log ) printf("%s filemap free buffer @%llX", _PID, (long long)access);
	}

	if ( fd != INVALID_HANDLE_VALUE )
	{
		CloseHandle(fd);
		if ( log ) printf("%s filemap close fd=%llX", _PID, (long long)fd);
	}

	fd = mapd = INVALID_HANDLE_VALUE; len = 0; access = NULL;
}//fn


FileMapArray::FileMapArray() 
{ for ( unsigned int i = 0; i < MAX_FILES; ++i ) Fm[i].init(); StartNumber = 0; }

int FileMapArray::init(unsigned int first, unsigned int l, cl_long sz, const char* name, unsigned int startn)
{
	if ( !name ) name = "";
	int retval = 0;
	if ( strlen(name) > sizeof(ActiveName)-5 )
	{ printf("%s Filemap: name too long '%s'", _PID, name); return 0; }
	strncpy(ActiveName, name, sizeof(ActiveName));
	StartNumber = startn;

	for ( unsigned int i = first; (i < (first + l)) && (i < MAX_FILES); ++i )
		if ( *name )
		{
			STR64 tmpname;
			snprintf(tmpname, sizeof(tmpname), SRC_FILENAME, ActiveName, StartNumber+i);
			retval += Fm[i].init(tmpname, sz, MyId);
		} else retval += Fm[i].init(NULL, sz, MyId);

		return(retval == l);
}



int FileMapArray::assert(unsigned int first, unsigned int nsrcs, cl_long sz, size_t MinReq, const char* name, unsigned int startn)
{ // check that all files have at least MinReq length
	CHECKPOINT("FileMapArray::assert ");
	int cnt = 0;
	if ( !name ) name = ActiveName;

	if ( strcmp(name, ActiveName) )
	{
		close();
		for ( int i = first; i < (first + nsrcs); ++i )
		{
			STR64 tmpname; snprintf(tmpname, sizeof(tmpname), SRC_FILENAME, name, i);
			cl_long fl = CheckFile(tmpname);
			if ( fl >= MinReq ) ++cnt;
		}
		if ( nsrcs != cnt ) return 0;
		return init(first, nsrcs, sz, name,startn);
	}

	return (NumOpen(first, nsrcs, MinReq) == nsrcs);
}

void FileMapArray::close(int I)
{
	if ( I >= 0 )
	{
		if ( I < MAX_FILES ) Fm[I].close();
	} else
		for ( int i = 0; i < MAX_FILES; ++i )	Fm[i].close();
	ActiveName[0] = '\0';
}

void FileMapArray::commit(void)
{
	STR64 tmpname;
	for ( unsigned int i = 0; i < MAX_FILES; ++i ) if ( Fm[i].access )
	{
		sprintf(tmpname, SRC_FILENAME, ActiveName, i); 
		FWrite(tmpname, Fm[i].access, Fm[i].len);
	}
}

int FileMapArray::NumOpen(unsigned int fst, unsigned int l, cl_long sz)
{
	int rv = 0;
	if ( (fst + l) > MAX_FILES ) return 0;
	for ( unsigned int i = fst; (i < (fst + l)); ++i )
		if ( (Fm[i].access) && (Fm[i].len >= sz) ) ++rv;
	return rv;
}

size_t  memblk::ctr = 0;

void memblk::WriteFile(const char* filename, size_t lim) { strncpy(name, filename, NAMELEN - 1); FWrite(filename, data, lim < len ? lim : len); }

size_t memblk::ReadFile(const char* filename) { free();  strncpy(name, filename, NAMELEN - 1); len = 0; data = FRead(filename, &len); return len; }

char* memblk::alloc(size_t s, const char* nm)
{
	if ( nm ) strncpy(name, nm, sizeof(name) - 1);
	if ( len == s ) return data;
	data = (char*)checked_malloc(s, name);
	len = data ? s : 0;
	if ( !data )	printf("%s memblk::alloc fail %s (%lli) from @%llX ", _PID, name ? name : "", s, (&s)[-2]);
	return data;
}

bool memblk::free(void)
{
	if ( !data ) { len = 0; return true; }
	checked_free(data);
	data = NULL; len = 0; return true;
}


void OPTDEF::Print(const char* base, FILE* fp) const
{
	const char* dest; int i;
	for ( const OPTDEF* p = this; p->type >= 0; p++ )
	{
		fprintf(fp, "\n%s  ", p->name);
		dest = base + p->offs;
		for ( i = 0; i < p->rep; i++ )
		{
			int t = p->type;
			switch ( t )
			{
				case T_INT:    fprintf(fp, OPTDEF::T_FMT[t], *(int*)dest);   break;
				case T_DOUBLE: fprintf(fp, OPTDEF::T_FMT[t], (float)(*(double*)dest));   break;
				case T_FLOAT: fprintf(fp, OPTDEF::T_FMT[t], *(float*)dest);  break;
				case T_STR64: fprintf(fp, OPTDEF::T_FMT[t], dest);  break;
			}
			fprintf(fp, " "); dest += OPTDEF::T_SIZE[t];
		}
	}
}

/*
void OPTDEF::Print(const char* base, char* Str) const
{
	if ( Str ) *Str = 0; else return;
	const char* dest; int i;
	for ( const OPTDEF* p = this; p->type >= 0; p++ )
	{
		sprintf(Str + strlen(Str), "\n%s  ", p->name);
		dest = base + p->offs;
		for ( i = 0; i < p->rep; i++ )
		{
			int t = p->type;
			switch ( t )
			{
				case T_INT:    sprintf(Str + strlen(Str), OPTDEF::T_FMT[t], *(int*)dest);  break;
				case T_DOUBLE: sprintf(Str + strlen(Str), OPTDEF::T_FMT[t], (float)(*(double*)dest));  break;
				case T_FLOAT:  sprintf(Str + strlen(Str), OPTDEF::T_FMT[t], *(float*)dest);  break;
				case T_STR64:  sprintf(Str + strlen(Str), OPTDEF::T_FMT[t], dest);  break;
			}
			sprintf(Str + strlen(Str), "%c", ' ');  dest += OPTDEF::T_SIZE[t];
		}
	}
}
*/

int OPTDEF::ParseFile(const char* name, const char* base) const
{
	CHECKPOINT("OPTDEF::ParseFile ");
	STR64 tmp;  const OPTDEF* p;
	size_t pt = 0;
	const char *dest;
	size_t len=0;
	char* text = FRead(name, &len, true);
	int i, a, retval = 0;
	if ( !text ) goto finish;

	do
	{
		if ( text[pt] != '-' ) { pt++; goto end; }

		for ( p = this; p->type >= 0; p++ )
			if ( !strncmp(text + pt, p->name, strlen(p->name)) ) goto found;
		pt++; goto end;

	found: //printf("\n INI item %s typ%i[%i] p=%03lX ",p->name,p->type,p->rep,pt);
		pt += strlen(p->name);
		if ( p->type > OPTDEF::MAXTYPE ) goto end;
		//printf(" p2=%03lX ",pt);
		// t -> "-name a,b,c"
		while ( (pt < len) && IsSpace(text[pt]) )  pt++;   if ( pt >= len ) goto finish;
		retval++;
		//printf(" p3=%03lX ",pt);

		for ( i = 0, dest = base + p->offs; i < p->rep; i++, dest += OPTDEF::T_SIZE[p->type] )
		{
			if ( sscanf(text + pt, "%63s", tmp) != 1 ) { printf("\n#Warning: incomplete '%s' ", p->name);  goto end; }
			pt += (strlen(tmp));

			a = sscanf(tmp, OPTDEF::T_FMT[p->type], dest);
			// printf("\n FIELD [%i] \"%s\" SCANNED %s %i TO %16lX p=%03lX ", strlen(tmp),tmp,OPTDEF::T_FMT[p->type],a, dest,pt);
			if ( a != 1 ) goto end;

			while ( IsSpace(text[pt]) && (pt < len) ) pt++;
			if ( (pt >= len) || !(text[pt]) ) goto finish;
		}
	end:;//prev=text[pt++];
	} while ( text[pt] && (pt < len) );
finish:
	checked_free(text, len);
	return retval;
}

int OPTDEF::ParseArgv(int argc, char** argv, const char* base, int argc_pos) const
{
	const OPTDEF* p; const char* dest;
	int i, arg = argc_pos;
	//printf("ParseA argc=%i", argc);
	while ( arg < argc )
	{
		if ( argv[arg][0] == '+' ) return arg;

		for ( p = this; p->type >= 0; p++ )
			if ( !strcmp(argv[arg], p->name) )
			{ //printf("\nOPT:  %i: '%s'  ", arg, argv[arg]); 
				goto found;
			}

		++arg; continue;

	found:  if ( ++arg >= argc ) break;
		if ( p->type > OPTDEF::MAXTYPE ) break;

//		printf("\nPARAM:  %i: '%s'  ", arg, argv[arg]);

		for ( i = 0, dest = base + p->offs; i < p->rep; i++, dest += OPTDEF::T_SIZE[p->type] )
		{
			if ( sscanf(argv[arg], OPTDEF::T_FMT[p->type], dest) != 1 ) { printf("\n Invalid parameter %i: '%s'  ", arg, argv[arg]); break; }
			if ( ++arg >= argc ) break;
		}
	}
	return arg;
}




int copyname(char* dest, const char* src, int nmax)
{
	char c = *src++; int i = 1;
	*dest++ = (isalnum(c)) ? c : 'x';
	while ( (c = *src++) && (++i <= nmax) ) *dest++ = isalnum(c) ? c : '_';
	if ( i < nmax ) *dest++ = 0;
	return i;
}


cl_long CheckFile(const char* fname)
{
	cl_long ret;
	FILE* fp = fopen(fname, "r");
	if ( !fp ) return -1;
	fseek(fp, 0, SEEK_END); ret = ftell(fp); fseek(fp, 0, SEEK_SET);
	fclose(fp);
	return ret;
}


char* FRead(const char* name, size_t* pLen, bool zero_pad)
{
	size_t len = pLen ? *pLen : 0, FullLen = len;
	if ( !pLen ) pLen = &len;

	FILE*fp = fopen(name, "rb");

	if ( !fp ) { printf("%s No file '%s' ", _PID, name); *pLen = 0;  return NULL; }

	if ( !len )
	{
		fseek(fp, 0, SEEK_END);
		len = ftell(fp);
		fseek(fp, 0, SEEK_SET);
	}

	char* r = (char*)checked_malloc(FullLen = len + (zero_pad ? 1 : 0));
	if ( !r ) { fclose(fp); printf("%s No memory for file '%s' %llu ", _PID, name, len); return NULL; }
	len = fread(r, 1, len, fp); fclose(fp); *pLen = FullLen;
	if ( G_Settings.LogLevel&LOG_FILES ) printf("\n*'%s' loaded: %lli* ", name, len);
	return r;
}


void FWrite(const char* name, void* r, size_t len, bool append)
{
	if ( !r ) { puts(" FWrite: NULL data ");  return; }
	FILE*fp = fopen(name, append ? "a+b" : "w+b");
	if ( !fp ) { fprintf(stderr, "%s Cannot write file '%s' ", _PID, name); return; }
	len = fwrite(r, 1, len, fp); fclose(fp);
	if ( G_Settings.LogLevel&LOG_FILES )
		printf("%s *'%s' saved %lli* ", _PID, name, len);
}


size_t LOAD(const char* name, void* buf, size_t size, int time)
{
	char nm[64];
	size_t n;
	if ( time >= 0 ) sprintf(nm, "%s%03i.dmp", name, time); else sprintf(nm, name);
	FILE* fp = fopen(nm, "rb");
	if ( !fp ) { if ( G_Settings.LogLevel&LOG_FILES ) printf("%s *Error opening file %s* ", _PID, nm); return 0; } else
	{
		n = fread(buf, 1, size, fp); fclose(fp);
		if ( G_Settings.LogLevel&LOG_FILES )	printf("%s *Loaded %s %lli* ", _PID, nm, n);
		return n;
	}
}

void DUMP(const char* name, void* buf, size_t size, int time)
{
	char nm[64]; size_t l;

	//  printf("\nDUMP %s, %lX, %lu, %i:  ",name,buf,size,time);

	if ( time >= 0 ) sprintf(nm, "%s%03i.dmp", name, time); else sprintf(nm, name);
	FILE* fp = fopen(nm, "w+b");
	//  printf(" fopen. ");
	if ( !fp )
	{
		if ( G_Settings.LogLevel&LOG_FILES ) printf("%s *Error creating file %s* ", _PID, nm);
	} else
	{
		l = fwrite(buf, 1, size, fp);
		//         printf(" fwrite. ");
		fclose(fp);
		if ( G_Settings.LogLevel&LOG_FILES ) printf("%s *Saved %s %lli* ", _PID, nm, l);
	}
};

void DumpRect(const char* name, void* buf, size_t height, size_t width, size_t stride)
{
	size_t l = 0;
	FILE* fp = fopen(name, "w+b");
	if ( !fp )
	{
		if ( G_Settings.LogLevel&LOG_FILES ) printf("%s *Error creating file %s* ", _PID, name);
		return;
	}
	for ( int i = 0; i < height; ++i )
		l += fwrite(((char*)buf) + i * stride, 1, width, fp);
	fclose(fp);
	if ( G_Settings.LogLevel&LOG_FILES ) printf("%s *Saved %s %lli* ", _PID, name, l);
};


size_t _cdecl ReturnRIP(size_t s)
{
	return  (&s)[-1];
}


#define _isInfPtr(p) 0
//( (~(*(cl_uint*)(p))) & 0xEF800000)

float diffstats(const float* F, const float* G, size_t count, float* dest)
{
	float SS = 0, Sum = 0, SumA = 0, Max = 0;
	if ( !F || !G )return -1;

	for ( int i = 0; i < count; ++i )
	{
		float D = F[i] - G[i];
		SS += D * D;
		Sum += D;
		SumA += (D = fabs(D));
		Max = __max(Max, D);
	}
	if ( dest )
	{
		dest[0] = sqrt(SS); dest[1] = Max; dest[2] = SumA; dest[3] = Sum;
	}
	return Max;
}

float maxupF(const float*  p, size_t len, int stride, float mm, size_t* retIdx, float* retMean)
{
	size_t Idx = 0;
	float  a = mm, b = mm, sum = 0;
	size_t l = len;
	while ( l-- )
	{
		if ( !_isInfPtr(p) ) b = fabs(*p);
		p += stride; sum += b;
		if ( a < b ) { a = b; Idx = len - l - 1; }
	}
	if ( retIdx ) *retIdx = Idx;
	if ( retMean ) *retMean = sum / len;
	return(a);
};


void DumpFloatImagePBM(const char* fname, const RectWin& RW, float lim)
{
	int I, J;
	if ( (!RW.ptr) || (RW.BPPixel != 32) ) return;
	FILE* fp = fopen(fname, "wb"); if ( !fp ) return;
	memblk tmp(RW.LWidth * 2);	if ( !tmp.data ) return;

	if ( lim <= 0.0f )
		for ( J = 0; J < RW.Height; ++J )
		{
			float l2 = maxupF((float*)RW.line(J), RW.LWidth);
			lim = __max(lim, l2);
		}

	float maxval = 32768.0f / lim;

	fprintf(fp, "P5\n# scale=%f \n%i %i\n65535 ", lim, RW.LWidth, RW.Height); fflush(fp);

	for ( J = RW.Height; J; --J )
	{
		float* p = (float*)RW.line(J);
		char* dest = tmp.data;
		for ( I = 0; I < RW.LWidth; ++I )
		{
			float V = round(*p++ * maxval) + 32768.0f;
			V = CLAMP(V, 0, 65535);
			const int VI = (int)V;
			dest[0] = (VI >> 8);		dest[1] = (VI & 0xFF); dest += 2;
		}
		fwrite(tmp.data, 2, RW.LWidth, fp);
	}
	fclose(fp);
}



size_t memcmp_ushort_tol(const void* p, const void* q, size_t len, int tol)
{
	if ( (!p) || (!q) ) return -1LL;
	len >>= 1;
	for ( size_t i = 0; i < len; ++i )
	{
		int diff = (int)(((unsigned short*)p)[i]) - (int)(((unsigned short*)q)[i]);
		if ( (diff < -tol) || (diff > tol) ) return i + 1;
	}
	return 0;
}

int diff_ushort(const void* p, const void* q, size_t len)
{
	int tol = 0;
	if ( (!p) || (!q) ) return -1LL;
	len >>= 1;
	for ( size_t i = 0; i < len; ++i )
	{
		int diff = (int)(((unsigned short*)p)[i]) - (int)(((unsigned short*)q)[i]);
		diff = (diff < 0) ? -diff : diff;
		tol = (tol > diff) ? tol : diff;
	}
	return tol;
}

size_t memcmp_float_tol(const float* p, const float* q, size_t len, float tol)
{
	if ( (!p) || (!q) ) return -1LL;
	len >>= 2;
	for ( size_t i = 0; i < len; ++i )	if ( fabs(p[i] - q[i]) > tol ) return i + 1;
	return 0;
}



const int MATHeader::ClassForType[16] = { 0,8,9,10,11,12,13, 7,0,6,0,0, 14,15,0,0 };
const int MATHeader::LenForType[16] = { 0,1,1,2,2,4,4, 4,0,8,0,0, 8,8,0,0 };

void MATHeader::Create(const int* pDim, int Type, const char* name)
{
	sub1 = sub2 = _pad = 0; MatrixHead = miMATRIX;
	aflagsHead = miUINT32; aflagsLen = 8;
	version = 0x100; endian = 'MI';
	aflags2 = 0; 	aflags1 = ClassForType[Type];
	DimHead = miINT32; DimLen = 16;
	NameHead = miINT8;

	DataType = Type &= 0x0F;
	Dim[0] = Dim[1] = Dim[2] = Dim[3] = 1;
	if ( pDim ) 	for ( int n = 0; (n < 4) && (pDim[n] > 0); ++n ) Dim[n] = pDim[n];
	DataLen = Dim[0] * Dim[1] * Dim[2] * Dim[3] * LenForType[Type];

	ActNameLen = align(NameLen = copyname(Name, name, sizeof(Name)), 8);
	HeadSize = offsetof(MATHeader, Name) + ActNameLen;
	MatrixLen = HeadSize + 8 + DataLen - offsetof(MATHeader, aflagsHead);

	memset(desc, ' ', sizeof(desc));
	snprintf(desc, sizeof(desc) - 1, "MATLAB 5.0 MAT - file, %s[%i,%i,%i,%i] '%s' @0x%X [0x%X].",
					 NameForType[Type], Dim[0], Dim[1], Dim[2], Dim[3], name, HeadSize + 8, DataLen);
}


FILE*  MATHeader::WriteFile(void* data, const char* fname) const
{
	STR64 localfname;
	if ( !fname ) { snprintf(localfname, 63, "%s.mat", Name); fname = localfname; }
	size_t wr = 0;
	IFLOGFILE printf("\nWriting MAT-file '%s' header '%s'", fname, desc);
	FILE* fp = fopen(fname, "wb");
	if ( !fp ) { IFLOGFILE puts("Cannot create file");  return NULL; }
	wr += fwrite(this, HeadSize, 1, fp);
	wr += fwrite(&DataType, 4, 2, fp);
	if ( !data ) { IFLOGFILE printf(" => %lli bytes header. ", (cl_ulong)wr); return fp; }
	wr += fwrite(data, DataLen, 1, fp);
	wr += fwrite(&_pad, align(DataLen, 8) - DataLen, 1, fp);
	fclose(fp);
	IFLOGFILE printf(" => %lli bytes data. ", (cl_ulong)wr);
	return NULL;
}

const char MATHeader::NameForType[16][8] = { "?0?","INT8","UINT8","INT16","UINT16",
	"INT32","UINT32", "SINGLE","?8?","DOUBLE","?10?","?11?","INT64","UINT64","?14?","?15?" };


