/*
WaveTomography software implements the algorithm for solving direct and inverse problems of wave tomography in application to medical ultrasound imaging for breast cancer diagnosis.
The software has been developed at Lomonosov Moscow State University. Copyright (c) 2021 Seryozhnikov S.Y, Romanov S.Y.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "k_defs.h"
#include <stdio.h>
#include "cmdl.h"
#include "cdata.h"
#include "gpu_util.h"
#include "gpudata.h"

GPU_DATA::GPU_DATA()
{
	CONST_DATA* t = static_cast<CONST_DATA*> (this);
	memset(t, 0, sizeof(*t));	 
	CScale = CData::C_SCALE; AScale = CData::A_SCALE;	
}

void GPU_DATA::Print(int n) const
{
	cl_ulong mask = 0x0FFFFFFFFFFFFULL;
	printf("%s--CONST--\n Inner:[%4i*%4i] Full:[%4i*%4i] Cell:[%3i*%3i]"
				 "\n DY=0x%04X   DZ=0x%08X   InOffs=0x%06X"
				 "\n NDets=%-4i:0x%X   NBnd=%-5i:0x%X  "
				 "\n CScale=%.4e  AScale=%.4e ", _PID,
				 InW, InH, FullW, FullH, BLKs.FwdW, BLKs.FwdY, DY, DZ, InOffsetB,
				 NumDetectors, NumDetectorsA, NumBoundary, NumBoundaryA,
				 CScale, AScale);
#ifdef TARGET_GPU
	if ( n & 4 )
		printf("%s---GPU---  Srcs=%i"
					 "\n COEFF->%012llX  end: %012llX  Coeff->%012llX"
					 "\n  LBUF->%012llX  end: %012llX   Grad->%012llX"
					 "\n    U0->%012llX   U1->%012llX     W0->%012llX   W1->%012llX ", _PID, GPUSrcs,
					 mask&(cl_ulong)GCoeffBuf.Addr, mask&(cl_ulong)(GCoeffBuf.Len + GCoeffBuf.Addr), mask&(cl_ulong)pgCoeff,
					 mask&(cl_ulong)GLayerBuf.Addr, mask&(cl_ulong)(GLayerBuf.Len + GLayerBuf.Addr), mask&(cl_ulong)pgGrad,
					 mask&(cl_ulong)pgLU.s[0], mask&(cl_ulong)pgLU.s[1],
					 mask&(cl_ulong)pgLW.s[0], mask&(cl_ulong)pgLW.s[1]);
#endif
} // Print()

#ifdef TARGET_GPU

#ifdef TARGET_WINDOWS
#include <io.h>
#include <ctime>
#include <windows.h>
#undef USE_MPI
#else // target linux
#undef USE_GL
#ifdef USE_MPI
#include "mpi.h"
#endif
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <errno.h>
#endif

#include <omp.h>

#define _USE_MATH_DEFINES 
#include <math.h>
#include <string.h>
#include <stdio.h>

#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>

#define cierr(s) if (cl_error != CL_SUCCESS) { printf("\n[%i] %s error %s\x07 ",MyId, s,ClErrorName(cl_error)); goto _cl_error; }


#define _K_(x) #x
const char* const KA_List[] = { DECLARE_KERNELS, NULL };
#undef _K_

cl_int GPU_DATA::AllocateOnGPU(void)	
{
	ENTER;	
	IFLOG(LOG_CHECKS) { printf("\n AllocGPU %i GPUSrcs ", GPUSrcs); fflush(stdout); }
	Release();
	cl_ulong Buf_a;
	cl_ulong memsz = __min(pGPU->MaxAllocationSize, pGPU->GlobalMemSize);
	const cl_ulong MSizeA = DY * FullH;

	pgParGrad = pgGrad = pgCoeff = pgLU.s[0] = pgLU.s[1] = pgLW.s[0] = pgLW.s[1] = 0;

	cl_int cl_error = 0;
	cl_image_format fmtC0 = { CL_RG, CL_UNSIGNED_INT16 };

	GConstBuf = clCreateBuffer(pGPU->CLContext, CL_MEM_READ_ONLY, sizeof(CONST_DATA), NULL, &cl_error);
	if ( cl_error != CL_SUCCESS ) { printf("\n %s CreateConstBuffer: error %i ", __FUNCTION__, cl_error); goto _cl_error; }

	for ( ; GPUSrcs > 0; GPUSrcs-- )	
	{
		size_t l1=BoundCh.BufSize(GPUSrcs),
			l2 = ExData.BufSize(GPUSrcs),
			l3= MSizeA * (GPUSrcs * 4 + 1) + 4096,
			l4= MSizeA * GPUSrcs + 4096,
		lmax = __max(l1, l2);
		lmax = __max(lmax, l3);
		if ( (lmax < pGPU->MaxAllocationSize) && ((l1 + l2 + l3 + l4) < pGPU->GlobalMemSize) ) break;
	}

	for ( ; GPUSrcs > 0; GPUSrcs-- )
	{
		IFLOG(LOG_CHECKS) { printf("\n TRY %i GPUSrcs ", GPUSrcs); fflush(stdout); 
		}
		if ( BoundCh.AllocateOnGPU(pGPU, GPUSrcs) ) continue;		
		if ( ExData.AllocateOnGPU(pGPU, GPUSrcs) ) goto Rls1;		
		GLayerBuf.Len = MSizeA * (GPUSrcs * 4 + 1) + 4096;
		if ( GLayerBuf.AllocBuffer(pGPU->CLContext, pGPU->CLQueue[0], (*pGPU)[k_GetAddr]) ) goto Rls2;		
		
		if ( ( GPUSrcs > 1 ) )//&& (InW+InH<MaxParGradSize*2))
		{
			GParGrad.Len = MSizeA * GPUSrcs + 4096;
			if ( GParGrad.AllocBuffer(pGPU->CLContext, pGPU->CLQueue[0], (*pGPU)[k_GetAddr]) ) goto Rls3;
		} else GParGrad.Len = 0;
		
		break;
		
	Rls3: GLayerBuf.Release();
	Rls2: ExData.Release();
	Rls1: BoundCh.Release();
	}
	
	IFLOG(LOG_CHECKS) { printf("\n AllocGPU %i GPUSrcs available", GPUSrcs); fflush(stdout); }

	if ( !GPUSrcs )
	{ fprintf(stderr, "Not enough GPU memory or allocation size"); cl_error = CL_INVALID_BUFFER_SIZE; goto _cl_error; }

	GCoeffBuf.Len = MSizeA + 4096;
	if ( cl_error |= GCoeffBuf.AllocBuffer(pGPU->CLContext, pGPU->CLQueue[0], (*pGPU)[k_GetAddr]) ) goto _cl_error;

	pgCoeff = align(GCoeffBuf.Addr, MemAlign) + MemoryShear[4] + InOffsetB;
	
	Buf_a = align(GParGrad.Addr, MemAlign);
	pgParGrad = Buf_a + MemoryShear[5] + InOffsetB;

	Buf_a = align(GLayerBuf.Addr, MemAlign);
	pgGrad = Buf_a + MemoryShear[5] + InOffsetB;
	pgLU.s[0] = Buf_a + MSizeA + MemoryShear[0] + InOffsetB;
	pgLU.s[1] = Buf_a + MSizeA * 2 + MemoryShear[1] + InOffsetB;
	pgLW.s[0] = Buf_a + MSizeA * 3 + MemoryShear[2] + InOffsetB;
	pgLW.s[1] = Buf_a + MSizeA * 4 + MemoryShear[3] + InOffsetB;

	IFLOG(LOG_INFO) printf("%s GPU Alloc Mb: %lli coeff +%lli main +%lli grad +%lli bound +%lli exdata ", _PID,
				 GCoeffBuf.Len >> 20, GLayerBuf.Len >> 20, GParGrad.Len >> 20, BoundCh.Len >> 20, ExData.Len >> 20);

_cl_error: 
	PRINT_CL_ERROR;
	RETURN(cl_error);
}

cl_int GPU_DATA::UploadToGPU(void) const
{
	CHECKPOINT("CDATA UploadtoGPU ");
	cl_int ret = CL_INVALID_VALUE;
	if ( !pGPU ) return CL_INVALID_CONTEXT;
	ret = clEnqueueWriteBuffer(pGPU->CLQueue[0], GConstBuf, CL_TRUE, 0,
		sizeof(CONST_DATA), this, 0, NULL, NULL);
	if ( ret ) return(ret);

	cl_ulong p[8] = { 0,0,0,0,0,0,0,0 };
	clFinish(pGPU->CLQueue[0]);
	SETARG0(k_CheckData);
	clSetKernelArg((*pGPU)[k_CheckData], 1, 8, &(GLayerBuf.Addr));
	ret |= clEnqueueTask(pGPU->CLQueue[0], (*pGPU)[k_CheckData], 0, NULL, NULL);
	clFinish(pGPU->CLQueue[0]);

	GLayerBuf.Xfer(false, p, GLayerBuf.Addr, 64, pGPU->CLQueue[0]);
	clFinish(pGPU->CLQueue[0]);
	if ( p[0] != pgLU.s[0] ) return CL_INVALID_VALUE;
	if ( p[1] != pgLU.s[1] ) return CL_INVALID_VALUE;
	if ( p[2] != pgLW.s[0] ) return CL_INVALID_VALUE;
	if ( p[3] != pgLW.s[1] ) return CL_INVALID_VALUE;
	if ( p[4] != pgGrad ) return CL_INVALID_VALUE;
	if ( p[5] != pgCoeff ) return CL_INVALID_VALUE;
	if ( p[6] != DZ ) return CL_INVALID_VALUE;
	return ret;
}


#endif 

cl_int GPU_DATA::Release()
{
	cl_int r = 0;
#ifdef TARGET_GPU
	r |= BoundCh.Release();
	r |= ExData.Release();
	r |= GLayerBuf.Release();
	r |= GCoeffBuf.Release();
	if ( GConstBuf ) { r |= clReleaseMemObject(GConstBuf); GConstBuf = NULL; }
	if ( r ) puts(ClErrorName(r));
#endif
	return r;
}//fn
