#include "setup.h"
#ifdef TARGET_GPU

#include "k_defs.h"

#define _USE_MATH_DEFINES 
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <CL/cl.h>

#include "cmdl.h"
#include "gpu_util.h"

#define cierr(s) if (cl_error != CL_SUCCESS) { printf("%s %s CL error %s\x07 ",_PID, s,ClErrorName(cl_error)); goto _cl_error; }



char* printkernel(cl_kernel k)
{
	cl_uint na = 0; size_t rvs;
	static char str[128];
	str[0] = '?'; str[1] = 0;
	if ( CL_SUCCESS != clGetKernelInfo(k, CL_KERNEL_NUM_ARGS, sizeof(cl_uint), &na, &rvs) ) return str;
	clGetKernelInfo(k, CL_KERNEL_FUNCTION_NAME, 112, str, &rvs);
	sprintf(str + rvs, "(%i) ", na);
	return str;
}


char* LoadProgSource(const char* header, const char* text, size_t* len, const char* addstr)
{
	FILE *fp1, *fp2; char*p = NULL; size_t l1, l2, l0 = addstr ? strlen(addstr) : 0;
	if ( len ) *len = 0;
	fp1 = fopen(header, "rb");  if ( !fp1 ) goto ret;
	fp2 = fopen(text, "rb"); if ( !fp2 ) goto close1;
	fseek(fp1, 0, SEEK_END); fseek(fp2, 0, SEEK_END);
	l1 = ftell(fp1);         l2 = ftell(fp2);
	fseek(fp1, 0, SEEK_SET); fseek(fp2, 0, SEEK_SET);
	p = (char*)malloc(l1 + l2 + l0 + 1); if ( !p ) goto close2;
	if ( addstr ) strcpy(p, addstr);
	fread(p + l0, 1, l1, fp1);     fread(p + l1 + l0, 1, l2, fp2); p[l1 + l2 + l0] = 0;
	if ( len ) *len = l1 + l2 + l0;
close2: fclose(fp2);
close1: fclose(fp1);
ret:    return p;
}


cl_int cl_error = 0;
const char* cl_errmsg = " no msg ";


cl_int GPUContext::Init(int CurPlatform, int GPU_ID, const char* const* KA_List, const char* src1, const char* src2)
{
	int QUEUE_OPTIONS = CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE;
	IFLOG(LOG_GPUPROF) { QUEUE_OPTIONS |= CL_QUEUE_PROFILING_ENABLE; }

	fReady = false;	
	constexpr int TMPSTR_LEN = 512;
	if ( (CurPlatform < 0) || (CurPlatform >= MAX_PLATFORMS) ||
		(GPU_ID < 0) || (GPU_ID > MAX_GPU_COUNT) ) return CL_INVALID_VALUE;

	ENTER;
	char tmpstr[TMPSTR_LEN];
	cl_platform_id cpPlatforms[MAX_PLATFORMS] = { 0 };
	cl_device_id cdDevices[MAX_GPU_COUNT] = { 0 };

	char* source = NULL; size_t program_length = 0;
	cl_int cl_error = CL_SUCCESS;
	cl_uint ciNumPlatforms, ciDeviceCount;
	int i;

	memset(this, 0, sizeof(this));

	cl_error = clGetPlatformIDs(8, cpPlatforms, &ciNumPlatforms); cierr("GetPlatformID");
	for ( i = 0; i < ciNumPlatforms; i++ )
	{
		*tmpstr = 0;  clGetPlatformInfo(cpPlatforms[i], CL_PLATFORM_NAME, TMPSTR_LEN - 1, tmpstr, NULL);
		if ( !MyId ) printf("%s PLATFORM #%i: %s ", _PID, i, tmpstr);
	}

	cl_error = clGetDeviceIDs(cpPlatforms[CurPlatform], CL_DEVICE_TYPE_GPU, MAX_GPU_COUNT, cdDevices, &ciDeviceCount); cierr("GetDevideIDs");
	printf("%s DEVICES: %i ", _PID, ciDeviceCount);

	for ( i = 0; i < ciDeviceCount; i++ )
	{
		size_t MaxWISize[8] = { 0,0,0,0,0,0,0,0 };

#define GETDEVINFO(name,typ,var) typ var; clGetDeviceInfo (cdDevices[i], name, sizeof(typ), &var, NULL);
#define GETDEVSTRING(name,l,var)  size_t l; \
    clGetDeviceInfo(cdDevices[i], name, 0,NULL, &l); \
    memblk var(l,"DEVString"); \
    clGetDeviceInfo(cdDevices[i], name, var.len, var.data, &l);
		//--end-macro--

		GETDEVINFO(CL_DEVICE_MAX_WORK_GROUP_SIZE, size_t, maxWGSize);
		GETDEVINFO(CL_DEVICE_LOCAL_MEM_SIZE, cl_ulong, LocSize);
		GETDEVINFO(CL_DEVICE_GLOBAL_MEM_SIZE, cl_ulong, GlobSize);
		GETDEVINFO(CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE, cl_ulong, ConstSize);
		GETDEVINFO(CL_DEVICE_MAX_CONSTANT_ARGS, cl_uint, ConstArgs);
		GETDEVINFO(CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, cl_uint, maxWIDim);
		GETDEVINFO(CL_DEVICE_MAX_COMPUTE_UNITS, cl_uint, maxCU);
		GETDEVINFO(CL_DEVICE_MAX_MEM_ALLOC_SIZE, cl_ulong, maxAlloc);
		GETDEVINFO(CL_DEVICE_GLOBAL_MEM_CACHE_SIZE, cl_ulong, CacheSize);
		GETDEVINFO(CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE, cl_uint, CLineSize);
		GETDEVINFO(CL_DEVICE_ENDIAN_LITTLE, cl_bool, Dev_LE);
		GETDEVINFO(CL_DEVICE_ADDRESS_BITS, cl_uint, AddrBits);
		clGetDeviceInfo(cdDevices[i], CL_DEVICE_MAX_WORK_ITEM_SIZES, maxWIDim * sizeof(size_t), MaxWISize, NULL);
		GETDEVSTRING(CL_DEVICE_EXTENSIONS, extLen, extensions);
		GETDEVSTRING(CL_DEVICE_NAME, nameLen, name);

		IFLOG(LOG_GPU)
		printf("%s *** DEVICE #%i: %s  Cores:%i  LE:%i\nGMem=%llu [%i bit] MaxAlloc=%llu \n  CacheSize=%llu,%u"\
												" LMem=%llu  Const=%lu,%u args\n  MaxWGSize=%i  dim=%i [ %i %i %i %i ]\nExt: %s\n", _PID,
												i, name.data, (int)maxCU, (int)Dev_LE, (cl_ulong)GlobSize, (int)AddrBits, (cl_ulong)maxAlloc, CacheSize, CLineSize,
												(cl_ulong)LocSize, (long)ConstSize, (int)ConstArgs, (int)maxWGSize, (int)maxWIDim,
												(int)MaxWISize[0], (int)MaxWISize[1], (int)MaxWISize[2], (int)MaxWISize[3], extensions.data);
		if ( i == GPU_ID )
		{
			MaxAllocationSize = maxAlloc;
			LocalMemSize = LocSize-256;
			GlobalMemSize = GlobSize;
			MaxBlockSize = maxWGSize;
		}
	} // for devices 

	{ cl_context_properties props[] =
	{
#ifdef USE_GL
		CL_GL_CONTEXT_KHR,  (cl_context_properties)wglGetCurrentContext(),
		CL_WGL_HDC_KHR,     (cl_context_properties)wglGetCurrentDC(),
#endif
		CL_CONTEXT_PLATFORM,(cl_context_properties)cpPlatforms[CurPlatform],    0 };
	CLContext = clCreateContext(props, 1, cdDevices + GPU_ID, NULL, NULL, &cl_error); cierr("CreateContext");
}


// Find out how many GPU's to compute on all available GPUs
	size_t nDeviceBytes;
	cl_error = clGetContextInfo(CLContext, CL_CONTEXT_DEVICES, ciDeviceCount * sizeof(cl_device_id), cxDevices, &nDeviceBytes);
	cierr("GetContextInfo");

	cxDeviceCount = (cl_uint)nDeviceBytes / sizeof(cl_device_id);
	if ( cxDeviceCount == 0 ) { printf("%s (%i) No OpenCL devices (%s) ", _PID, cl_error, ClErrorName(cl_error));  goto _cl_error; }
	printf("%s Devices in context: %i ", _PID, cxDeviceCount);


	// create command-queues
	for ( i = 0; i < cxDeviceCount; ++i )
	{
		
		cl_command_queue_properties props = 0; size_t sz;
		CLQueue[i] = clCreateCommandQueue(CLContext, cxDevices[i], QUEUE_OPTIONS, &cl_error); cierr("CreateCommandQueue");
		CLQueu2[i] = clCreateCommandQueue(CLContext, cxDevices[i], QUEUE_OPTIONS, &cl_error); cierr("CreateCommandQueu2");
		clGetCommandQueueInfo(CLQueue[i], CL_QUEUE_PROPERTIES, sizeof(cl_command_queue_properties), &props, &sz);
		props = 0;
		clGetCommandQueueInfo(CLQueu2[i], CL_QUEUE_PROPERTIES, sizeof(cl_command_queue_properties), &props, &sz);
	}
	CHECKPOINT("CreateQueue ");

	// Program Setup
	sprintf(tmpstr, "\n#define GPU_LOCAL_MEM_SIZE %i\n", LocalMemSize);
	source = LoadProgSource(src1, src2, &program_length,tmpstr);
	if ( !source ) { printf("%s Error: Failed to load the script! ", _PID);  return 1; }
	cpProgram = clCreateProgramWithSource(CLContext, 1, (const char **)&source, &program_length, &cl_error); cierr("CreateProgram");
	free(source);
	CHECKPOINT("LoadProg ");

	cl_error = clBuildProgram(cpProgram, cxDeviceCount, cxDevices, "-cl-fast-relaxed-math -cl-mad-enable -cl-no-signed-zeros ", NULL, NULL);
	if ( cl_error )
	{
		char tmpname[32];
		clGetProgramBuildInfo(cpProgram, cxDevices[0], CL_PROGRAM_BUILD_LOG, sizeof(tmpstr) - 16, tmpstr, NULL);
		puts(tmpstr); puts("Error building program.");
		snprintf(tmpname, 31, "oclbuild%02i.log", MyId);
		DUMP(tmpname, tmpstr, strlen(tmpstr));
		goto _cl_error;
	}

	CHECKPOINT("BuildProg ");
	if ( G_Settings.LogLevel&TEST_BUILDPROG ) { puts("\n*** TEST MODE - OpenCL Build successful - EXITING ***"); cl_error = 1; goto _cl_error; }

	LinClampSampler = clCreateSampler(CLContext, CL_FALSE, CL_ADDRESS_CLAMP_TO_EDGE, CL_FILTER_LINEAR, &cl_error);  cierr("CreateSampler1");
	LinNormSampler = clCreateSampler(CLContext, CL_TRUE, CL_ADDRESS_CLAMP_TO_EDGE, CL_FILTER_LINEAR, &cl_error);  cierr("CreateSampler2");
	NearClampSampler = clCreateSampler(CLContext, CL_FALSE, CL_ADDRESS_CLAMP_TO_EDGE, CL_FILTER_NEAREST, &cl_error);  cierr("CreateSampler3");
	KernelArray::Init(KA_List, cpProgram);
	if ( !cl_error ) fReady = true;
_cl_error: RETURN(cl_error);
} // Init()


cl_int GPUContext::Close(void)
{
	ENTER;
	cl_error = 0;
  
	IFLOG(LOG_GPUPROF) { if ( !MyId ) KernelArray::PrintStats(1); };

	CHECKPOINT("KrnlArray.Release ");  KernelArray::Release();
	CHECKPOINT("ReleaseCommandQueue"); for ( int i = 0; i < cxDeviceCount; ++i ) { clReleaseCommandQueue(CLQueue[i]); clReleaseCommandQueue(CLQueu2[i]); }
	CHECKPOINT("ReleaseProgram");      clReleaseProgram(cpProgram);
	cpProgram = 0;
#ifdef USE_GL
	clReleaseMemObject(GLViewer.cl_pbo); GLViewer.cl_pbo = 0;
#endif
	cl_error = clReleaseContext(CLContext); //cierr("ReleaseContext");			
	RETURN(cl_error);
}



const char* ClErrorName(cl_int n)
{
	static const char* const errnames[64] = {
" CL_SUCCESS-0 ",
" CL_DEVICE_NOT_FOUND-1 ",
" CL_DEVICE_NOT_AVAILABLE-2 ",
" CL_COMPILER_NOT_AVAILABLE-3 ",
" CL_MEM_OBJECT_ALLOCATION_FAILURE-4 ",
" CL_OUT_OF_RESOURCES-5 ",
" CL_OUT_OF_HOST_MEMORY-6 ",
" CL_PROFILING_INFO_NOT_AVAILABLE-7 ",
" CL_MEM_COPY_OVERLAP-8 ",
" CL_IMAGE_FORMAT_MISMATCH-9 ",
" CL_IMAGE_FORMAT_NOT_SUPPORTED-10 ",
" CL_BUILD_PROGRAM_FAILURE-11 ",
" CL_MAP_FAILURE-12 ",
" (?)-13 ", " (?)-14 ", " (?)-15 ", " (?)-16 ",
" (?)-17 ", " (?)-18 ", " (?)-19 ", " (?)-20 ",
" (?)-21 ", " (?)-22 ", " (?)-23 ", " (?)-24 ",
" (?)-25 ", " (?)-26 ", " (?)-27 ", " (?)-28 ",
" (?)-29 ",
" CL_INVALID_VALUE-30 ",
" CL_INVALID_DEVICE_TYPE-31 ",
" CL_INVALID_PLATFORM-32 ",
" CL_INVALID_DEVICE-33 ",
" CL_INVALID_CONTEXT-34 ",
" CL_INVALID_QUEUE_PROPERTIES-35 ",
" CL_INVALID_COMMAND_QUEUE-36 ",
" CL_INVALID_HOST_PTR-37 ",
" CL_INVALID_MEM_OBJECT-38 ",
" CL_INVALID_IMAGE_FORMAT_DESCRIPTOR-39 ",
" CL_INVALID_IMAGE_SIZE-40 ",
" CL_INVALID_SAMPLER-41 ",
" CL_INVALID_BINARY-42 ",
" CL_INVALID_BUILD_OPTIONS-43 ",
" CL_INVALID_PROGRAM-44 ",
" CL_INVALID_PROGRAM_EXECUTABLE-45 ",
" CL_INVALID_KERNEL_NAME-46 ",
" CL_INVALID_KERNEL_DEFINITION-47 ",
" CL_INVALID_KERNEL-48 ",
" CL_INVALID_ARG_INDEX-49 ",
" CL_INVALID_ARG_VALUE-50 ",
" CL_INVALID_ARG_SIZE-51 ",
" CL_INVALID_KERNEL_ARGS-52 ",
" CL_INVALID_WORK_DIMENSION-53 ",
" CL_INVALID_WORK_GROUP_SIZE-54 ",
" CL_INVALID_WORK_ITEM_SIZE-55 ",
" CL_INVALID_GLOBAL_OFFSET-56 ",
" CL_INVALID_EVENT_WAIT_LIST-57 ",
" CL_INVALID_EVENT-58 ",
" CL_INVALID_OPERATION-59 ",
" CL_INVALID_GL_OBJECT-60 ",
" CL_INVALID_BUFFER_SIZE-61 ",
" CL_INVALID_MIP_LEVEL-62 ",
" CL_INVALID_GLOBAL_WORK_SIZE-63 " };
	static char tmp[64];
	if ((n > 0) || (n < -63) ) 
	{ sprintf(tmp, " ??cl_error=%i ", n); return tmp; } else return errnames[-n];
}

#ifdef CL_REPORT_ERRORS
inline cl_int clEnqueueReadBuffer1(cl_command_queue command_queue, cl_mem buffer, cl_bool blocking_read,
																	 size_t offset, size_t cb, void *ptr, cl_uint num_events_in_wait_list, const cl_event *event_wait_list, cl_event *evt)
{
	cl_int err = clEnqueueReadBuffer(command_queue, buffer, blocking_read, offset, cb, ptr, num_events_in_wait_list, event_wait_list, evt);
	if ( err ) printf("\n clEnqRead offs=%llXh size=%llXh ERROR: %s ", (cl_long)offset, (cl_long)cb, ClErrorName(err));
	return err;
}

inline cl_int clEnqueueWriteBuffer1(cl_command_queue command_queue, cl_mem buffer, cl_bool blocking_read,
																		size_t offset, size_t cb, void *ptr, cl_uint num_events_in_wait_list, const cl_event *event_wait_list, cl_event *evt)
{
	cl_int err = clEnqueueWriteBuffer(command_queue, buffer, blocking_read, offset, cb, ptr, num_events_in_wait_list, event_wait_list, evt);
	if ( err ) printf("\n clEnqWrite offs=%llXh size=%llXh ERROR: %s ", (cl_long)offset, (cl_long)cb, ClErrorName(err));
	return err;
}

#else
#define clEnqueueWriteBuffer1 clEnqueueWriteBuffer
#define clEnqueueReadBuffer1 clEnqueueReadBuffer
#endif


cl_int AllocateGPUImage(cl_context& cx, cl_command_queue& cq, cl_image_format* pFmt, cl_uint4& FwdDim, cl_mem *retImg)
{
	size_t res = 0, res_sz = 0, org[3] = { 0,0,0 }, rgn[3] = { 1,1,1 };
	cl_int ret, test[16];
	*retImg = clCreateImage3D(cx, CL_MEM_READ_ONLY, pFmt, FwdDim.s[0], FwdDim.s[1], FwdDim.s[2], 0, 0, NULL, &ret);
	ret |= clEnqueueReadImage(cq, *retImg, CL_TRUE, org, rgn, 0, 0, test, 0, NULL, NULL);
	ret |= clGetMemObjectInfo(*retImg, CL_MEM_SIZE, sizeof(size_t), &res, &res_sz);
	if ( ret != CL_SUCCESS ) { printf("%sAllocGPUImage [%i*%i*%i] failed: %s\n", _PID, FwdDim.s[0], FwdDim.s[1], FwdDim.s[2], ClErrorName(ret)); *retImg = NULL; } else if ( G_Settings.LogLevel&(LOG_GPU | LOG_MALLOC | LOG_INFO) ) printf("%s AllocGPUImage [%iMb] ", _PID, (int)(res >> 20));
	return ret;
}

cl_int AllocateGPUImage2D(cl_context& cx, cl_command_queue& cq, cl_image_format* pFmt, cl_uint4& FwdDim, cl_mem *retImg)
{
	size_t res = 0, res_sz = 0, org[3] = { 0,0,0 }, rgn[3] = { 1,1,1 };
	cl_int ret, test[16];
	*retImg = clCreateImage2D(cx, CL_MEM_READ_ONLY, pFmt, FwdDim.s[0], FwdDim.s[1], 0, NULL, &ret);
	ret |= clEnqueueReadImage(cq, *retImg, CL_TRUE, org, rgn, 0, 0, test, 0, NULL, NULL);
	ret |= clGetMemObjectInfo(*retImg, CL_MEM_SIZE, sizeof(size_t), &res, &res_sz);
	if ( ret != CL_SUCCESS ) { printf("%sAllocGPUImage [%i*%i] failed: %s\n", _PID, FwdDim.s[0], FwdDim.s[1], ClErrorName(ret)); *retImg = NULL; } else if ( G_Settings.LogLevel&(LOG_GPU | LOG_MALLOC | LOG_INFO) ) printf("%s AllocGPUImage [%iMb] ", _PID, (int)(res >> 20));
	return ret;
}


cl_int AllocateGPUBuffer(cl_context& cx, cl_ulong TotalMemReq, cl_command_queue& cq, cl_kernel& kGetAddr, cl_mem *retBuf, cl_ulong* retAddr)
{
	cl_mem membuf;
	cl_long test[16], res[16]; cl_int ret = 0;
	*test = -2LL;
	*retBuf = NULL; if ( !TotalMemReq ) return 0;
	membuf = *retBuf = clCreateBuffer(cx, CL_MEM_READ_WRITE, TotalMemReq, NULL, &ret);  if ( ret != CL_SUCCESS ) goto _ret;
	ret = clSetKernelArg(kGetAddr, 0, sizeof(cl_mem), (void *)(&membuf));  if ( ret != CL_SUCCESS ) goto _ret;
	ret = clEnqueueWriteBuffer(cq, membuf, CL_TRUE, 0, sizeof(cl_long), test, 0, NULL, NULL);  if ( ret != CL_SUCCESS ) goto _ret;
	*test = 0;
	ret = clEnqueueReadBuffer(cq, membuf, CL_TRUE, 0, sizeof(cl_long), test, 0, NULL, NULL);   if ( ret != CL_SUCCESS ) goto _ret;
	if ( *test != -2 ) { ret = 1; goto _ret; }

	ret = clEnqueueTask(cq, kGetAddr, 0, NULL, NULL); if ( ret != CL_SUCCESS ) goto _ret;
	clFinish(cq); *res = -2;

	ret = clEnqueueReadBuffer(cq, membuf, CL_TRUE, 0, sizeof(cl_long) * 4, res, 0, NULL, NULL);  if ( ret != CL_SUCCESS ) goto _ret;
	if ( *res == -2 ) { ret = 2; goto _ret; }
 // printf("\nGetAddr: %llXh %llXh %llXh \n",res[0],res[1],res[2]);
	if ( retAddr ) *retAddr = *res;

_ret:
	if ( ret ) printf("%sAllocGPUBuf [%iMb] : %s", _PID, (int)(TotalMemReq >> 20), ClErrorName(ret));
	else if ( G_Settings.LogLevel&(LOG_MALLOC | LOG_GPU) ) printf("%sAllocGPUBuf [%iMb] @0x%llX ", _PID, (int)(TotalMemReq >> 20), *res);
	return ret;
}




//============================================================================
//BLK: GPU UTIL



size_t MRangeBase::TotalItems(void) const
{
	size_t rv = 1; int i;
	if ( (dim > 3) || (dim < 1) ) return 0;
	for ( i = 0; i < dim; ++i ) rv *= GWS[i];
	return rv;
}

size_t MRangeBase::TotalGroups(void) const
{
	size_t rv = 1; int i;
	if ( (dim > 3) || (dim < 1) ) return 0;
	for ( i = 0; i < dim; ++i ) rv *= (GWS[i] / LWS[i]);
	return rv;
}



Kernel::Kernel(cl_program cpProgram, const char* kname)
{
	err = 0; k = clCreateKernel(cpProgram, kname, &err);
	strncpy(name, kname, 48);
	evt = NULL; elapsed_ns = overlap_ns = 0; n_run = 0; max_ns = 0; min_ns = -1; st = en = 0;
}


cl_ulong Kernel::Profile(bool wait, cl_ulong intersect_st, cl_ulong intersect_en)
{
	cl_ulong elapsed, e1;
	if ( !evt ) return 0;
	if ( wait ) clWaitForEvents(1, &evt);

	clGetEventProfilingInfo(evt, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &st, NULL);
	clGetEventProfilingInfo(evt, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &en, NULL);

	clReleaseEvent(evt); evt = NULL;
	elapsed = en - st;
	elapsed_ns += elapsed; if ( elapsed < min_ns ) min_ns = elapsed; if ( elapsed > max_ns ) max_ns = elapsed; ++n_run;

	if ( intersect_en > intersect_st )
	{
		elapsed = ((en < intersect_en) ? en : intersect_en);
		e1 = ((st > intersect_st) ? st : intersect_st);
		if ( elapsed > e1 ) overlap_ns += (elapsed - e1);
	}
	return en;
}

void Kernel::Release(void)
{
	cl_int err1 = CL_SUCCESS;
	if ( k ) err1 = clReleaseKernel(k);    //printf("\nReleaseK %08X",(int)k); 
	if ( !err ) err = err1;
	if ( evt ) clReleaseEvent(evt); evt = NULL;
}

inline cl_int Kernel::Read(cl_command_queue& cq, cl_mem& cm, cl_bool blocking, size_t pos, size_t sz, void* ptr)
{
	cl_errmsg = name;
	err = k ? CL_INVALID_KERNEL : clEnqueueReadBuffer(cq, cm, blocking, pos, sz, ptr, 0, NULL, &evt);
	if ( G_Settings.LogLevel&LOG_KRNL ) printf("%s %s Read%c buf=%llX @%llX [%llX] =%s",
																						 _PID, name, blocking ? 'W' : ' ', (cl_ulong)cm, (cl_ulong)pos, (cl_ulong)sz, ClErrorName(err));
	if ( blocking && !err ) Profile();
	cl_error |= err; return err;
}

inline cl_int Kernel::Write(cl_command_queue& cq, cl_mem& cm, cl_bool blocking, size_t pos, size_t sz, void* ptr)
{
	cl_errmsg = name;
	err = k ? CL_INVALID_KERNEL : clEnqueueWriteBuffer(cq, cm, blocking, pos, sz, ptr, 0, NULL, &evt);
	if ( G_Settings.LogLevel&LOG_KRNL ) printf("%s %s Write%c buf=%llX @%llX [%llX] =%s",
																						 _PID, name, blocking ? 'W' : ' ', (cl_ulong)cm, (cl_ulong)pos, (cl_ulong)sz, ClErrorName(err));
	if ( blocking && !err ) Profile();
	cl_error |= err; return err;
}


cl_int Kernel::ExecWait(cl_command_queue& cq, MRangeBase& NDR)
{
	cl_int rv = CL_SUCCESS;
	cl_errmsg = name;
	if ( k ) rv = clEnqueueNDRangeKernel(cq, k, NDR.dim, NULL, NDR.GWS, NDR.LWS, 0, NULL, &evt);
	if ( G_Settings.LogLevel&LOG_KRNL ) printf("%s %s ExecW [%i*%i*%i/%i*%i*%i] =%s",
																						 _PID, name, NDR.GWS[0], NDR.GWS[1], NDR.GWS[2], NDR.LWS[0], NDR.LWS[1], NDR.LWS[2], ClErrorName(err));
	if ( !rv ) { clWaitForEvents(1, &evt);  Profile(); }
	cl_error |= rv;  return rv;
}

cl_int Kernel::ExecNoWait(cl_command_queue& cq, MRangeBase& NDR, bool use_event)
{
	cl_errmsg = name;
	err = k ? clEnqueueNDRangeKernel(cq, k, NDR.dim, NULL, NDR.GWS, NDR.LWS, 0, NULL, use_event ? &evt : NULL) : CL_INVALID_KERNEL;
	if ( G_Settings.LogLevel&LOG_KRNL ) printf("%s %s Exec  [%i*%i*%i/%i*%i*%i] =%s",
																						 _PID, name, NDR.GWS[0], NDR.GWS[1], NDR.GWS[2], NDR.LWS[0], NDR.LWS[1], NDR.LWS[2], ClErrorName(err));
	cl_error |= err;  return err;
}

cl_int Kernel::ExecNoWait(cl_command_queue& cq, MRangeBase& NDR, cl_event* wait)
{
	cl_errmsg = name;
	err = k ? clEnqueueNDRangeKernel(cq, k, NDR.dim, NULL, NDR.GWS, NDR.LWS, 1, wait, &evt) : CL_INVALID_KERNEL;;
	if ( G_Settings.LogLevel&LOG_KRNL ) printf("%s %s Exec  [%i*%i*%i/%i*%i*%i] =%s",
																						 _PID, name, NDR.GWS[0], NDR.GWS[1], NDR.GWS[2], NDR.LWS[0], NDR.LWS[1], NDR.LWS[2], ClErrorName(err));
	cl_error |= err;  return err;
}


cl_int KernelArray::GetErrors(void) const
{ cl_int e = 0;  for ( cl_uint i = 0; i < used; ++i ) if ( !e ) e = pK[i].err; return(e); }

cl_int KernelArray::Init(const char* const * name, cl_program cpProgram)
{
	cl_uint i, n = 0;
	if ( used ) Release();
	while ( name[n] ) ++n;
	used = n; pK = new Kernel[n];
	for ( i = 0; i < used; ++i ) pK[i] = Kernel((*(name[i]) == '$') ? NULL : cpProgram, name[i]);
	// use ".Name" for buffer profiling items
	return GetErrors();
}

cl_int KernelArray::SetArg(cl_uint krn, cl_uint argn, size_t argsz, const void* argptr) const
{
	if ( krn >= used || !pK[krn].k ) return -1;
	return clSetKernelArg(pK[krn], argn, argsz, argptr);
}

float KernelArray::TotalElapsed(int sel) const
{
	cl_ulong ns = 0; cl_uint i;
	if ( sel & 0x80 )
	{
		if ( sel & 1 ) for ( i = 0; i < used; ++i ) if ( pK[i].k ) ns += pK[i].overlap_ns;
		if ( sel & 2 ) for ( i = 0; i < used; ++i ) if ( !pK[i].k ) ns += pK[i].overlap_ns;
	} else
	{
		if ( sel & 1 ) for ( i = 0; i < used; ++i ) if ( pK[i].k ) ns += pK[i].elapsed_ns - pK[i].overlap_ns;
		if ( sel & 2 ) for ( i = 0; i < used; ++i ) if ( !pK[i].k ) ns += pK[i].elapsed_ns - pK[i].overlap_ns;
	}
	return (float)ns*1.0e-9f;
}

float KernelArray::PrintStats(bool reset) const
{
	static const float msec = 0.000001f, usec = 0.001f, sec = 0.000000001f;
	static char strS[] = "s ", strMS[] = "ms";
	cl_ulong CurTime = Microseconds();
	float total = TotalElapsed(1), totalxfer = TotalElapsed(2), // seconds
		cputime = CPUStartTime ? (CurTime - CPUStartTime)*1e-6f : 0.0f;
	total += totalxfer;
	totalxfer = totalxfer * 100.0f / total; // percentage
	cl_uint i;
	time_t SystemTime = time(0);
	printf("\n%s  Runs  |  Avg  |  Min  |  Max  |  Overlap  |%9.2fs GPU, %3.2f%% xfer\n"\
				 "--------+-------+-------+-------+-----------+%9.2fs %6.2f%% *CPU*\n",
				 asctime(localtime(&SystemTime)),
				 total, totalxfer, cputime, cputime*100.0f / total);
	total *= 1.0e7f; // 1e7= nanosecond percentage
	float tdiv = __max(total, 1);

	for ( i = 0; i < used; ++i )
	{ //clGetKernelInfo(pK[i],CL_KERNEL_FUNCTION_NAME,128,str,&rvs);
		if ( pK[i].n_run )
		{
			bool bOverlapS = (pK[i].overlap_ns > 9999000000ULL),
				bElapsedS = (pK[i].elapsed_ns > 9999000000ULL);
			cl_uint avg = ((pK[i].elapsed_ns) / 1000 / pK[i].n_run), tmin = pK[i].min_ns / 1000, tmax = pK[i].max_ns / 1000;

#define xxscale(val,mod) (val)>999999? (int)((val)/1000): (int)(val), (val)>999999? mod:" "

			printf(" %6i%s|%6i%s:%6i%s:%6i%s|%9.2f%s|%9.2f%s %5.2f%% %s\n",
						 xxscale(pK[i].n_run, "K"), xxscale(avg, "m"), xxscale(tmin, "m"), xxscale(tmax, "m"),
						 (bOverlapS ? sec : msec)*(float)pK[i].overlap_ns, bOverlapS ? strS : strMS,
						 (bElapsedS ? sec : msec)*(float)pK[i].elapsed_ns, bElapsedS ? strS : strMS,
						 (float)(pK[i].elapsed_ns - pK[i].overlap_ns) / tdiv, pK[i].name);
#undef xxscale
		}
		if ( reset ) { pK[i].elapsed_ns = pK[i].max_ns = pK[i].overlap_ns = 0; pK[i].min_ns = -1; pK[i].n_run = 0; }
	}
	return total * 0.0001; // return milliseconds
}

void KernelArray::Release(void)
{
	for ( cl_uint i = 0; i < used; ++i )  pK[i].Release();
	used = 0;
	if ( pK ) delete[] pK; pK = NULL; // puts("ReleaseKernelArray");  
}



//============================================================================


void MRange::Init(const char* nm, int w, int h, int ww, int wh, int cw, int ch, int halo, int vs)
{
	dim = 2;
	Wid = w; Hei = h; WorkW = ww; WorkH = wh; CellW = cw; CellH = ch; name = nm;
	if ( vs > 0 ) VectorSize = vs; else // vs<0 specify max. # of threads, then autoselect vs
	{
		VectorSize = (cw*ch) / (-vs);
		if ( VectorSize >= ch )  VectorSize = ch; else
			while ( ch%VectorSize ) ++VectorSize;
	}
	BlockW = cw - halo - halo; BlockH = ch - halo - halo;
	OffsX = WorkW % VectorSize; OffsY = WorkH % VectorSize;
	if ( OffsX ) OffsX = VectorSize - OffsX;
	if ( OffsY ) OffsY = VectorSize - OffsY; // shift the last group inwards by this value
	GWS[0] = ((WorkW - halo - halo + OffsX + BlockW - 1) / BlockW)*(LWS[0] = cw);
	GWS[1] = ((WorkH - halo - halo + OffsY + BlockH - 1) / BlockH)*(LWS[1] = ch / VectorSize);
	GWS[2] = LWS[2] = 0;
	//if( (!MyId) && (G_Settings.LogLevel&LOG_INFO) ) Print();
};

void MRange::Print(void) const
{
	printf("\n%s: %i*%i => %i*[%i*%i] / %i*[%i*%i] -%i,%i; blk=[%i*%i] ",
				 name, Wid, Hei, GWS[0], GWS[1], VectorSize, LWS[0], LWS[1], VectorSize, OffsX, OffsY, BlockW, BlockH);
}

int MRange::ExecWait(cl_command_queue& cq, Kernel& K, cl_event* pE) const
{
	cl_event* pe = pE ? pE : &(K.evt);
	int err = clEnqueueNDRangeKernel(cq, K.k, 2, NULL, GWS, LWS, 0, NULL, pe);
	if ( err )  printf("%s Exec '%s' Error: %s \x07", _PID, name, ClErrorName(err));
	else if ( !pE ) clFinish(cq);
	return err;
}

int MRange::ExecNoWait(cl_command_queue& cq, cl_kernel& k) const
{
	int err = clEnqueueNDRangeKernel(cq, k, 2, NULL, GWS, LWS, 0, NULL, NULL);
	if ( err )  printf("%s Exec '%s' Error: %s \x07", _PID, name, ClErrorName(err));
	return err;
}


void MRange3::Init(const char* nm, int w, int h, int d, int ww, int wh, int wd, int cw, int ch, int cd, int halo, int vs)
{
	dim = 3;
	Wid = w; Hei = h; Dep = d; WorkW = ww; WorkH = wh; WorkD = wd; CellW = cw; CellH = ch; CellD = cd; name = nm;
	VectorSize = vs;
	BlockW = cw - halo - halo; BlockH = ch - halo - halo; BlockD = cd - halo - halo;
	OffsX = WorkW % vs; OffsY = WorkH % vs; OffsZ = WorkD % vs;
	if ( OffsX ) OffsX = vs - OffsX;  if ( OffsZ ) OffsZ = vs - OffsZ;
	if ( OffsY ) OffsY = vs - OffsY; // shift the last group inwards by this value
	GWS[0] = ((WorkW - halo - halo + OffsX + BlockW - 1) / BlockW)*(LWS[0] = cw);
	GWS[1] = ((WorkH - halo - halo + OffsY + BlockH - 1) / BlockH)*(LWS[1] = ch);
	GWS[2] = ((WorkD - halo - halo + OffsZ + BlockD - 1) / BlockD)*(LWS[2] = cd / vs);
	//if( (!MyId) && (G_Settings.LogLevel&LOG_INFO) ) Print();
};

void MRange3::Print(void) const
{
	printf("\n%s: %i*%i*%i => %i*%i*%i / %i*[%i*%i*%i] -%i,%i,%i; blk=[%i*%i*%i] ",
				 name, Wid, Hei, Dep, GWS[0], GWS[1], GWS[2], VectorSize, LWS[0], LWS[1], LWS[2], OffsX, OffsY, OffsZ, BlockW, BlockH, BlockD);
}

int MRange3::ExecWait(cl_command_queue& cq, cl_kernel& k, cl_event* pE) const
{
	int err = clEnqueueNDRangeKernel(cq, k, 3, NULL, GWS, LWS, 0, NULL, pE);
	if ( err )  printf("%s Exec3 '%s' Error: %s \x07", _PID, name, ClErrorName(err));
	else  if ( !pE ) clFinish(cq);
	return err;
}

int MRange3::ExecNoWait(cl_command_queue& cq, cl_kernel& k) const
{
	int err = clEnqueueNDRangeKernel(cq, k, 3, NULL, GWS, LWS, 0, NULL, NULL);
	if ( err )  printf("%s Exec3 '%s' Error: %s \x07", _PID, name, ClErrorName(err));
	return err;
}


cl_int CLBuffer::Release(void)
{
	cl_int ret = 0;  if ( Mem ) ret = clReleaseMemObject(Mem);
	Len = Addr = 0;  Status = -1; Mem = NULL;
	return ret;
}


cl_int CLBuffer::AllocBuffer(cl_context& cx, cl_command_queue& cq, cl_kernel& kGetAddr)
{
	if ( !Len ) return 0;
	cl_int cl_error = AllocateGPUBuffer(cx, Len, cq, kGetAddr, &Mem, &Addr);
	Status = cl_error;
	if ( !cl_error ) return 0;
	Mem = NULL;  Len = 0;  Addr = 0;
	printf("\nAllocBuffer[%u Mb] failed (%i - %s) ", (unsigned int)(Len >> 20),
				 cl_error, ClErrorName(cl_error));
	return cl_error;
}

cl_int CLBuffer::AllocImage(cl_image_format* pFmt, cl_uint4& FwdDim, cl_context& cx, cl_command_queue& cq)
{
	cl_int cl_error = AllocateGPUImage(cx, cq, pFmt, FwdDim, &Mem);
	Status = cl_error;
	Addr = 0;
	if ( !cl_error ) return 0;
	Mem = NULL;  Len = 0;  Addr = 0;
	printf("\nAllocImage[%i*%i*%i] failed (%i - %s) ", (int)FwdDim.s[0], (int)FwdDim.s[1], (int)FwdDim.s[2],
				 cl_error, ClErrorName(cl_error));
	return cl_error;
}

void CLBuffer::Dump(const char* name, cl_command_queue& cq, cl_ulong start, cl_ulong l) const
{
	if ( !Addr || (start > Len) ) return;

	size_t bl, total = 0, len = Len - start;
	if ( l ) len = __min(Len, l);

	size_t BLKSZ = __min(len, 0x100000);
	memblk tmp(BLKSZ);
	if ( !tmp.data ) return;

	FILE* fp = fopen(name, "wb");
	if ( !fp ) return;
	IFLOGFILE  printf("\n GPU Dump @0x%012llX [0x%012llX] -> '%s' ", Addr + start, len, name);
	while ( len )
	{
		bl = BLKSZ;
		if ( bl > len ) bl = len;
		len -= bl;
		clEnqueueReadBuffer(cq, Mem, CL_TRUE, start, bl, tmp.data, 0, NULL, NULL);
		start += bl;
		fwrite(tmp.data, 1, bl, fp);
		total += bl;
	}
	fclose(fp);
}




void DumpGPU(const char* name, cl_mem& membuf, size_t start, size_t len, cl_command_queue& cq, float Amp)
{
	const size_t BLKSZ = 0x100000;
	size_t bl, total = 0;
	memblk tmp(BLKSZ);
	if ( !tmp.data ) return;

	FILE* fp = fopen(name, "wb");
	if ( !fp ) return;
	while ( len )
	{
		bl = BLKSZ;
		if ( bl > len ) bl = len;
		len -= bl;
		clEnqueueReadBuffer(cq, membuf, CL_TRUE, start, bl, tmp.data, 0, NULL, NULL);
		if ( Amp != 0.0f ) for ( int I = (bl >> 2) - 1; I >= 0; --I ) ((float*)tmp.data)[I] *= Amp;
		start += bl;
		fwrite(tmp.data, 1, bl, fp);
		total += bl;
	}
	fclose(fp);
	tmp.free();
	IFLOGFILE  printf("%s *DumpGPU: %s %lli* ", _PID, name, total);
}


cl_int SumupGPUCheck(int GpuMemGuard, cl_int* src, const char* str)
{
	cl_int sum = 0;
	for ( int i = (GpuMemGuard >> 2) - 1; i >= 0; --i ) sum |= src[i];
	if ( sum ^= 0x80000000 ) printf("%s GPU GUARDZONE ERROR %s: %08X ", _PID, str ? str : "", (int)sum);
	return sum;
}

static inline void check(void) {};


#endif 
// TARGET_GPU