#pragma once

#ifndef TARGET_GPU

typedef int GPUContext;
typedef int CLBuffer;

#else

char* printkernel(cl_kernel k);
cl_int SumupGPUCheck(int guard, cl_int* src, const char* str = NULL);
void DumpGPU(const char* name, cl_mem& membuf, size_t start, size_t len, cl_command_queue& cq, float Amp = 0.0f);
char* LoadProgSource(const char* header, const char* text, size_t* len, const char *addstr=NULL);
const char* ClErrorName(cl_int n);

extern cl_int cl_error;
extern const char* cl_errmsg;


cl_int AllocateGPUImage(cl_context& cx, cl_command_queue& cq, cl_image_format* pFmt, cl_uint4& FwdDim, cl_mem *retImg);
cl_int AllocateGPUImage2D(cl_context& cx, cl_command_queue& cq, cl_image_format* pFmt, cl_uint4& FwdDim, cl_mem *retImg);
cl_int AllocateGPUBuffer(cl_context& cx, cl_ulong TotalMemReq, cl_command_queue& cq, cl_kernel& kGetAddr, cl_mem *retBuf, cl_ulong* retAddr = NULL);


struct MRangeBase
{
	int dim;
	size_t GWS[3], LWS[3];
	size_t TotalItems(void) const;
	size_t TotalGroups(void) const;
};

struct Kernel
{
	cl_kernel k;  cl_int err;
	cl_uint n_run; cl_event evt;
	cl_ulong st, en, elapsed_ns, overlap_ns, min_ns, max_ns;
	char name[48];

	inline Kernel() { err = 0; evt = NULL; elapsed_ns = overlap_ns = 0; max_ns = 0; min_ns = -1; n_run = 0; };
	inline Kernel(cl_kernel K) { k = K; evt = NULL; elapsed_ns = overlap_ns = 0; max_ns = 0; min_ns = -1; }
	Kernel(cl_program cpProgram, const char* kname);

	cl_ulong Profile(bool wait = false, cl_ulong intersect_st = 0, cl_ulong intersect_en = 0);
	void Release(void);
	inline cl_int Read(cl_command_queue& cq, cl_mem& cm, cl_bool blocking, size_t pos, size_t sz, void* ptr);
	inline cl_int Write(cl_command_queue& cq, cl_mem& cm, cl_bool blocking, size_t pos, size_t sz, void* ptr);
	inline operator cl_kernel&() { return k; }
	cl_int ExecWait(cl_command_queue& cq, MRangeBase& NDR);
	cl_int ExecNoWait(cl_command_queue& cq, MRangeBase& NDR, bool use_event = true);
	inline cl_int ExecNoWait(cl_command_queue& cq, MRangeBase& NDR, cl_event* wait);
};

struct KernelArray
{
	Kernel *pK; Kernel NullK; 
	cl_ulong CPUStartTime;
	cl_uint used;
	KernelArray() { pK = NULL; NullK = Kernel(NULL); used = 0; CPUStartTime = 0; }
	cl_int GetErrors(void) const;
	cl_int Init(const char* const * name, cl_program cpProgram);
	KernelArray(char** name, cl_program cpProgram) { Init(name, cpProgram); }
	cl_int SetArg(cl_uint krn, cl_uint argn, size_t argsz, const void* argptr) const;
	float TotalElapsed(int sel = 1) const;
	float PrintStats(bool reset = false) const;
	void Release(void);
	~KernelArray() { Release(); }
	Kernel& operator[](unsigned int n) { return ((n >= used) ? NullK : pK[n]); }
};

struct GPUContext :KernelArray
{
	static constexpr int  MAX_PLATFORMS = 8, MAX_GPU_COUNT = 32;

	cl_command_queue CLQueue[MAX_GPU_COUNT] = { 0 };
	cl_command_queue CLQueu2[MAX_GPU_COUNT] = { 0 };
	cl_device_id 	cxDevices[MAX_GPU_COUNT] = { 0 };
	//cl_ulong GPULocMemSize[MAX_GPU_COUNT] = { 0 };

	size_t MaxAllocationSize = 0,		LocalMemSize = 0, GlobalMemSize = 0,
	MaxBlockSize = 0;

	cl_context CLContext = 0;
	int cxDeviceCount = 0;
	cl_program cpProgram = 0;
	bool fReady = false;

	cl_sampler LinClampSampler, NearClampSampler, LinNormSampler;

	cl_int Init(int CurPlatform, int GPU_ID, const char* const* KA_List, const char *src1, const char *src2);
	cl_int Close(void);
	GPUContext() { cxDeviceCount = 0; CLContext = 0; };
};

struct MRange :MRangeBase
{
	cl_int Wid, Hei, // aligned array
		WorkW, WorkH, // working part, can be odd
		CellW, CellH,
		BlockW, BlockH, VectorSize,
		OffsX, OffsY;
	const char* name;

	MRange() { Init("NULL", 0, 0, 0, 0, 1, 1, 0, 1); }

	MRange(const char* nm, int w, int h, int ww, int wh, int cw, int ch, int halo, int vs, bool print = false)
	{ Init(nm, w, h, ww, wh, cw, ch, halo, vs); if ( print ) Print(); }

	void Init(const char* nm, int w, int h, int ww, int wh, int cw, int ch, int halo, int vs);
	void Print(void) const;
	int ExecWait(cl_command_queue& cq, Kernel& K, cl_event* pE = NULL) const;
	int ExecNoWait(cl_command_queue& cq, cl_kernel& k) const;
};

struct MRange3 :MRangeBase
{
	cl_int Wid, Hei, Dep, // aligned array
		WorkW, WorkH, WorkD,// working part, can be odd
		CellW, CellH, CellD,
		BlockW, BlockH, BlockD, VectorSize,
		OffsX, OffsY, OffsZ;
	const char* name;

	void Init(const char* nm, int w, int h, int d, int ww, int wh, int wd, int cw, int ch, int cd, int halo, int vs);
	MRange3() { Init("NULL", 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1); }
	MRange3(const char* nm, int w, int h, int d, int ww, int wh, int wd, int cw, int ch, int cd, int halo, int vs, bool print = false)
	{ Init(nm, w, h, d, ww, wh, wd, cw, ch, cd, halo, vs); if ( print ) Print(); }
	void Print(void) const;
	int ExecWait(cl_command_queue& cq, cl_kernel& K, cl_event* pE = NULL) const;
	int ExecNoWait(cl_command_queue& cq, cl_kernel& k) const;
};


struct CLBuffer
{
	cl_mem Mem;
	cl_ulong Len, Addr;
	cl_long Status;

	inline operator cl_mem&() { return Mem; }
	inline CLBuffer() { Mem = NULL; Len = 0; Addr = 0; Status = -1LL; }
	cl_int AllocBuffer(cl_context& cx, cl_command_queue& cq, cl_kernel& kGetAddr);
	cl_int AllocImage(cl_image_format* pFmt, cl_uint4& FwdDim, cl_context& cx, cl_command_queue& cq);

	cl_int Release(void);
	void Dump(const char* name, cl_command_queue& cq, cl_ulong start = 0, cl_ulong l = 0) const;

	inline bool includes(cl_ulong addr, cl_ulong size) const
	{ return (!Status && Mem && Addr && (addr >= Addr) && (addr < Addr + Len) && (addr + size <= Addr + Len)); }

	cl_int Xfer(bool Up, void* src, cl_ulong linear, cl_ulong size, cl_command_queue cq) const
	{
		cl_int ret = 0;
		if ( !size ) size = Len;
		if ( !linear ) linear = Addr;
		if ( !includes(linear, size) ) return CL_INVALID_VALUE;
		ret = Up ? clEnqueueWriteBuffer(cq, Mem, CL_FALSE, linear - Addr, size, src, 0, NULL, NULL)
			: clEnqueueReadBuffer(cq, Mem, CL_FALSE, linear - Addr, size, src, 0, NULL, NULL);
		if ( G_Settings.LogLevel&LOG_KRNL || ret )
			printf("%sGPU %sload: %i %s \n rel=0x%08llX bytes=%08llX [Len=0x%08llX] Host:0x%012llX GPU:0x%012llX ",
						 _PID, Up ? "up" : "down", ret, ClErrorName(ret), linear - Addr, size, Len, src, linear);
		clFinish(cq);
		return ret;
	}
	void Print(const char* name = NULL) const
	{
		printf("%s CLBuffer '%s' Addr=%012llX Len=%08llX ", _PID, name ? name : "", Addr, Len);
	}
}; // struct



#endif 