#pragma once

#include "k_defs.h"
#include "gpu.h"


#define SETARG0(name) clSetKernelArg((*pGPU)[name], 0, sizeof(cl_mem),(void*)&(GConstBuf))
#define CL_TRY(smth) { if(cl_error|=(smth)) goto cl_error_occured; }

#define DECLARE_KERNELS _K_(GetAddr), _K_(CheckData), _K_(CLS), \
_K_(BackwardWave44s),_K_(BackwardWave44pg),_K_(BackwardWave44), \
_K_(ForwardWave44), _K_(Initialize),_K_(SumParGrad),\
 _K_(SaveExData), _K_(LoadExData),_K_(LoadBound), _K_(CompareData),\
 _K_(NRBCond_Hrz), _K_(NRBCond_Vrt),\
_K_($LoadFromGPU), _K_($SaveToGPU),  _K_($SaveExData),  _K_($BufferOp)
//-end-macro

#define _K_(x) k_##x
enum { DECLARE_KERNELS };
#undef _K_

extern const char* const KA_List[];

class DataCache : public memblk
#ifdef TARGET_GPU
	, public CLBuffer
#endif
{
public:
	cl_uint LineWidthB = 0, LinesCommon = 0, LinesPerSrc = 0;

	DataCache() {}

	void Initialize(cl_uint LW, cl_uint LC, cl_uint LPS)
	{
		Free();
		LineWidthB = LW, LinesCommon = LC, LinesPerSrc = LPS;
	}

	inline cl_ulong BufSize(int Srcs)
	{ return (cl_ulong)LineWidthB * (LinesCommon + (cl_ulong)LinesPerSrc * Srcs); }


	cl_int Alloc(int HostSrcs, const char* name = NULL)
	{
		size_t l = BufSize(HostSrcs);
		return (l ? (memblk::alloc(l, name) ? 0 : -1) : 0);
	}


	cl_int Free(void)
	{
		memblk::free();
#ifdef TARGET_GPU
		return CLBuffer::Release();
#else
		return 0;
#endif
	}


	inline cl_ulong offset(int src, cl_uint line = 0) const
	{
		return  LineWidthB * (line + ((src < 0) ? 0 : (LinesCommon + LinesPerSrc * src)));
	}

	inline char* HostAddr(int src, cl_uint line = 0) const
	{
		cl_ulong o = offset(src, line);
		return (((o + LineWidthB) <= memblk::len) ? (data + o) : NULL);
	}

#ifdef TARGET_GPU
	inline cl_ulong GPUAddr(int src, cl_uint line = 0) const
	{
		if ( !CLBuffer::Addr ) return 0;
		cl_ulong o = offset(src, line);
		return (o + LineWidthB) < CLBuffer::Len ? CLBuffer::Addr + o : NULL;
	}


	cl_int AllocateOnGPU(GPUContext* ctx, int GPUSrcs)
	{
		ENTER;
		cl_int cl_error = 0;
		if ( 0 == (CLBuffer::Len = BufSize(GPUSrcs)) ) return 0;

		CL_TRY(CLBuffer::AllocBuffer(ctx->CLContext, ctx->CLQueue[0], (*ctx)[k_GetAddr]));
		IFLOG(LOG_GPU) CLBuffer::Print();
		if ( LinesCommon && data )
		{ CL_TRY(CLBuffer::Xfer(true, data, Addr, LinesCommon*LineWidthB, ctx->CLQueue[0])); }

		PRINT_CL_ERROR;
		RETURN(cl_error);
	}

	cl_int  TransferGPU(int src, bool upload, cl_command_queue& cq, char* external = NULL) const
	{
		size_t startGPU, startHost, amt;
		char* ptr = external ? external : data;
		if ( !ptr ) { printf("\n Buffer not allocated ");  return CL_INVALID_BUFFER_SIZE; }

		if ( src < 0 ) { startGPU = startHost = 0; amt = LinesCommon; } else { startGPU = src * LinesPerSrc + LinesCommon; startHost = LinesCommon;  amt = LinesPerSrc; }
		startGPU *= LineWidthB; startHost *= LineWidthB; amt *= LineWidthB;
		if ( (startGPU + amt > CLBuffer::Len) || (startHost + amt > memblk::len) )
		{ printf("\n Transfer out of range"); return CL_INVALID_ARG_VALUE; }
		return CLBuffer::Xfer(upload, ptr, CLBuffer::Addr + startGPU, amt, cq);
	}
#endif

	void Dump(const char* name, int src, cl_command_queue* pCQ = NULL) const
	{
		cl_ulong offs = offset(src, 0), amt = (src < 0 ? LinesCommon : LinesPerSrc) * LineWidthB;
		IFLOG(LOG_FILES) printf("%s DataCache Dump src=%i '%s' %uMb ", _PID, src, name, (unsigned int)(amt >> 20));
#ifdef TARGET_GPU
		if ( pCQ )
		{
			if ( CLBuffer::Status || !CLBuffer::Addr || (CLBuffer::Len < (offs + amt)) )
			{ fprintf(stderr, "\nDataCache::Dump GPU address out of range! "); 	return; }
			CLBuffer::Dump(name, *pCQ, offs, amt);
		} else
#endif
		{
			if ( (!data) || (memblk::len < (offs + amt)) )
			{ fprintf(stderr, "\nDataCache::Dump address out of range! "); 	return; }
			FWrite(name, data + offs, amt);
		}
	}
	void Print(void) const
	{
		memblk::Print();
#ifdef TARGET_GPU
		CLBuffer::Print(name);
#endif
	}
};


struct BlockSizes
{
	int FwdY, BwdY, FwdW, BwdW;
	void SetDefaults()
	{
		if ( FwdY <= 0 ) FwdY = 20;	 if ( BwdY <= 0 ) BwdY = 20;
		if ( FwdW <= 0 ) FwdW = 64; 	if ( BwdW <= 0 ) BwdW = 64;
	}
};

struct GPU_DATA :CONST_DATA
{ // POD data structure. Included in a larger struct. No virtuals.	
	cl_mem   GConstBuf=NULL;
	CLBuffer  GCoeffBuf, GLayerBuf, GParGrad;
	DataCache BoundCh, ExData;
	BlockSizes BLKs= { 20,20,64,64 };

	cl_uint MemoryShear[6] = { 0,128,256,384,512,768 }; // bytes. Will crash AVX if less that CPU vector size
	cl_uint MemAlign = 128; // bytes
	//cl_uint MaxParGradSize = 8000; // pixels

	GPU_DATA();
	~GPU_DATA() { Release(); }

	cl_int Release(void);
	void Print(int n = 0) const;

#ifdef TARGET_GPU
	GPUContext *pGPU;
	cl_int AllocateOnGPU(void);
	cl_int UploadToGPU(void) const;

#endif // ifdef TARGET_GPU

};


