// Based on Catch and Doctest open-source frameworks
// Implements a subset of features not depending on the STL.

#define UNIT_TEST

#ifndef CATCH_CONFIG_RUNNER
#include "minicatch.hpp"
namespace Catch {
#endif

#define CATCH_TEST_PASSED_MODE ((mode<2) || ((mode>4)&&(mode<8)))
#define CATCH_TEST_FAILED_MODE ((mode>=2) && (mode<=4))
#define CATCH_TEST_REENTER_MODE (mode<4)

	Context BaseContext;
	Context *CurrentContext = &BaseContext;

	TestCase *  GlobalTail = NULL;
	TestCase *  CurrentTest = NULL;
	char LastLoggedMsg[256];


	TestCase::TestCase(TestFunction _pfn, const char* _file, int _line, const char* _name, const char* _desc)
		: pFn(_pfn), SectionData(_file, _line, _name, _desc)
	{
		//if ( !GlobalTail ) Abort("TestCase List tail is NULL", _file, _line);
		pNext = GlobalTail;
		GlobalTail = this;
		Duration = 0;
		passed = executed = passed_chk = executed_chk = 0;
		mode = MD_SKIPPED;
		if ( desc )
			mode |= ((MD_MAY_FAIL & (strstr(desc, "[!mayfail]") ? -1 : 0)) |
			(MD_SHOULD_FAIL & (strstr(desc, "[!shouldfail]") ? -1 : 0)));
		SectionData::Print(); printf("<- Test added.");
	}

	int TestCase::invoke(void)
	{
		int ret = 0;
		sprintf(failmsg, "No function");
		LastLoggedMsg[0] = 0;
		if ( !pFn ) return ret;
		*failmsg = 0;
		CurrentTest = this;
		CurrentContext->currentSubcaseLevel = 0;
		Start = CATCH_CLOCK::now();
		Duration = 0;
		++CurrentContext->TestCasesRun;
		mode &= ~MD_SKIPPED;
		do
		{
			CurrentContext->should_reenter = false;
			CurrentContext->subcasesCurrentMaxLevel = 0;
			CurrentContext->subcasesStack.clear();
			try { pFn(); }
			catch ( const std::exception &exc )
			{
				mode |= MD_ERRORED; ++CurrentContext->Errors;
				snprintf(failmsg, sizeof(failmsg), "%s |> Exception:%s", LastLoggedMsg, exc.what());
			}
		} while ( CurrentContext->should_reenter && CATCH_TEST_FAILED_MODE );
		std::chrono::duration<double> D = CATCH_CLOCK::now() - Start;
		if ( CATCH_TEST_FAILED_MODE ) ++CurrentContext->Failures;
		if ( CATCH_TEST_PASSED_MODE ) ++CurrentContext->Successes;
		Duration = D.count() * 1000000.0;
		PrintInTest();
		if ( CurrentContext->Failures >= CurrentContext->MaxFailures )
		{ printf(" %i TESTS FAILED - STOP", CurrentContext->Failures); return CurrentContext->Failures; } else return 0;
	}

	bool TestCase::Log(const char* expr, const char* file, int line, int code)
	{
		static constexpr int MAX_RETCODES = 8;
		static const char* retcodes[MAX_RETCODES] = { "SUCCESS","FAIL","CHECK","CHECK_FAIL","CODE 4","CODE 5","CODE 6","CODE 7" };


		if ( code&RES_FAIL )
		{
			mode |= MD_FAILED;
			snprintf(LastLoggedMsg, sizeof(LastLoggedMsg) - 1, "%5i:  '%s' -> %s", line, expr, retcodes[code&(MAX_RETCODES - 1)]);
		}
		bool Failed = CATCH_TEST_FAILED_MODE;

		if ( (BaseContext.Mode&Context::F_REPORT_SUCCESS) || (code&RES_FAIL) )
			printf("\n%s %5i:  '%s' -> %s ", file, line, expr, retcodes[code&(MAX_RETCODES - 1)]);

		if ( !failmsg[0] ) strncpy(failmsg, LastLoggedMsg, sizeof(failmsg) - 1);

		switch ( code )
		{
			case RES_SUCCESS: ++passed;     case RES_FAIL:  ++executed; 		        break;
			case RES_CHECK:   ++passed_chk; case RES_CHECK_FAIL: ++executed_chk; 	 break;
		}
		return Failed;
	}

	int TestCase::PrintInTest(void)
	{
		bool failed = CATCH_TEST_FAILED_MODE;
		if ( failed || (CurrentContext->Mode & (Context::F_REPORT_SUCCESS | Context::F_REPORT_DURATION))
				|| (Duration >= CurrentContext->MinReportDuration) )
		{
			Print();
			printf("  Req: %i/%i, Chk: %i/%i Elapsed %i ms \n*** %s *** %s", passed, executed, passed_chk, executed_chk,
				(int)(Duration / 1000), failed ? "FAILED" : "PASSED", failed ? failmsg : "");
			//printf("\nthis %llX next %llX  ", (long long int)this, (long long int)pNext);
			return failed;
		}
		return 0;
	}

	void TestCase::PrintXML(FILE* fp)
	{
		fprintf(fp, "<testcase classname = \"%s\" name = \"%s\" time = \"%f\">\n", file, desc, (float)Duration*1.0e-6f);
		if ( mode&MD_SKIPPED ) fprintf(fp, "<skipped />\n");
		else
			if ( mode&MD_ERRORED ) 	fprintf(fp, "<error message =\"%s\">%s</error>\n", failmsg, "Exception caught");
			else
				if ( CATCH_TEST_FAILED_MODE )
				{
					const char* failreason = " ";
					switch ( mode )
					{
						case 2: case 3: failreason = "Test should fail but didn't";
						case 4: failreason = "Assertion failed";
					}
					fprintf(fp, "<failure message =\"%s\">%s</failure>\n", failmsg, failreason);
				}
		fprintf(fp, "</testcase>\n");
	}


	Subcase::Subcase(const char* _file, int _line, const char* _name, const char* _desc)
		: SectionData(_file, _line, _name, _desc)
	{
					// if a Subcase on the same level has already been entered
		if ( CurrentContext->subcasesStack.size() < size_t(CurrentContext->subcasesCurrentMaxLevel) )
		{
			CurrentContext->should_reenter = true;
			return;
		}

		// push the current signature to the stack so we can check if the
		// current stack + the current new subcase have been traversed
		CurrentContext->subcasesStack.push(*this);
		if ( CurrentContext->subcasesPassed.count(CurrentContext->subcasesStack) != 0 )
		{
// pop - revert to previous stack since we've already passed this
			CurrentContext->subcasesStack.pop();
			return;
		}

		CurrentContext->subcasesCurrentMaxLevel = CurrentContext->subcasesStack.size();
		m_entered = true;
		CurrentContext->currentSubcaseLevel++;

		//printf("\nSUBCASE ENTERED %i: ", CurrentContext->currentSubcaseLevel);
		Print();
	}

	Subcase::~Subcase()
	{
		if ( m_entered )
		{
	// only mark the subcase stack as passed if no subcases have been skipped
			if ( CurrentContext->should_reenter == false )
				CurrentContext->subcasesPassed.insert_unique(CurrentContext->subcasesStack);
			CurrentContext->subcasesStack.pop();
			CurrentContext->currentSubcaseLevel--;
			//printf("\nSUBCASE EXITED %i: ", CurrentContext->currentSubcaseLevel);		Print();
		}
	}


	void Context::ReportJUnit(char* name) const
	{
		char tmp[64];
		strftime(tmp, 63, "", localtime(&Timestamp));
		std::chrono::duration<float> duration = CATCH_CLOCK::now() - Start;
		FILE* XmlLog = fopen("junit.xml", "wb"); 		if ( !XmlLog )return;
		fprintf(XmlLog, "<?xml version=\"1.0\" encoding=\"UTF - 8\"?>\n<testsuites>\n"
						"<testsuite name = \"minicatch_JUnitXml\" errors = \"%i\" tests = \"%i\" failures = \"%i\" time = \"%f\" timestamp = \"%s\" />\n",
						Errors, TestCasesRun, Failures, duration.count(), tmp);
		for ( TestCase* pT = GlobalTail; pT; pT = pT->pNext )	pT->PrintXML(XmlLog);
		fprintf(XmlLog, "</testsuite>\n</testsuites>\n");
		fclose(XmlLog);
	}



	/*
	<testcase classname = "JUnitXmlReporter.constructor" name = "should default path to an empty string" time = "0.006">
		<failure message = "test failure">Assertion failed< / failure>
		</testcase>
		<testcase classname = "JUnitXmlReporter.constructor" name = "should default consolidate to true" time = "0">
		<skipped / >
		< / testcase>
		<testcase classname = "JUnitXmlReporter.constructor" name = "should default useDotNotation to true" time = "0" / >
		*/



/*

<?xml version="1.0" encoding="UTF-8"?>
<testsuites>
	 <testsuite name="JUnitXmlReporter" errors="0" tests="0" failures="0" time="0" timestamp="2013-05-24T10:23:58" />
	 <testsuite name="JUnitXmlReporter.constructor" errors="0" skipped="1" tests="3" failures="1" time="0.006" timestamp="2013-05-24T10:23:58">
			<properties>
				 <property name="java.vendor" value="Sun Microsystems Inc." />
				 <property name="compiler.debug" value="on" />
				 <property name="project.jdk.classpath" value="jdk.classpath.1.6" />
			</properties>
			<testcase classname="JUnitXmlReporter.constructor" name="should default path to an empty string" time="0.006">
				 <failure message="test failure">Assertion failed</failure>
			</testcase>
			<testcase classname="JUnitXmlReporter.constructor" name="should default consolidate to true" time="0">
				 <skipped />
			</testcase>
			<testcase classname="JUnitXmlReporter.constructor" name="should default useDotNotation to true" time="0" />
	 </testsuite>
</testsuites>

<testsuites>        => the aggregated result of all junit testfiles
	<testsuite>       => the output from a single TestSuite
		<properties>    => the defined properties at test execution
			<property>    => name/value pair for a single property
			...
		</properties>
		<error></error> => optional information, in place of a test case - normally if the tests in the suite could not be found etc.
		<testcase>      => the results from executing a test method
			<system-out>  => data written to System.out during the test run
			<system-err>  => data written to System.err during the test run
			<skipped/>    => test was skipped
			<failure>     => test failed
			<error>       => test encountered an error
		</testcase>
		...
	</testsuite>
	...
</testsuites>

*/

	bool Filtered(SectionData* pS, int argc, const char* const* argv)
	{
		bool Enabled = ((pS->desc[0] != '.') || (CurrentContext->Mode&Context::F_ENABLE_HIDDEN));
		int Default = 0, ret = 0;
		if ( !argv ) argc = 0;	if ( argc <= 0 ) return Enabled;
		for ( int i = 0; i < argc; ++i )
		{
			if ( !argv[i] ) break;
			if ( argv[i][0] == '~' )
			{	// exclusive arg			
				if ( !Default ) Default = ret = 1;
				if ( strstr(pS->desc, argv[i] + 1) ) ret = 2; // exclude any item containing this str												
			} else
			{	// inclusive arg			
				if ( !Default ) Default = ret = 2;
				if ( (strstr(pS->desc, argv[i])) &&
					(Enabled || (argv[i][0] == '.')) ) ret = 1;
			}		// inclusive arg					
		} // for args
		if ( !ret ) ret = Default;
		return(ret == 1);
	}

	int Session::RunFilteredFunction(int argc, const char* const* argv, int(TestCase::*pFn)(void))
	{
		int ret = 0;
		TestCase* pT = GlobalTail;// BaseTest.pNext;
		while ( pT && !ret )
		{
			if ( Filtered(pT, argc, argv) ) 	ret = (pT->*pFn)();
			pT = pT->pNext;
		}
		return ret;
	}


	int Session::Run(int argc, const char* const* argv)
	{
		bool bList = false;
		int args = 0;
		while ( ++args < argc )
		{
			if ( argv[args][0] == '-' )
				switch ( argv[args][1] )
				{
					case 'l': bList = true; break;
					case 'a': BaseContext.MaxFailures = 1; break;
					case 'b': BaseContext.Mode |= Context::F_BREAK; break;
					case 's': BaseContext.Mode |= Context::F_REPORT_SUCCESS; break;
					case 'd': BaseContext.Mode |= Context::F_REPORT_DURATION; break;
					case 'r': BaseContext.Mode |= Context::F_REPORT_JUNIT; break;
					case 'x': if ( ++args >= argc ) break; if ( !argv[args] ) break;
						BaseContext.MaxFailures = atoi(argv[args]);
						printf("\nMax failures set to %i ", BaseContext.MaxFailures);
						break;
					case 'D': if ( ++args >= argc ) break; if ( !argv[args] ) break;
						BaseContext.MinReportDuration = atol(argv[args]) * 1000;
						printf("\nMin report duration set to %i ms ", (int)(BaseContext.MinReportDuration / 1000));
						break;
					case 'h': case '?': default:
						puts("\nUsage: <exename> [-options] [*] [~][tests_to_run]"
								 "\n * - include hidden tests \n ~name - exclude tests containing name"
								 "\nOptions: \n -l list tests\n -a abort on failure\n -x N abort after N failures"
								 "\n -r report in JUnit.xml"
								 "\n -b breakpoint on failure\n -s report successes \n -d report durations"
								 "\n -D N report durations if longer than N ms"); return 0;
				} else if ( argv[args][0] == '*' ) BaseContext.Mode |= Context::F_ENABLE_HIDDEN;
				else break;
		}
		if ( bList ) return RunFilteredFunction(argc - args, argv + args, &TestCase::Print);
		else
		{
			int ret = RunFilteredFunction(argc - args, argv + args, &TestCase::invoke);
			if ( BaseContext.Mode&Context::F_REPORT_JUNIT ) BaseContext.ReportJUnit();
			BaseContext.Print();
			return ret;
		}
	}

#ifndef CATCH_CONFIG_RUNNER
};//namespace
#endif


/*

#ifdef UNIT_TEST

TEST_CASE("mayfail", "[!mayfail]") {	REQUIRE(0); }

TEST_CASE("mayfail2", "[!mayfail]") { 	REQUIRE(1); }

TEST_CASE("shouldfail", "[!shouldfail]") { 	REQUIRE(0); }

TEST_CASE("hidden", ".hidden")
{
	CHECK(1);
	throw new std::exception();
}

#endif

*/