#pragma once
#include <stdlib.h>
#include <stdio.h>
#include <cstring>
#include <ctime>
#include <chrono>
#include <exception>

#ifdef CATCH_CONFIG_MAIN
#define CATCH_CONFIG_RUNNER
#endif

#ifndef CATCH_SECTION_STACK_CAPACITY
#define CATCH_SECTION_STACK_CAPACITY 20
#endif

#ifndef CATCH_SECTION_MAP_CAPACITY
#define CATCH_SECTION_MAP_CAPACITY 256
#endif

#define CATCH_BREAKPOINT do { *(int*)(0)=0; } while(0)

#define CATCH_CLOCK std::chrono::steady_clock

#define CATCH_INTERNAL_LINEINFO  __FILE__, static_cast<int>( __LINE__ )
#define INTERNAL_CATCH_UNIQUE_NAME_LINE( name, line ) INTERNAL_CATCH_UNIQUE_NAME_LINE2( name, line )
#define INTERNAL_CATCH_UNIQUE_NAME_LINE2( name, line ) name##line

#ifdef CATCH_CONFIG_COUNTER
#  define INTERNAL_CATCH_UNIQUE_NAME( name ) INTERNAL_CATCH_UNIQUE_NAME_LINE( name, __COUNTER__ )
#else
#  define INTERNAL_CATCH_UNIQUE_NAME( name ) INTERNAL_CATCH_UNIQUE_NAME_LINE( name, __LINE__ )
#endif

#define TEST_CASE( Name, Desc ) INTERNAL_CATCH_TESTCASE( Name, Desc)

#define INTERNAL_CATCH_TESTCASE( Name, Desc ) INTERNAL_CATCH_TESTCASE2( INTERNAL_CATCH_UNIQUE_NAME( ____C_A_T_C_H___T_ ), INTERNAL_CATCH_UNIQUE_NAME( ____C_A_T_C_H___To_ ), Name, Desc )

#define SECTION( Name, Desc ) INTERNAL_CATCH_SECTION( Name, Desc )
#define INTERNAL_CATCH_SECTION( Name, Desc ) INTERNAL_CATCH_SECTION2( INTERNAL_CATCH_UNIQUE_NAME( ____C_A_T_C_H____S_ ), Name, Desc )


#define INTERNAL_CATCH_TESTCASE2( TestName, ObjName, Name, Desc ) \
        static void TestName(void); \
        static Catch::TestCase ObjName( TestName, CATCH_INTERNAL_LINEINFO,  Name, Desc ); \
        static void TestName(void)
//-end-macro


#define INTERNAL_CATCH_SECTION2( SecName, name, desc )  if(  const Catch::Subcase & SecName = Catch::Subcase(CATCH_INTERNAL_LINEINFO, name, desc) )


#define REQUIRE( expr )   {  if(expr) Catch::CurrentTest->Log(#expr,CATCH_INTERNAL_LINEINFO,Catch::RES_SUCCESS); \
                       else if (Catch::CurrentTest->Log(#expr,CATCH_INTERNAL_LINEINFO,Catch::RES_FAIL)) \
                          { if (Catch::CurrentContext->Mode&Catch::Context::F_BREAK) CATCH_BREAKPOINT; else return; } }
//-end-macro

#define CHECK( expr )   {  Catch::CurrentTest->Log(#expr,CATCH_INTERNAL_LINEINFO,(expr)? Catch::RES_CHECK : Catch::RES_CHECK_FAIL); }
//-end-macro

#define SCENARIO( name, tags ) TEST_CASE( "Scenario: " name, tags )
#define GIVEN( desc )    SECTION(  "Given: "  desc, "" )
#define WHEN( desc )     SECTION(  " When: "  desc, "" )
#define AND_WHEN( desc ) SECTION(  "  And: "  desc, "" )
#define THEN( desc )     SECTION(  " Then: "  desc, "" )
#define AND_THEN( desc ) SECTION(  "  And: "  desc, "" )


#define CATCH_DECLARE_STD_OPER(name)		name& operator=(const name& that) = default;\
		  name& operator=(name&& that) = default; \
		  name(const name& that) = default; \
		  name(name&& that) = default;
//-end-macro

namespace Catch {

	template <class T, int Cap> 	class FixedStack
	{
		T Data[Cap];
		int len;

	public:
		inline int size(void) const { return len; };
		inline void clear(void) { memset(Data, 0, Cap * sizeof(T)); len = 0; }
		inline FixedStack<T, Cap>() { clear(); }

		inline int compare(const FixedStack<T, Cap>& that) const
		{ return memcmp(this, &that, sizeof(FixedStack<T, Cap>)); }

		inline void push(const T& ref)
		{
			if ( len < 0 ) len = 0;
			if ( len < Cap - 1 ) { memmove(Data + len, &ref, sizeof(T)); ++len; }
		}

		inline void pop(void)
		{ if ( len > 0 ) --len; else len = 0;  memset(Data + len, 0, sizeof(T)); }

		inline void pop(T* s)
		{
			if ( len <= 0 ) len = 1;
			memmove(s, Data + len - 1, sizeof(T));
			memset(Data + len - 1, 0, sizeof(T));  --len;
		}

		int count(const T& ref) const
		{
			int ret = 0;
			for ( int i = 0; i < len; ++i )
				ret += (memcmp(Data + i, &ref, sizeof(T)) ? 0 : 1);
			return ret;
		}

		inline bool insert_unique(const T& ref)
		{ if ( count(ref) ) return false; else push(ref); return true; }
#define stname_n_ FixedStack<T,Cap>
		CATCH_DECLARE_STD_OPER(stname_n_);
#undef stname_n_
	};


	static constexpr int RES_SUCCESS = 0, RES_FAIL = 1, RES_CHECK = 2, RES_CHECK_FAIL = 3;

	typedef void(*TestFunction)();

	static inline void Abort(const char* msg, const char*_file, int _line)
	{ printf("\nCatch aborted: %s : %s : %i ", msg, _file, _line);  exit(1); }


	struct SectionData
	{
		const char* name;
		const char* desc;
		const char* file;
		int line;

		SectionData(const char* _file, int _line, const char* _name = "", const char* _desc = "")
			: name(_name), desc(_desc), file(_file), line(_line) {}

		SectionData() {};

		int Print(void) const
		{
			printf("\n%s %5i: %s %s ", file ? file : "(NULL)", line, name ? name : "(NULL)", desc ? desc : "(NULL)");
			return 0;
		}

		CATCH_DECLARE_STD_OPER(SectionData);
	};

	typedef FixedStack <SectionData, CATCH_SECTION_STACK_CAPACITY> SubcaseStack;
	typedef FixedStack <SubcaseStack, CATCH_SECTION_MAP_CAPACITY>  StackList;


	struct Context
	{
		static constexpr int F_REPORT_SUCCESS = 1, F_REPORT_DURATION = 2, F_ENABLE_HIDDEN = 4, F_REPORT_JUNIT=8, F_BREAK = 0x10;

		SubcaseStack subcasesStack;
		StackList subcasesPassed;

		int subcasesCurrentMaxLevel = 0;
		int currentSubcaseLevel = 0;

		bool should_reenter = false;

		unsigned int Mode = 0, MaxFailures = 10,
			TestCasesRun = 0, Successes = 0, Failures = 0, Errors = 0, MinReportDuration = 10000000;

		std::time_t Timestamp;
		std::chrono::time_point<CATCH_CLOCK> Start;

		Context()
		{
			auto time_point = std::chrono::system_clock::now();
			Timestamp = std::chrono::system_clock::to_time_t(time_point);
			Start = CATCH_CLOCK::now();
		}

		void ReportJUnit(char* name = NULL) const;

		int Print(void) const { printf("\n*** Tests: %i; pass: %i; fail: %i; Error: %i", 
																	 TestCasesRun, Successes, Failures, Errors); return 0;}

	};


	struct TestCase :SectionData
	{
		static constexpr int MD_MAY_FAIL = 1, MD_SHOULD_FAIL = 2, MD_FAILED = 4, MD_ERRORED = 8, MD_SKIPPED=16, MD_MASK = 0x1F;

		TestFunction pFn;
		unsigned int mode, passed, executed, passed_chk, executed_chk;
		std::chrono::time_point<CATCH_CLOCK> Start;
		size_t Duration;
		char failmsg[256];
		TestCase *pNext;

		TestCase(TestFunction _pfn, const char* _file, int _line, const char* _name = "", const char* _desc = "");
		int invoke(void);
		bool Log(const char* expr, const char* file, int line, int code);
		int PrintInTest(void);
		void PrintXML(FILE* fp);
		inline int Print(void) { return SectionData::Print(); }
	};


	struct Subcase :SectionData
	{
		bool   m_entered = false;
		operator bool() const { return m_entered; }
		Subcase(const char* _file, int _line, const char* _name = "", const char* _desc = "");
		~Subcase();
		CATCH_DECLARE_STD_OPER(Subcase);
	};


	struct Session
	{
		Session() {}
		int RunFilteredFunction(int argc, const char* const* argv, int(TestCase::*pFn)(void));
		int Run(int argc, const char* const* argv);
	};

	extern TestCase *  GlobalTail;
	extern char LastLoggedMsg[256];
	extern TestCase *  CurrentTest;
	extern Context BaseContext;
	extern Context * CurrentContext;
	extern bool Filtered(SectionData* pS, int argc, const char* const* argv);

#ifdef CATCH_CONFIG_RUNNER

#include "minicatch.cpp"

#endif

}; // end namespace


#ifdef CATCH_CONFIG_MAIN
int main(int argc, char** argv)
{
	return Catch::Session().Run(argc, argv);
}
#endif
