
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <omp.h>

#include "k_defs.h"
#include "parallel.h"
#include "rdtsc.h"


void Parallel::ResetOne(void)
{
	nRuns = 0;
	TimeSpent = MaxTime = MinTime = 0;

	if ( !pObj ) return;

	if ( pObj->el_wid && (Width % pObj->el_wid) )
	{
		printf("\nParallel::Reset @0x%llX: Width=%i non-divisible by el_wid of: ",
			(cl_ulong)this, Width); pObj->Print();
	}
}

int Parallel::Reset(void)
{
	int ret = 0;
	for ( Parallel* ptr = this; ptr->pObj; ++ptr )
	{
		ptr->ResetOne(); ++ret;
		if ( ptr->Mode&Parallel::RUN_LAST_ITEM ) break;
	}
	return ret;
}

int Parallel::Print(int level)
{
	int ret = 0;
	if ( level > 0 ) printf("\n Runs   Avg    Min    Max    Total   Fn  ");
	for ( Parallel* ptr = this; ptr->pObj; ++ptr )
	{
		if ( level > 0 )
			printf("\n%6u  %8llu  %8llu %8llu  %8llu %c %s.%i ",
						 ptr->nRuns, ptr->nRuns ? (ptr->TimeSpent / ptr->nRuns) : 0LL,
						 ptr->MinTime, ptr->MaxTime, ptr->TimeSpent,
						 (ptr->Mode&Parallel::RUN_ASYNC) ? 'A' : ' ', ptr->pObj->Name, ptr->el_wid);
		else
		{
			ptr->pObj->Print();
			printf(" .%c [%4i/%3i * %4i/%3i]", (ptr->Mode&Parallel::RUN_ASYNC) ? 'A' : 'S',
						 ptr->Width, ptr->CellX, ptr->Height, ptr->CellY);
		}
		++ret;
		if ( ptr->Mode&Parallel::RUN_LAST_ITEM ) break;
	}

	return ret;
}

int Parallel::Run(int RunMethod, int T0, int NSteps)
{
	static constexpr int STRIDE = sizeof(float);
	bool fPrint = RunMethod & RUN_PRINT,
		fHyper2 = RunMethod & RUN_HYPER2;
	int  Hyper = (RunMethod & RUN_HYPER) ? 1 : 0;
	int StepDir = (NSteps < 0) ? -1 : 1;		NSteps *= StepDir;
	int NThreads = omp_get_max_threads();
	if ( NThreads & 1 ) Hyper = 0;

	cl_ulong BaseTime = Rdtsc();

	for ( Parallel* ptr = this; ptr->pObj; ++ptr )
	{
		if ( !(ptr->pObj) ) ptr->el_wid = 0; else  ptr->el_wid = ptr->pObj->el_wid;
		if ( ptr->Mode&RUN_LAST_ITEM ) break;
	}

	IFLOG(LOG_KRNL) Print(0);

	omp_set_num_threads(NThreads);
	volatile int BAR = NThreads, LOCK = 0;

#pragma omp parallel default(none), shared(NSteps,T0,BaseTime,Hyper,fHyper2,NThreads,StepDir,fPrint,BAR,LOCK)
	{ int LOC = 0, LocalT = T0,
		MyThread = omp_get_thread_num();
	if ( fHyper2 ) MyThread = ((MyThread+MyThread)% NThreads) +((MyThread>=(NThreads >> 1))? 1:0);
		int MyCore = (MyThread >> Hyper), MyHyper = (MyThread&Hyper);
 //	printf(" T%iC%i",MyThread,MyCore);
	for ( int TT = 0; TT < NSteps; ++TT )
	{
		Parallel* ptr = this;
		while ( ptr->pObj )
		{
			do
			{
				if ( !ptr->pObj ) break;
				int NStrips = ptr->Width / ptr->el_wid, TotalHei = ptr->Height;
				int BlockNY = 0, BlockNN = 0, StripsPerCore = 0, StripsPerThread = 0,
					StartX = 0, RangeX = 0;

				switch ( ptr->Mode&RUN_MASK )
				{
					case RUN_2D_YMARCH: 
						if ( !ptr->el_wid ) break; // skip non-existant shaders
						BlockNY = (ptr->Height + ptr->CellY - 1) / ptr->CellY;
						BlockNN = NStrips * BlockNY;
						StripsPerThread = (BlockNN + NThreads - 1) / NThreads;
						StripsPerCore = (StripsPerThread << Hyper);
						if ( fPrint )
							printf("\n--- T=%i BNY=%i BNN=%i NS=%i SPC=%i H=%i", MyThread, BlockNY, BlockNN, NStrips, StripsPerCore, Hyper);

						if ( Hyper )
							for ( int ST = MyCore * StripsPerCore;
							(ST < (MyCore + 1)*StripsPerCore) && ((ST + Hyper) < BlockNN);
									 ST += (Hyper + 1) )
						{
							StartX = (ST + MyHyper) % NStrips;
							const int Y = ((ST + MyHyper) / NStrips)*ptr->CellY;
							const cl_ulong StartOffs = ptr->el_wid * StartX * STRIDE + Y * ptr->DY;
							int  BlockHei = TotalHei - Y; BlockHei = __min(BlockHei, ptr->CellY);
							if ( fPrint ) 	printf("\n%2i, %4i, %4i, %4i, %8llu, %4i; ",
																		 MyThread, ST, StartX, Y, StartOffs, BlockHei);
							ptr->pObj->pF(ptr->Arg, StartOffs, BlockHei, StartX*ptr->el_wid, Y, LocalT);
						} else
							{
								int RNG = 1, EN = (MyCore + 1)*StripsPerCore;	EN = __min(EN, BlockNN);
								for ( int ST = MyCore * StripsPerCore; ST < EN; ST += RNG )
								{
									StartX = ST % NStrips;
									RNG = __min(NStrips - StartX, EN - ST);
									int Y = (ST / NStrips)*ptr->CellY;
									cl_ulong StartOffs = ptr->el_wid * StartX * STRIDE + Y * ptr->DY;
									int  BlockHei = TotalHei - Y; BlockHei = __min(BlockHei, ptr->CellY);
									if ( fPrint )	printf("\nTH=%2i, ST=%4i, X0=%4i, Y0=%4i, Offs=%8llu, H=%2i,  RNG=%3i; ",
																			 MyThread, ST, StartX, Y, StartOffs, BlockHei, RNG);
									
									ptr->pObj->pF(ptr->Arg, StartOffs, BlockHei | ((RNG - 1) << 16), StartX*ptr->el_wid, Y, LocalT);
								}
							}
						break;

					case RUN_1D: if ( !ptr->el_wid ) break; // skip non-existant shaders
						StartX = NStrips * MyThread / NThreads;
						RangeX = NStrips * (MyThread + 1) / NThreads - StartX;
						StartX *= ptr->el_wid;
						if(RangeX) ptr->pObj->pF(ptr->Arg, StartX*STRIDE, RangeX, StartX, ptr->Height, LocalT);
						break;

					case RUN_TASK_SINGLE: default: if ( MyThread ) break; // fallthrough
					case RUN_TASK_ALL: ptr->pObj->pF(ptr->Arg, MyThread, ptr->CellY, ptr->Width, ptr->Height, LocalT);

				}// switch mode

				if ( !MyThread )
				{
					const cl_ulong CurTime = Rdtsc();
					BaseTime = (CurTime - BaseTime);
					ptr->nRuns++; ptr->TimeSpent += BaseTime;
					if ( BaseTime > ptr->MaxTime ) ptr->MaxTime = BaseTime;
					if ( !(ptr->MinTime) || (BaseTime < ptr->MinTime) )  ptr->MinTime = BaseTime;
					BaseTime = CurTime;
				}

			} while ( (ptr++)->Mode & RUN_ASYNC );

#pragma omp barrier

			if ( fPrint ) printf("\n---BAR T=%2i ", MyThread);


			if ( (ptr - 1)->Mode&RUN_LAST_ITEM ) break;
		}// for shaders

		LocalT += StepDir;		
	} // for TT

	}// parallel
	IFLOG(LOG_KRNL) Print(1);
	return 0;
}// fn



#ifdef UNIT_TEST
#include "minicatch.hpp"

#endif

