/*
WaveTomography software implements the algorithm for solving direct and inverse problems of wave tomography in application to medical ultrasound imaging for breast cancer diagnosis.
The software has been developed at Lomonosov Moscow State University. Copyright (c) 2021 Seryozhnikov S.Y, Romanov S.Y.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "setup.h"


#include <cstdio>
#include <cstring>
#include <fcntl.h>

#include "k_defs.h"
#include "cdata.h"
#include "cmdl.h"

#ifndef TARGET_WINDOWS
#include <sys/time.h>
#include <sys/mman.h>
#include <unistd.h>
#endif

#ifdef USE_MPI
#include <mpi.h>
#endif

#include <omp.h>

OPTDEF G_Options[] = {
{         "//---Globals",OPTDEF::T_COMMENT,0,0 },
{"-log",      OPTDEF::T_INT,    1,  offsetof(GlobalSettings,LogLevel)      },
{"-GPU",      OPTDEF::T_INT,    1,  offsetof(GlobalSettings,GPU_device)  },
{"-NGPUs",    OPTDEF::T_INT,    1,  offsetof(GlobalSettings,GPUsPerNode)  },
{"",-1,0,0} };


#ifdef UNIT_TEST
#include "minicatch.hpp"
#endif

GlobalSettings G_Settings = { 0,0,1 };

GPUContext GPU;



int main(int argc, char** argv)
{

#ifdef TARGET_WINDOWS
	_fmode = O_BINARY;
#endif

	CData CD;

#ifdef USE_MPI
	MPI_Init((int*)(&argc), &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &NumProcesses);
	MPI_Comm_rank(MPI_COMM_WORLD, &MyId);
	sprintf(_PID, "\n[%02i] ", MyId);
#endif
		
	if(!MyId) puts("\n*** WaveTomography v." _WAVETM_VERSION_);
	
	G_Options->ParseFile("wavetm.ini", (char*)(&G_Settings));
	G_Options->ParseArgv(argc, argv, (char*)(&G_Settings), 1);
	
	IFLOG(LOG_MULTI)
	{
		if ( !MyId ) printf("\nMPI: %i processes ", NumProcesses);
		printf("%s %i threads ", _PID,(int)omp_get_max_threads());
	}

#ifdef UNIT_TEST
	if ( (argc >= 2) && !strncmp("--test", argv[1], 8) )
		return Catch::Session().Run(argc - 1, argv + 1);
#endif

	CData::Options->ParseFile("wavetm.ini", (char*)(&CD));
	CData::Options->ParseArgv(argc, argv, (char*)(&CD), 1);
	CD.EnableGPU = !(G_Settings.GPU_device & 0xFFFF0000);
	CD.Interactive();

	if ( CD.InitGPU(&GPU, G_Settings.GPU_device+(MyId%G_Settings.GPUsPerNode)) ) return 1;	

	int rv=CD.Process();	
	CHECKPOINT(" Processed. ")

#ifdef TARGET_GPU
	GPU.Close();
#endif

#ifdef USE_MPI
	MPI_Finalize();
#endif
	return rv;

};

